#ifndef matrix_h
#define matrix_h

#include <list>
#include <vector>

#include "ncutMemory.h"

/**
 * @brief Simple matrix class.
 *
 * @c Matrix is a simple quadratic matrix class containing all operations
 * needed by the ncut library.
 */
class Matrix{

  public:

    // functions for read & write access to data members ------
   
    /**
     * @brief the value of a cell of the matrix
     * @param row the row
     * @param col the column
     * @return the value of the specified cell
     */
    inline double& val(unsigned int row, unsigned int col)
    {return value_[row][col];};

    // functions for read-only access to data members ------

    /**
     * @brief read-only access to the values of a matrix cell
     * @param row the row
     * @param col the column
     * @return constant value of the specified cell
     */
    inline const double& val(unsigned int row, unsigned int col) const
    {return value_[row][col];};
    /**
     * @brief get the dimension of the matrix
     * @return the dimension of the matrix
     */
    inline unsigned int dim() const
    {return dimension_;};

    // class-specific functions ------

    /**
     * @brief Reset matrix to new dimension.
     *
     * @param dimension The new dimension of the matrix.
     *
     * Changes the dimension, deletes the old matrix and creates a new
     * one with the given dimension 
     */
    void changeDimension(const unsigned int dimension);

    /**
     * @brief Adds the matrix to the current matrix.
     *
     * @param m The matrix which should be added.
     */
    void add(const Matrix* m);
    /**
     * @brief Subtracts a matrix from the current matrix.
     *
     * @param m The matrix which should be subtracted.
     */
    void subtract(const Matrix* m);

    /**
     * @brief Returns the matrix in bandfrom which is needed for arpack++.
     *
     * @param n      The dimension of the matrix will be written into this
     *               variable.
     * @param nsdiag The number of the non-zero diagonals will be written 
     *               into this variable.
     * @return An array of doubles (which is needed for aprack++).
     */
    auto_array<double> band(int&n, int& nsdiag) const;

    /**
     * @brief Splits the matrix into two new matrices.
     *
     * @param a        One part will be written into this matrix.
     * @param b        The second part will be written into this matrix.
     * @param indexMap A mapping of index numbers in the old matrix to the
     *                 associated index in one of the new matrices.
     * @param indices  The list of indices in matrix @c a.
     * @return 0 if there was no error
     */
    int split(Matrix& a, Matrix& b, std::vector<unsigned int>& indexMap,
              std::list<unsigned int>& indices) const;

    /**
     * @brief Injects a small matrix into a larger one at an arbitrary position.
     *
     * @param m The @c Matrix to be injected.
     * @param index The position where @c m should be injected.
     * @return This object.
     *
     * @c Matrix m is copied into this @c Matrix starting at position
     * (@c index , @c index ). @c m does not have to fit inside this matrix,
     * but if it doesn't, only the values of @c m that do fit are copied.
     * All values in this matrix remain unchanged except the ones overwritten
     * by @c m.
     */
    Matrix& inject(const Matrix& m, unsigned int index);

    /**
     * @brief Transforms a vector.
     *
     * @param vec The vector to be transformed.
     *
     * The vector vec is multiplied from the left with this matrix. The size of
     * the vec array has to match the dimension of this matrix, otherwise the
     * result is undefined.
     */
    void transform(double* vec) const;

    /**
     * @brief Resets the matrix to have diag with dimension dim on its diagonal.
     *
     * @param diag The new diagonal vector of this matrix.
     * @param dim  The new dimension of this matrix.
     */
    void diag(const double* diag, unsigned int dim);

    /**
     * @brief Sets all elements in the matrix to the given value @c val.
     *
     * @param val All elements of the matrix will be set to this value.
     * @return This matrix.
     */
    Matrix& operator=(double val);
    /**
     * @brief Convience function for add().
     *
     * @param operand he matrix to be added.
     * @return This object.
     */
    Matrix& operator+=(const Matrix& operand);
    /**
     * @brief Convience function for subtract().
     *
     * @param operand The matrix to be subtracted.
     * @return This object.
     */
    Matrix& operator-=(const Matrix& operand);

    // standard functions ------

    /**
     * @brief Creates an empty matrix with the given dimension.
     *
     * @param dimension The dimension for the new matrix (default=0).
     */
    Matrix(unsigned int dimension=0);
    /**
     * @brief Create a new diagonal matrix.
     *
     * @param diag The vector that will be on the matrix's diagonal.
     * @param dim The dimension for the new matrix (default=0).
     *
     * Creates a matrix of dimension @c dim with @c diag on its diagonal
     */
    Matrix(const double *diag, unsigned int dim);
    /**
     * @brief %Matrix copy constructor.
     *
     * @param clone The matrix to be cloned.
     */
    Matrix(const Matrix& clone);
    /**
     * @brief %Matrix destructor.
     */
    virtual ~Matrix();

    /**
     * @brief %Matrix assignment operator.
     *
     * @param clone The matrix to be cloned.
     * @return This object.
     */
    virtual Matrix& operator=(const Matrix& clone);

  protected:

    double** value_; /**< All elements in the matrix*/
    unsigned int dimension_; /**< The dimension of the matrix*/
};

#endif
