#ifndef MOTIONPROFILE_H
#define MOTIONPROFILE_H 1

#include "ncutSetting.h"
#include "matrix.h"
#include "ncutImage.h"
#include "ncutException.h"

// #include <time.h>
#include <math.h>
#include <cstdlib>
#include <iostream>

// kommt dann weg:
// #define MAX_SEG_NUM 15

#define MAX_SEG_ID 2000000000

namespace ncut
{

const double PI = 3.14159265;

struct pxlInfo_t {
  double hueVecX;
  double hueVecY;
  double lit;
  unsigned int elm;
};

/**
 * @brief Profile handles similarity information of one frame window.
 *
 * The base class for @c ElementProfile and @c SegmentProfile.
 * A Profile contains Information about its associated type.
 * Segmentation is done on Elements (e.g. from a prioir segmentation step) 
 * using information stored in an @c ElementProfile and the information of the
 * resulting segments is stored in a @c SegmentProfile.
 * Elements and segments are identified by numbers unique in all of the
 * sequence. To get the index used for an element or a segment in a @c Profile
 * (e.g. the index of this element in the similarity matrix) use the function
 * @c idx(elementID).
 */
template <class IMG_ELM_TYPE = unsigned char,
          class MSK_ELM_TYPE = signed long>
class Profile
{
  public:

    // functions for read-only access to data members ------

    /**
     * @brief Similarity matrix of this profile.
     * @return the similarity matrix of this profile
     */
    inline const Matrix* similarity() const
    {return &similarity_;};
    /**
     * @brief Get the indices of the regions in this profile.
     * @return the indices of the regions
     */
    inline const std::map<MSK_ELM_TYPE, unsigned int>* indices() const
    {return &indices_;};
    /**
     * @brief Get the identifiers of the regions in this profile.
     * @return the identifiers of the regions
     */
    inline const std::vector<MSK_ELM_TYPE>* identifiers() const
    {return &identifiers_;};
    /**
     * @brief Get the index of a region identified by an id.
     * @param idx the id of the region
     * @return the index of the specified region
     */
    unsigned int idx(const MSK_ELM_TYPE& id) const;
    /**
     * @brief Get the identifier of the region with the given index.
     * @param idx the idex of the region
     * @return the id of the specified region
     */
    inline const MSK_ELM_TYPE& id(unsigned int idx) const
    {return identifiers_[idx];};

    // standard functions ------

    /**
     * @brief %Profile empty constructor.
     *
     * The resulting @c Profile will contain empty indices and identifier maps
     * and an empty similarity Matrix.
     */
    Profile(const Setting* setting);

    /**
     * @brief %Profile copy constructor
     *
     * @param clone The @c Profile to be cloned.
     */
    Profile(const Profile& clone);

    /**
     * @brief %Profile can be constructed by manually setting all members.
     *
     * @param similarity  A similarity matrix.
     * @param indices     An indices map.
     * @param identifiers An identifiers map.
     * @param setting     The parameters of the current segmentation.
     *
     * @c similarity , @c indices and @c identifiers must have the same
     * dimension and @c indices and @c identifiers must be valid.
     */
    Profile(const Matrix& similarity,
            const std::map<MSK_ELM_TYPE, unsigned int>& indices,
            const std::vector<MSK_ELM_TYPE>& identifiers,
            const Setting* setting);

    /**
     * @brief %Profile destructor.
     */
    virtual ~Profile();

    /**
     * @brief %Profile assignment operator.
     *
     * @param clone The @c Profile to be cloned.
     * @return This object.
     */
    virtual Profile& operator=(const Profile& clone);

  protected:

    Matrix similarity_; /**< The similarity matrix stores the weight
                        (=similarity) from every element to every
                        other element*/

    std::map<MSK_ELM_TYPE, unsigned int> indices_; /**< A map from global
        element identifiers that are consistently used in every iteration to
        the local indices these elements have in this profile (e.g. the
        index in the similarity matrix)*/
    std::vector<MSK_ELM_TYPE> identifiers_; /**< A map from local element
        indices to global element identifiers (inverse of @c indices_)*/

    const Setting* setting_; /**< Contains all needed parameters*/
};

/**
 * @brief ElementProfile handles the similarity information of the elements in a frame window.
 *
 * Segmentation is only done on a few frames (= frame window) of the sequence.
 * The frame window is moved over the whole sequence to get a complete
 * segmentation.
 * An @c ElementProfile is used to store information about the elements in one
 * frame window (= the frames that are used for one segmentation step). It is
 * also used to calculate the similarity between those elements, which is then
 * stored in the @c similarity matrix.
 * @c ElementProfile administers the frame and mask sequences.
 */
template <class IMG_ELM_TYPE = unsigned char,
          class MSK_ELM_TYPE = signed long>
class ElementProfile : public Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>
{
  public:

    // factor to normalize a IMG_ELM_TYPE
    const double normFac; /**< a constant factor to normalize a datatype of
        the image sequence to the interval [0,1]*/

    // functions for read-only access to data members ------

   /**
    * @brief Get the n-th frame in this frame window.
    * @param num the number of the frame
    * @return a pointer to the specified frame
    */
    inline const Image<IMG_ELM_TYPE>* frame(unsigned int num) const
    {return frameSeq_[num];};
   /**
    * @brief Get the n-th mask in this frame window.
    * @param num the number of the mask
    * @return a pointer the specified mask
    */
    inline const Image<MSK_ELM_TYPE>* mask(unsigned int num) const
    {return maskSeq_[num];};
   /**
    * @brief Get a pixel in the n-th mask of this frame window.
    * @param num the number of the mask
    * @param pxl the number of the pixel
    * @return the element number of the element the pixel belongs to
    */
    inline const MSK_ELM_TYPE& mask(unsigned int num, unsigned int pxl) const
    {return maskSeq_[num]->val(pxl);};

   /**
    * @brief Get a pixel in the n-th frame of this frame window
    * @param num the number of the frame
    * @param pxlelm e.g. to get channel c of pixel p of an n-channel image use p*n+c
    * @return the value of one channel of a pixel in the specified frame
    */
    inline const IMG_ELM_TYPE& frame(unsigned int num, unsigned int pxlelm) const
    {return frameSeq_[num]->val(pxlelm);};

    /**
     * @brief Get the width of the frames.
     * @return The width of the frames.
     */
    inline unsigned int width() const
    {return width_;};
    /**
     * @brief Get the height of the frames.
     * @return The height of the frames.
     */
    inline unsigned int height() const
    {return height_;};
    /**
     * @brief Get the number of channels of the frames.
     * @return The number of channels of the frames.
     */
    inline unsigned int nChl() const
    {return nChl_;};
    /**
     * @brief Get the number of frames in the current frame window.
     * @return The number of frames in the current frame window.
     * note: this may be smaller than wSize if the frame window is not completely filled yet.
     */
    inline unsigned int nFrm() const
    {return frameSeq_.length();};
    /**
     * @brief Get the preset size of the frame window in frames.
     * @return The preset size of the frame window in frames.
     */
    inline unsigned int wSize() const
    {return wSize_;};
    /**
     * @brief Get the number of elements in this profile.
     * @return The number of elements in this profile.
     */
    inline unsigned int nElm() const
    {return nElm_;};
    /**
     * @brief Get a list if pixels for each element.
     * @return A list of pixels for each element.
     * This function returns a vector containing vectors of pixel numbers.
     * Vector number i holds the pixel numbers of element i.
     */
    inline const std::vector<std::vector<unsigned int> >* pixels() const
    {return &pxls_;};
    /**
     * @brief Get the pixels of an element.
     * @param idx The index of the element.
     * @return A list of pixel numbers for an element.
     */
    inline const std::vector<unsigned int>* pixels(unsigned int idx) const
    {return &pxls_[idx];};
    /**
     * @brief Get the frame containing the element with index @c idx.
     * @param idx The index of the element.
     * @return The number of the frame containing the element.
     */
    unsigned int frameOf(unsigned int idx) const;
    /**
     * @brief Get the neighbourhood matrix of this profile.
     * @return The neighbourhood matrix of this profile.
     */
    inline const Matrix* neighbors() const
    {return &neighbors_;};
    /**
     * @brief Get the mean hue vector x coordinate of an element.
     * @param elm The index of the element.
     * @return The hue vector x coordinate of an element.
     */
    inline double hueVecX(unsigned int elm) const
    {return hueVecX_[elm];};
    /**
     * @brief Get the mean hue vector y coordinate of an element.
     * @param elm The index of the element.
     * @return The hue vector xycoordinate of an element.
     */
    inline double hueVecY(unsigned int elm) const
    {return hueVecY_[elm];};
    /**
     * @brief Get the mean lightness of an element.
     * @param elm The index of the element.
     * @return The lightness of an element.
     */
    inline double lightness(unsigned int elm) const
    {return lightness_[elm];};
    /**
     * @brief Get the mean saturation of an element.
     * @param elm The index of the element.
     * @return The saturation of an element.
     */
    inline double saturation(unsigned int elm) const
    {return saturation_[elm];};
    /**
     * @brief Get the mean intensity for one channel of an element.
     * @param elm The index of the element.
     * @param chl The colour channel.
     * @return The mean intensity of the element.
     */
    inline double intensity(unsigned int elm, unsigned int chl) const
    {return intensity_[elm][chl];};
    /**
     * @brief Get the mean location x-coordinate of an element.
     * @param elm The index of the element.
     * @return The mean location x-coordinate of an element (in pixels).
     */
    inline double locationX(unsigned int elm) const
    {return locationX_[elm];};
    /**
     * @brief Get the mean location y-coordinate of an element.
     * @param elm The index of the element.
     * @return The mean location y-coordinate of an element (in pixels).
     */
    inline double locationY(unsigned int elm) const
    {return locationY_[elm];};
    /**
     * @brief Get the mean location z-coordinate of an element (frame number).
     * @param elm The index of the element.
     * @return The mean location z-coordinate of an element (in frames).
     * This function returns the "mean" frame number in this frame window.
     * This number does not have to be an integer if the element spans multiple frames.
     */
    inline double locationZ(unsigned int elm) const
    {return locationZ_[elm];};
    /**
     * @brief Get the number of pixels of an element.
     * @param elm The index of the element.
     * @return The number of pixels of an element.
     */
    inline unsigned int nPixels(unsigned int elm) const
    {return nPixels_[elm];};
    /**
     * @brief Get the minimum element index in the given frame.
     * @param frm The number of the frame in tghe current frame window.
     * @return The minimum element index in the given frame.
     */
    inline unsigned int minIdx(unsigned int frm) const
    {return (frm == 0) ? 0 : maxElmIdx_[frm-1]+1;};
    /**
     * @brief Get the maximum element index in the given frame.
     * @param frm The number of the frame in tghe current frame window.
     * @return The maximum element index in the given frame.
     */
    inline unsigned int maxIdx(unsigned int frm) const
    {return maxElmIdx_[frm];};

    // class-specific functions ------

    /**
     * @brief Calculate element info of a new frame.
     *
     * @param frame An new frame for segmentation.
     * @param mask Segment mask from a prior segmentation step.
     * @return 1 if sequence is not long enough to start segmentation yet
     *         (smaller then @c wSize ), 0 otherwise
     *
     * Gathers all needed element info and calculates the similarity matrix.
     */
    int push(const Image<IMG_ELM_TYPE>* frame, const Image<MSK_ELM_TYPE>* mask);

    // standard functions ------

    /**
     * @brief %ElementProfile is constucted passing a frame window size.
     *
     * @param wSize The size of the frame window.
     * @param setting The parameters of the current segmentation.
     *
     * A new empty ElementProfile is constructed. Use @c push until the frame
     * window is filled to start info and similarity calculation.
     */
    ElementProfile(unsigned int wSize, const Setting* setting);

    /**
     * @brief %ElementProfile copy constructor.
     *
     * @param clone The @c ElementProfile to be cloned.
     */
    ElementProfile(const ElementProfile& clone);

    /**
     * @brief %ElementProfile destructor.
     */
    virtual ~ElementProfile();

    /**
     * @brief %ElementProfile assignment operator.
     *
     * @param clone The @c ElementProfile to be cloned.
     */
    virtual ElementProfile& operator=(const ElementProfile& clone);

  protected:

    Sequence<IMG_ELM_TYPE> frameSeq_; /**< contains all frames in the current
                                      frame window*/
    Sequence<MSK_ELM_TYPE> maskSeq_; /**< contains the segment masks of all
                                     frames in the current frame window*/

    unsigned int wSize_; /**< the size of the frame window (in frames)*/

    unsigned int width_; /**< the width of each image in the sequence in
                         pixels*/
    unsigned int height_; /**< the height of each image in the sequence in
                          pixels*/
    unsigned int nChl_; /**< the number of channels of each image in the
                        sequence*/
    unsigned int nElm_; /**< the number of elements in the current frame
                        window*/

    std::vector<std::vector<unsigned int> > pxls_; /**< a list of pixels in
                                                   each element*/

    std::vector<unsigned int> maxElmIdx_;/**< the highest used element index in
                                         each frame*/

    Matrix neighbors_;/**< stores the number of neighboring pixels of every
                      element to every other element*/

    // element info arrays
    double* hue_;           /**< the mean hue of each element*/
    double* hueVariation_;  /**< the hue variation of each element*/
    double* hueVecX_;       /**< the hue vector X component of each element*/
    double* hueVecY_;       /**< the hue vector Y component of each element*/
    double* lightness_;     /**< the mean lightness of each element*/
    double* litVariation_;  /**< the lightness variation of each element*/
    double* saturation_;    /**< the mean saturation of each element*/
    double* satVariation_;  /**< the saturation variation of each element*/
    double** intensity_;    /**< the mean intensity of each image channel*/
    double* locationX_;     /**< the center-of-gravity X coordinate*/
    double* locationY_;     /**< the center-of-gravity Y coordinate*/
    double* locationZ_;     /**< the center-of-gravity Z coordinate*/
    unsigned int* nPixels_; /**< the number of pixels in each element*/

  private:

    /**
     * @brief Adapt base data to the new frame window
     *
     * @note base data is @c indices_ , @c identifiers_ , @c maxElmIdx_ and
     *       @c pxls_.
     */
    void readaptBase();
    /**
     * @brief Get base info for one frame
     *
     * @param frmNo Info will be gathered from the frame with this number.
     *
     * @note base info is @c indices_ , @c identifiers_ , @c maxElmIdx_ ,
     *       @c pxls_ and the number of elements.
     */
    void chartElements(unsigned int frmNo);
    /**
     * @brief Initialize the element information arrays.
     */
    void initInfoArrays();
    /**
     * @brief Initialize all info.
     */
    void initInfo();
    /**
     * @brief Adapt all info to the new frame window (keep valid info).
     *
     * Keep info from previous frames that is still valid. Only needed info
     * will be calculated later on.
     */
    void readaptInfo();
    /**
     * @brief Gather all needed information from elements.
     *
     * @param from Info will be gathered from this frame on.
     * @param to   Info will be gathered until this frame.
     *
     * Calculate information from the images and segment masks in the current
     * frame window.
     */
    void getInfo(unsigned int from, unsigned int to);
    /**
     * @brief Calculate the similarity matrix using the gathered information.
     *
     * @param from Elements from this frame on will be calculated.
     * @param to   Elements until this frame on will be calculated.
     *
     * Calculates the similarity between the elements of frames @c from to
     * @c to and all other elements in the current frame window.
     */
    void calcSim(unsigned int from, unsigned int to);

    /**
     * @brief Frees all memory dynamically allocated directly by this object.
     */
    void cleanup();

    unsigned int nElm_old_; /**< the number of elements in the last frame
                            window*/
    unsigned int nElm_front_; /**< the number of elements in the first frame
                              of the last frame window (= the frame that was
                              removed from the frame window)*/
};

/**
 * @brief SegmentProfile handles the similarity information of the segments in a frame window.
 *
 * A @c SegmentProfile is used to store information about the segments
 * resulting of the segmentation of one frame window.
 * @c SegmentProfile is also used to determine the connections of segments from
 * this frame window to segments from the last frame window.
 * Throghout the sequence, segments are connected through their identifiers.
 * segments that appear in consecutive frames and have the same identifiers are
 * to be considered the same segment.
 */
template <class IMG_ELM_TYPE = unsigned char,
          class MSK_ELM_TYPE = signed long>
class SegmentProfile : public Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>
{
  public:

    // functions for read-only access to data members ------

    /**
     * @brief Get a list of elements for each segment in this profile.
     * @return A list of elements for each segment in this profile.
     * This function returns a vector containing vectors of element indices.
     * Vector i contains the element indices of the elements contained in the segment with index i.
     */
    inline const std::vector<std::vector<MSK_ELM_TYPE> >* elements() const
    {return &elements_;};
    /**
     * @brief Get a list of elements for a segment.
     * @param idx The index of the segment.
     * @return A list of elements for the given segment.
     */
    inline const std::vector<MSK_ELM_TYPE>* elements(unsigned int idx) const
    {return &elements_[idx];};
    /**
     * @brief Get the element profile for the elements contained in the segments.
     * @return The element profile for the elements contained in the segments.
     */
    inline const ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>* elmProfile() const
    {return elmProfile_;};
    /**
     * @brief Get the segment id for the element identified by @c elmId.
     * @param elmId The identifier of the element.
     * @return The segment id for an element identified by @c elmId.
     */
    const MSK_ELM_TYPE& mask(const MSK_ELM_TYPE& elmId) const;
    /**
     * @brief Get the segment index for the element with the given index.
     * @param elmIdx The index of the element.
     * @return The segment id for an element identified by @c elmId.
     */
    inline unsigned int idxMask(unsigned int elmIdx) const
    {return idxMask_[elmIdx];};
    /**
     * @brief Get the age of a segment.
     * @param segId The identifier of the segment.
     * @return The age of a segment with the given identifier.
     */
    unsigned int age(const MSK_ELM_TYPE& segId) const;
    /**
     * @brief Get the active/inactive status of a segment.
     * @param segId The identifier of the segment.
     * @return true if the segment is inactive, false otherwise.
     */
    bool inactive(const MSK_ELM_TYPE& segId) const;
    /**
     * @brief Get the active/inactive list.
     * @return A list of booleans indicating wether a segment is active or inactive
     * (true for inactive, false for active).
     */
    inline const std::map<MSK_ELM_TYPE, bool>* inactive() const
    {return &inactive_;};
    /**
     * @brief Get the age list.
     * @return A map from segment identifiers to integers indicating the age of a segment with that identifier.
     */
    inline const std::map<MSK_ELM_TYPE, unsigned int>* age() const
    {return &age_;};
    /**
     * @brief Get the most similar segment to the given segment.
     * @param segId The identifier of the segment.
     * @return The segment most similar to the given segment.
     */
    const MSK_ELM_TYPE& simSeg(const MSK_ELM_TYPE& segId) const;
    /**
     * @brief Get the number of elements in this segment profile.
     * @return The number of elements in this segment profile.
     */
    inline unsigned int nElm() const
    {return (elmProfile_==NULL)?0:elmProfile_->nElm();};
    /**
     * @brief Get the number of segments in this segment profile.
     * @return The number of segments in this segment profile.
     */
    inline unsigned int nSeg() const
    {return nSeg_;};
    /**
     * @brief Get the number of channels of the frames in this frame window.
     * @return The number of channels of the frames in this frame window.
     */
    inline unsigned int nChl() const
    {return nChl_;};
    /**
     * @brief Get the mean location x-coordinate of a segment.
     * @param seg The index of the segment.
     * @return The mean location x-coordinate of a segment (in pixels).
     */
    inline double locationX(unsigned int seg) const
    {return locationX_[seg];};
    /**
     * @brief Get the mean location y-coordinate of a segment.
     * @param seg The index of the segment.
     * @return The mean location y-coordinate of a segment (in pixels).
     */
    inline double locationY(unsigned int seg) const
    {return locationY_[seg];};
    /**
     * @brief Get the number of pixels in a segment
     * @param seg The index of the segment.
     * @return The number of pixels of the given segment.
     */
    inline unsigned int nPixels(unsigned int seg) const
    {return nPixels_[seg];};

    // class-specific functions ------

    /**
     * @brief Calculate the difference between two segments.
     *
     * @param idx1     Index of segment 1.
     * @param idx2     Index of segment 2.
     * @param elements A segmentation, this paramter is optional.
     * @return The difference between segment 1 and segment 2.
     *
     * If the segments don't touch, a maximal value of 500 is returned.
     * Optionally, a segmentation different from the segmentation in this
     * SegmentProfile can be specified.
     */
    double
    dif(unsigned int idx1, unsigned int idx2,
        const std::vector<std::vector<MSK_ELM_TYPE> >* elements = NULL) const;

    /**
     * @brief Calculate segment info of a new frame window and segmentation.
     *
     * @param elmProfile An @c ElementProfile.
     * @param segments A segmentation on the elements of @c elmProfile.
     * @return 0 if successful
     *
     * Calculate all needed segment info using information from @c elmProfile
     * and connect the new segments to the segments from the last frame window.
     */
    int push(const ElementProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>* elmProfile,
             const std::vector<std::vector<MSK_ELM_TYPE> >& segments);

    /**
     * @brief Calculate segment similarity.
     *
     * @return 0 if successful
     *
     * Calculate the similarity of every segment to every other segment.
     */
    int calcSim();

    // standard functions ------

    /**
     * @brief %SegmentProfile empty constructor.
     *
     * @param setting The parameters of the current segmentation.
     *
     * Construct an empty @c SegmentProfile.
     */
    SegmentProfile(const Setting* setting);

    /**
     * @brief %SegmentProfile copy constructor.
     *
     * @param clone The @c SegmentProfile to be cloned.
     */
    SegmentProfile(const SegmentProfile& clone);

    /**
     * @brief %SegmentProfile destructor.
     */
    virtual ~SegmentProfile();

    /**
     * @brief %SegmentProfile assignment operator.
     *
     * @param clone The @c SegmentProfile to be cloned.
     * @return This object.
     */
    virtual SegmentProfile& operator=(const SegmentProfile& clone);

  protected:

    std::vector<std::vector<MSK_ELM_TYPE> > elements_; /**< a list of
        elements for each segment*/
    const ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>* elmProfile_; /**< the
        @c ElementProfile containing information about the elements the
        segments consist of*/

    std::map<MSK_ELM_TYPE, MSK_ELM_TYPE> mask_; /**< a map from element
        identifiers to the segment identifers of the segments they are part of*/
    std::vector<unsigned int> idxMask_; /**< a map from element indices
        to the segment indices of the segments they are part of*/
    std::map<MSK_ELM_TYPE, unsigned int> age_; /**< the number of frames
        each segment has been active (= in the sequence) so far*/
    std::map<MSK_ELM_TYPE, bool> inactive_; /**< true if a segment from the
        last frame window wasn't assigned to any segment in this frame window*/
    std::map<MSK_ELM_TYPE, MSK_ELM_TYPE> simSeg_; /**< the most similar
        segment of each segment*/

    unsigned int nSeg_; /**< the number of segments in this frame window*/
    unsigned int nChl_; /**< the number of channels of the images*/
    MSK_ELM_TYPE nextSegId_;

    Matrix neighbors_; /**< the number of neigboring pixels of every segment
        to every other segment*/

    // segment info arrays
    double** intensity_;      /**< the mean intensity of each image channel*/
    double* locationX_;       /**< the center-of-gravity X coordinate*/
    double* locationY_;       /**< the center-of-gravity Y coordinate*/
    unsigned int* nElements_; /**< the number of elements in each segment*/
    unsigned int* nPixels_;   /**< the number of pixels in each segment*/

  private:

    /**
     * @brief Connect to last frame window's segments.
     *
     * @param last The @c SegmentProfile of the last frame window.
     *
     * For every segment, find a matching segment in the last frame window (=
     * last iteration of @c push ), or declare the segment as newly spawned by
     * giving it an unused identifier.
     */
    void attachIds(const SegmentProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>& last);

    /**
     * @brief Connect to the last frame window.
     *
     * @param last The @c SegmentProfile of the last frame window.
     *
     * Update the @c age_ , @c simSeg_ , @c mask_ , @c indices_ and
     * @c identifiers_ maps from the last frame window.
     */
    void attach(const SegmentProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>& last);

    /**
     * @brief Frees all memory dynamically allocated directly by this object.
     */
    void cleanup();
};


/* template implementations ------------------------------------------------- */

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
unsigned int Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    idx(const MSK_ELM_TYPE& id) const
{
  typename std::map<MSK_ELM_TYPE, unsigned int>::const_iterator elmIdx_p;

  elmIdx_p = indices_.find(id);
  if (elmIdx_p == indices_.end())
    throw ncutException("element not found");

  return elmIdx_p->second;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>::Profile(const Setting* setting)
  : similarity_(), indices_(), identifiers_(), setting_(setting)
{
  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>::Profile(const Profile& clone)
  : similarity_(), indices_(), identifiers_(), setting_(clone.setting_)
{
  *this = clone;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    Profile(const Matrix& similarity,
            const std::map<MSK_ELM_TYPE, unsigned int>& indices,
            const std::vector<MSK_ELM_TYPE>& identifiers,
            const Setting* setting)
  : similarity_(similarity), indices_(indices), identifiers_(identifiers),
    setting_(setting)
{
  if (similarity_.dim() == 0 ||
      similarity_.dim() != indices_.size() ||
      similarity_.dim() != identifiers_.size()) {

    throw ncutException("dimension = 0 or dimension mismatch");
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>::~Profile()
{
  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Profile<IMG_ELM_TYPE,MSK_ELM_TYPE>& Profile<IMG_ELM_TYPE,MSK_ELM_TYPE>::
    operator=(const Profile& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  if (!indices_.empty()) indices_.clear();
  if (!identifiers_.empty()) identifiers_.clear();

  // copy the similarity matrix
  this->similarity_.changeDimension(clone.similarity_.dim());
  this->similarity_ = clone.similarity_;

  // copy indices and identifiers maps
  this->indices_ = clone.indices_;
  this->identifiers_ = clone.identifiers_;

  // copy the setting
  this->setting_ = clone.setting_;

  return *this;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
unsigned int ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    frameOf(unsigned int idx) const
{
  for (unsigned int frm=0; frm<maxElmIdx_.size(); frm++)
    if (maxElmIdx_[frm] >= idx) return frm;

  throw ncutException("element index out of bounds");
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::readaptBase()
{
  #ifdef debugout
  std::cout << "readaptBase" << std::endl << std::flush;
  #endif
  nElm_front_ = maxElmIdx_.front() + 1;
  nElm_old_ = nElm_;

  // readapt the element index map
  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator indicesItr;
  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator eraseItr;
  for (indicesItr = this->indices_.begin();
       indicesItr != this->indices_.end(); ) {
    if (indicesItr->second < nElm_front_) {
      eraseItr = indicesItr++;
      this->indices_.erase(eraseItr);
    }
    else {
      indicesItr->second -= nElm_front_;
      indicesItr++;
    }
  }

  // readapt the element identifiers map
  this->identifiers_.resize(this->indices_.size());
  for (indicesItr = this->indices_.begin();
       indicesItr != this->indices_.end();
       indicesItr++) {
    this->identifiers_[indicesItr->second] = indicesItr->first;
  }

  // readapt the maximum element index list
  maxElmIdx_.erase(maxElmIdx_.begin());
  typename std::vector<unsigned int>::iterator maxElmIdxItr;
  for (maxElmIdxItr=maxElmIdx_.begin();
       maxElmIdxItr!=maxElmIdx_.end();
       maxElmIdxItr++) {
    (*maxElmIdxItr) -= nElm_front_;
  }

  // readapt the element pixels map
  typename std::vector<std::vector<unsigned int> >::iterator pxlsItr =
      pxls_.begin();
  for (unsigned int elm=0; elm <nElm_front_; elm++)
    pxlsItr++;
  pxls_.erase(pxls_.begin(), pxlsItr);

}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    chartElements(unsigned int frmNo)
{
  #ifdef debugout
  std::cout << "chartElements" << std::endl << std::flush;
#endif

  if (frmNo >= maskSeq_.length())
    throw ncutException("invalid frame number, corrupt mask sequence?");

  Image<MSK_ELM_TYPE>* maskImg = maskSeq_[frmNo];
  unsigned int maskPxlCount = width_*height_;
  MSK_ELM_TYPE currentElementName;
  unsigned int startNumber;
  if (maxElmIdx_.empty())
    startNumber = 0;
  else
    startNumber = maxElmIdx_.back()+1;
  unsigned int elmNumber = startNumber;

  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator firstOccurrence;

  for (unsigned int pxl=0; pxl<maskPxlCount; pxl++) {
    currentElementName = maskImg->val(pxl);
    firstOccurrence = this->indices_.find(currentElementName);
    if (firstOccurrence == this->indices_.end()) {
      // current element name not found
      this->identifiers_.push_back(currentElementName);
      this->indices_[currentElementName] = elmNumber++;
      this->pxls_.push_back(std::vector<unsigned int>());
      this->pxls_.back().push_back(pxl);
    }
    else if (firstOccurrence->second < startNumber) {
      // this is a duplicate element name from another frame (not allowed)
      throw ncutException("duplicate element name");
    }
    else {
      this->pxls_[firstOccurrence->second].push_back(pxl);
    }
  }
  maxElmIdx_.push_back(elmNumber-1);

  nElm_ = elmNumber;

  if (nElm_ <= 0)
    throw ncutException("no elements found");
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
    void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::initInfoArrays()
{
#ifdef debugout
  std::cout << "initInfoArrays" << std::endl << std::flush;
#endif

  // create element info arrays
  hue_          = new double [nElm_];
  hueVariation_ = new double [nElm_];
  hueVecX_      = new double [nElm_];
  hueVecY_      = new double [nElm_];
  lightness_    = new double [nElm_];
  litVariation_ = new double [nElm_];
  saturation_   = new double [nElm_];
  satVariation_ = new double [nElm_];
  intensity_    = new double* [nElm_];
  for (unsigned int elm=0; elm<nElm_; elm++)
    intensity_[elm] = new double[frameSeq_.nChl()];
  locationX_    = new double [nElm_];
  locationY_    = new double [nElm_];
  locationZ_    = new double [nElm_];
  nPixels_      = new unsigned int [nElm_];

  // init element info arrays to 0
  for (unsigned int elm=0; elm<nElm_; elm++) {
    hue_[elm]          = 0.0;
    hueVariation_[elm] = 0.0;
    hueVecX_[elm]      = 0.0;
    hueVecY_[elm]      = 0.0;
    lightness_[elm]    = 0.0;
    litVariation_[elm] = 0.0;
    saturation_[elm]   = 0.0;
    satVariation_[elm] = 0.0;
    for (unsigned int chl=0; chl<frameSeq_.nChl(); chl++)
      intensity_[elm][chl] = 0.0;
    locationX_[elm]     = 0.0;
    locationY_[elm]     = 0.0;
    locationZ_[elm]     = 0.0;
    nPixels_[elm]       = 0;
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::initInfo()
{
#ifdef debugout
  std::cout << "initInfo" << std::endl << std::flush;
#endif

  // initialize the matrices
  this->similarity_.changeDimension(nElm_);
  this->neighbors_.changeDimension(nElm_);

  // initialize the element info arrays
  initInfoArrays();
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::readaptInfo()
{
#ifdef debugout
  std::cout << "readaptInfo" << std::endl << std::flush;
#endif

  // readapt the similarity matrix and the neighborhood matrix
  Matrix firstFrameSim, otherFramesSim;
  Matrix firstFrameNeighb, otherFramesNeighb;
  Matrix firstFrameBordrs, otherFramesBordrs;

  std::vector<unsigned int> indexMap;
  std::list<unsigned int> firstFrameIndices;

  for (unsigned int elm=0; elm<nElm_front_; elm++)
    firstFrameIndices.push_back(elm);

  this->similarity_.split(firstFrameSim, otherFramesSim, indexMap,
                          firstFrameIndices);
  this->neighbors_.split(firstFrameNeighb, otherFramesNeighb, indexMap,
                         firstFrameIndices);

  this->similarity_.changeDimension(nElm_);
  this->neighbors_.changeDimension(nElm_);

  this->similarity_.inject(otherFramesSim, 0);
  this->neighbors_.inject(otherFramesNeighb, 0);

  // save pointers to the old element info arrays
  double* hue_old = hue_;
  double* hueVariation_old = hueVariation_;
  double* hueVecX_old = hueVecX_;
  double* hueVecY_old = hueVecY_;
  double* lightness_old = lightness_;
  double* litVariation_old = litVariation_;
  double* saturation_old = saturation_;
  double* satVariation_old = satVariation_;
  double** intensity_old = intensity_;
  double* locationX_old = locationX_;
  double* locationY_old = locationY_;
  double* locationZ_old = locationZ_;
  unsigned int* nPixels_old = nPixels_;

  // initialize the new element info arrays
  initInfoArrays();

  // copy the needed element information from the old arrays to the new ones
  // (and decrease the Z coordinate by one)
  unsigned int old_i, new_i;
  for (old_i=nElm_front_, new_i=0; old_i<nElm_old_; old_i++, new_i++) {
    hue_[new_i] = hue_old[old_i];
    hueVariation_[new_i] = hueVariation_old[old_i];
    hueVecX_[new_i] = hueVecX_old[old_i];
    hueVecY_[new_i] = hueVecY_old[old_i];
    lightness_[new_i] = lightness_old[old_i];
    litVariation_[new_i] = litVariation_old[old_i];
    saturation_[new_i] = saturation_old[old_i];
    satVariation_[new_i] = satVariation_old[old_i];
    for (unsigned int chl=0; chl<frameSeq_.nChl(); chl++) {
      intensity_[new_i][chl] = intensity_old[old_i][chl];
    }
    locationX_[new_i] = locationX_old[old_i];
    locationY_[new_i] = locationY_old[old_i];
    locationZ_[new_i] = locationZ_old[old_i] - 1.0;
    nPixels_[new_i] = nPixels_old[old_i];
  }

  // delete old element info arrays
  if (hue_old != NULL) {
    delete [] hue_old;
    hue_old = NULL;
  }
  if (hueVariation_old != NULL) {
    delete [] hueVariation_old;
    hueVariation_old = NULL;
  }
  if (hueVecX_old != NULL) {
    delete [] hueVecX_old;
    hueVecX_old = NULL;
  }
  if (hueVecY_old != NULL) {
    delete [] hueVecY_old;
    hueVecY_old = NULL;
  }
  if (lightness_old != NULL) {
    delete [] lightness_old;
    lightness_old = NULL;
  }
  if (litVariation_old != NULL) {
    delete [] litVariation_old;
    litVariation_old = NULL;
  }
  if (saturation_old != NULL) {
    delete [] saturation_old;
    saturation_old = NULL;
  }
  if (satVariation_old != NULL) {
    delete [] satVariation_old;
    satVariation_old = NULL;
  }
  if (intensity_old != NULL) {
    for (unsigned int elm=0; elm < nElm_old_; elm++) {
      if (intensity_old[elm] != NULL) delete [] intensity_old[elm];
    }
    delete [] intensity_old;
    intensity_old = NULL;
  }
  if (locationX_old != NULL) {
    delete [] locationX_old;
    locationX_old = NULL;
  }
  if (locationY_old != NULL) {
    delete [] locationY_old;
    locationY_old = NULL;
  }
  if (locationZ_old != NULL) {
    delete [] locationZ_old;
    locationZ_old = NULL;
  }
  if (nPixels_old != NULL) {
    delete [] nPixels_old;
    nPixels_old = NULL;
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    getInfo(unsigned int from, unsigned int to)
{
#ifdef debugout
  std::cout << "getInfo" << std::endl << std::flush;
#endif

  if (to >= frameSeq_.length() || from > to)
    throw ncutException("invalid frame range, corrupt sequences?");

  unsigned int varBins = this->setting_->variation_bins;
  double satMult = this->setting_->saturation_multiplier;

  if (varBins < 3 || varBins > IMAGE_DATATYPE_MAXVAL)
    throw ncutException(
        "invalid number of bins, must be in [3,IMAGE_DATATYPE_MAXVAL]" );
  if (satMult < 0.0)
    throw ncutException("invalid saturation multiplier, must be >= 0");

  // factor "number of bins" / "max. intensity value"
  const double binFac = normFac * varBins;

  // total number of pixels per frame
  const unsigned int pxlCount = width_*height_;

  // total number of components per frame
  const unsigned int cmpCount = pxlCount*frameSeq_.nChl();

  // pixel history (histPxl + pxlHistOffset = pxl)
  std::deque<pxlInfo_t> pxlHist;
  unsigned int pxlHistSize = (3*width_)+3; //setting_->pxlHist_size;
  unsigned int pxlHistOffset =  width_+1; //setting_->pxlHist_offset;
  if (pxlHistOffset >= pxlHistSize)
    throw ncutException(
        "invalid pixel history offset specified (offset >= size)" );

  // saturation histogram for calculation of saturation variation
  double** satBuffer = new double* [nElm_];
  for (unsigned int i=0; i<nElm_; i++) {
    satBuffer[i] = new double[varBins];
    for (unsigned int j=0; j<varBins; j++) {
      satBuffer[i][j] = 0.0;
    }
  }

  // lightness histogram for calculation of lightness variation
  double** litBuffer = new double* [nElm_];
  for (unsigned int i=0; i<nElm_; i++) {
    litBuffer[i] = new double[varBins];
    for (unsigned int j=0; j<varBins; j++) {
      litBuffer[i][j] = 0.0;
    }
  }

  // intensity values for one pixel (red = 0, green = 1, blue = 2)
  double intVal [frameSeq_.nChl()];

  // counters
  unsigned int elm;
  unsigned int chl;
  unsigned int cmp;
  unsigned int pxl;
  unsigned int col;
  unsigned int row;
  unsigned int histPxl;
  unsigned int histCol;
  unsigned int histRow;

  for (unsigned int frm=from; frm<=to; frm++) {

    // reset counters
    chl     = 0;
    cmp     = 0;
    pxl     = 0;
    col     = 0;
    row     = 0;
    histPxl = 0;
    histCol = 0;
    histRow = 0;

    // clear the pixel history
    pxlHist.clear();

    // get current image and mask
    Image<MSK_ELM_TYPE>* mask = maskSeq_[frm];
    Image<IMG_ELM_TYPE>* image = frameSeq_[frm];

    while (histPxl<pxlCount) {

      //---- do everything that only needs information of a single pixel
      if (cmp < cmpCount) {

        //-- get the element index of this pixel
        if (chl == 0) { // pixel start
          typename std::map<MSK_ELM_TYPE, unsigned int>::iterator currentElmItr
              = this->indices_.find(mask->val(pxl));
          if (currentElmItr == this->indices_.end())
            throw ncutCritical(
                "data corrupt: unknown element, index map is incomplete" );
          elm = currentElmItr->second;

          if (elm >= nElm_)
            throw ncutCritical(
                "data corrupt: invalid element index in index map" );
        }

        //-- gather raw information from each channel of the pixel
        IMG_ELM_TYPE curInt = image->val(cmp);
        intVal[chl] = curInt;
        intensity_[elm][chl] += curInt;

        //-- calulate pixel information
        if (chl == frameSeq_.nChl()-1) { // pixel end

          // location
          locationX_[elm] += col;
          locationY_[elm] += row;
          locationZ_[elm] += frm;

          // number of pixels
          nPixels_[elm]++;

          // lightness
          double pxlLightness = (0.2126*intVal[0]) +
                                (0.7152*intVal[1]) +
                                (0.0722*intVal[2]);
          lightness_[elm] += pxlLightness;

          // saturation
          double maxIntVal, minIntVal;
          maxIntVal = minIntVal = intVal[0];
          for (unsigned int i=1; i<frameSeq_.nChl(); i++) {
            if (intVal[i] > maxIntVal) maxIntVal = intVal[i];
            if (intVal[i] < minIntVal) minIntVal = intVal[i];
          }
          double pxlSaturation = maxIntVal - minIntVal;
          pxlSaturation *= satMult;
          if (pxlSaturation > 255.0) pxlSaturation = 255.0;

          saturation_[elm] += pxlSaturation;

          // hue and hue vectors
          double pxlHue, pxlHueVecX, pxlHueVecY;
          if (pxlSaturation != 0.0) {
            pxlHue = acos((intVal[0] - (intVal[1]/2.0) - (intVal[2]/2.0)) /
                sqrt((intVal[0]*intVal[0]) + (intVal[1]*intVal[1]) +
                     (intVal[2]*intVal[2]) - (intVal[0]*intVal[1]) -
                     (intVal[0]*intVal[2]) - (intVal[2]*intVal[1])));
            pxlHue = (intVal[2] > intVal[1]) ? ((2*PI) - pxlHue) : pxlHue;

            pxlHueVecX = pxlSaturation * cos(pxlHue);
            pxlHueVecY = pxlSaturation * sin(pxlHue);
          }
          else {
            pxlHueVecX = 0.0;
            pxlHueVecY = 0.0;
          }
          hueVecX_[elm] += pxlHueVecX;
          hueVecY_[elm] += pxlHueVecY;

          // fill saturation and lightness buffers for calulation of saturation
          // and lightness variation
          double binPos = (pxlSaturation * binFac) - 0.5;
          int binNum;
          if (binPos < 0.0) satBuffer[elm][0]++;
          else if (binPos >= (double) (varBins-1))
            satBuffer[elm][varBins-1]++;
          else {
            binNum = (int) binPos;
            binPos -= binNum;
            satBuffer[elm][binNum+1]+=binPos;
            satBuffer[elm][binNum]+=1.0-binPos;
          }
          binPos = (pxlLightness * binFac) - 0.5;
          if (binPos < 0.0) litBuffer[elm][0]++;
          else if (binPos >= (double) (varBins-1))
            litBuffer[elm][varBins-1]++;
          else {
            binNum = (int) binPos;
            binPos -= binNum;
            litBuffer[elm][binNum+1]+=binPos;
            litBuffer[elm][binNum]+=1.0-binPos;
          }

          // update the pixel history
          pxlInfo_t pxlInfo;
          pxlInfo.hueVecX = pxlHueVecX;
          pxlInfo.hueVecY = pxlHueVecY;
          pxlInfo.lit = pxlLightness;
          pxlInfo.elm = elm;

          if (pxl >= pxlHistSize) pxlHist.pop_back();
          pxlHist.push_front(pxlInfo);
        }
      }
      else {
        // reached the end of the image, push empty pixel infos to the history
        // (so its length remains the same) until edge detection and
        // neighborhood search are finished
        if (chl == frameSeq_.nChl()-1) {
          pxlInfo_t pxlInfo;
          pxlInfo.hueVecX = 0.0;
          pxlInfo.hueVecY = 0.0;
          pxlInfo.lit = 0.0;
          pxlInfo.elm = 0;

          if (pxl >= pxlHistSize) pxlHist.pop_back();
          pxlHist.push_front(pxlInfo);
        }
      }

      //---- do everything that needs information from more than one pixel
      //     (pixel history is needed)
      if (chl == frameSeq_.nChl()-1) { // pixel end

        //-- look for neighboring elements
        if (histPxl >= width_) { // neighbor pixel above ...
          unsigned int thisElm = pxlHist[pxlHistOffset].elm;
          unsigned int neighborElm = pxlHist[pxlHistOffset+width_].elm;
          if (neighborElm != thisElm) { // ... belongs to another element
            neighbors_.val(neighborElm,thisElm) =
                ++neighbors_.val(thisElm,neighborElm);
          }
        }
        if (histCol >= 1) { // neighbor to the left ...
          unsigned int thisElm = pxlHist[pxlHistOffset].elm;
          unsigned int neighborElm = pxlHist[pxlHistOffset+1].elm;
          if (neighborElm != thisElm) { // ... belongs to another element
            neighbors_.val(neighborElm,thisElm) =
                ++neighbors_.val(thisElm,neighborElm);
          }
        }
        if (frm >= 1 && pxl>=pxlHistOffset) { // neighbor in the last frame
          unsigned int thisElm = pxlHist[pxlHistOffset].elm;
          typename std::map<MSK_ELM_TYPE, unsigned int>::iterator lastFrmElmItr
              = this->indices_.find(maskSeq_[frm-1]->val(histPxl));
          if (lastFrmElmItr == this->indices_.end())
            throw ncutException(
                "data corrupt: invalid element index in index map" );
          unsigned int lastFrmElm = lastFrmElmItr->second;
          if (lastFrmElm != thisElm) {
            if (neighbors_.val(lastFrmElm,thisElm) < 0.5)
              neighbors_.val(lastFrmElm,thisElm) =
                  neighbors_.val(thisElm,lastFrmElm) = 1.0;
          }
        }
      }

      //---- update counters
      if (chl == frameSeq_.nChl()-1) { // pixel end
        if (pxl>=pxlHistOffset) {
          histPxl++;
          if (++histCol >= width_) {
            histCol = 0;
            histRow++;
          }
        }

        pxl++;

        if (++col >= width_) {
          col = 0;
          row++;
        }
      }

      if (++chl >= frameSeq_.nChl())
        chl = 0;

      cmp++;
    }
  }

  // calculate means of the accumulated element information
  for (unsigned int elm=minIdx(from); elm<=maxIdx(to); elm++) {

    // zero out everything if an element has 0 pixels
    if (nPixels_[elm]==0) {

      hue_[elm]          = 0.0;
      hueVecX_[elm]      = 0.0;
      hueVecY_[elm]      = 0.0;
      hueVariation_[elm] = 0.0;
      lightness_[elm]    = 0.0;
      saturation_[elm]   = 0.0;
      litVariation_[elm] = 0.0;
      satVariation_[elm] = 0.0;
      for (unsigned int chl=0; chl<frameSeq_.nChl(); chl++)
        intensity_[elm][chl] = 0;
      locationX_[elm]    = 0;
      locationY_[elm]    = 0;
      locationZ_[elm]    = 0;

      continue;
    }

    // calculate element hue
    if (hueVecX_[elm] > 0.0) {
      if (hueVecY_[elm] >= 0.0)
        hue_[elm] = atan(hueVecY_[elm] / hueVecX_[elm]);
      else
        hue_[elm] = atan(hueVecY_[elm] / hueVecX_[elm]) + (2*PI);
    }
    else if (hueVecX_[elm] < 0.0) {
      hue_[elm] = atan(hueVecY_[elm] / hueVecX_[elm]) + PI;
    }
    else { // hueVecX_[elm] == 0.0
      if (hueVecY_[elm] > 0.0)
        hue_[elm] = PI*0.5;
      else if (hueVecY_[elm] < 0.0)
        hue_[elm] = PI*1.5;
      else   //hueVecX_[elm] == 0.0 && hueVecY_[elm] == 0.0 (no hue at all)
        hue_[elm] = -1.0;
    }
    if (hue_[elm] > 0.0) hue_[elm] *= 180.0 / PI; // to degrees

    // calculate element hue variation
    if (saturation_[elm] < 0.01)
      // aussume no variation if there is (practically) no hue
      hueVariation_[elm] = 1.0;
    else
      // the variation is the length of the weighted mean vector
      hueVariation_[elm] = sqrt((hueVecX_[elm]*hueVecX_[elm]) +
                                (hueVecY_[elm]*hueVecY_[elm])  )
                           / saturation_[elm];
    hueVariation_[elm] = 1.0 - hueVariation_[elm];
    if (hueVariation_[elm] < 0.0) hueVariation_[elm] = 0.0;

    // calculate element hue vector
    hueVecX_[elm] /= nPixels_[elm];
    hueVecY_[elm] /= nPixels_[elm];

    // calculate element lightness and saturation
    lightness_[elm]  /= nPixels_[elm];
    saturation_[elm] /= nPixels_[elm];

    // calculate element lightness and saturation variation
    for (unsigned int bin=0; bin<varBins; bin++) {
      if (litBuffer[elm][bin] <= 0.0) continue;
      litVariation_[elm] += fabs( (((double)bin)/binFac) - lightness_[elm])
                            * litBuffer[elm][bin];
    }
    litVariation_[elm] /= nPixels_[elm];
    litVariation_[elm] /= IMAGE_DATATYPE_MAXVAL;

    for (unsigned int bin=0; bin<varBins; bin++) {
      if (satBuffer[elm][bin] <= 0.0) continue;
      satVariation_[elm] += fabs( (((double)bin)/binFac) - saturation_[elm])
                            * satBuffer[elm][bin];
    }
    satVariation_[elm] /= nPixels_[elm];
    satVariation_[elm] /= IMAGE_DATATYPE_MAXVAL;

    // calculate element intensity for each image channel
    for (unsigned int chl=0; chl<frameSeq_.nChl(); chl++) {
      intensity_[elm][chl] = intensity_[elm][chl]/nPixels_[elm];
    }

    // calculate element center of gravity location
    locationX_[elm] = locationX_[elm]/nPixels_[elm];
    locationY_[elm] = locationY_[elm]/nPixels_[elm];
    locationZ_[elm] = locationZ_[elm]/nPixels_[elm];
  }

  // delete lightness and saturation histograms
  if (litBuffer != NULL) {
    for (unsigned int elm=0; elm<nElm_; elm++) {
      if (litBuffer[elm] != NULL)
        delete [] litBuffer[elm];
    }
    delete [] litBuffer;
    litBuffer = NULL;
  }
  if (satBuffer != NULL) {
    for (unsigned int elm=0; elm<nElm_; elm++) {
      if (satBuffer[elm] != NULL)
        delete [] satBuffer[elm];
    }
    delete [] satBuffer;
    satBuffer = NULL;
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    calcSim(unsigned int from, unsigned int to)
{
#ifdef debugout
  std::cout << "calcSim" << std::endl << std::flush;
#endif

  if (to >= frameSeq_.length() || from > to)
    throw ncutException("invalid frame range, corrupt sequences?");

  // the weight of the intensity similarity
  double intWeight = this->setting_->intensity_weight;
  // the weight of the position similarity
  double posWeight = this->setting_->position_weight;
  // the falloff rate of the intensity similarity
  double falloff_int = this->setting_->intensity_falloff;//
  // the falloff rate of the position similarity
  double falloff_pos = this->setting_->distance_falloff;//0.1
  // true if the hue vectors should be stretched
  bool stretchSat = (this->setting_->stretch_saturation == 1);

  if (intWeight < 0.0 || intWeight > 1.0)
    throw ncutException("invalid intensity weight, must be in [0,1]");
  if (posWeight < 0.0 || posWeight > 1.0)
    throw ncutException("invalid position weight, must be in [0,1]");
  if (falloff_int < 0.000001)
    throw ncutException("invalid intensity falloff, must be >= 0.000001");
  if (falloff_pos < 0.000001)
    throw ncutException("invalid distance falloff, must be >= 0.000001");

  // the points that define the piecewise linear saturation stretching function
  double point0_x = this->setting_->stretch_point0_x;
  double point0_y = this->setting_->stretch_point0_y;
  double point1_x = this->setting_->stretch_point1_x;
  double point1_y = this->setting_->stretch_point1_y;
  double point2_x = this->setting_->stretch_point2_x;
  double point2_y = this->setting_->stretch_point2_y;
  double point3_x = this->setting_->stretch_point3_x;
  double point3_y = this->setting_->stretch_point3_y;

  if (point0_x < 0.0 || point0_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 0 coordinate x");
  if (point0_y < 0.0 || point0_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 0 coordinate y");
  if (point1_x <= point0_x || point1_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 1 coordinate x");
  if (point1_y < 0.0 || point1_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 1 coordinate y");
  if (point2_x <= point1_x || point2_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 2 coordinate x");
  if (point2_y < 0.0 || point2_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 2 coordinate y");
  if (point3_x <= point2_x || point3_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 3 coordinate x");
  if (point3_y < 0.0 || point3_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 3 coordinate y");

  // factor 100 / "maximum possible image intensity value"
  const double perctFac = 100.0*normFac;

  // intensity and position similarity
  double intSim;
  double posSim;

  unsigned int firstIdx = minIdx(from);
  unsigned int lastIdx = maxIdx(to);

  for (unsigned int i=0; i<=lastIdx; i++) {
    this->similarity_.val(i,i) = 0.0;
    for (unsigned int j=(i<firstIdx)?firstIdx:(i+1); j<=lastIdx; j++) {

      if (intWeight > 0.0) {
        double hueVecX_i, hueVecY_i, hueVecX_j, hueVecY_j;
        if (stretchSat) {
          double sat_i = saturation_[i];
          double sat_j = saturation_[j];

          double interPos;
          double stretch_i;
          if (sat_i <= point1_x) {
            interPos = (sat_i-point0_x) / (point1_x-point0_x);
            stretch_i = (interPos*point1_y) + (1-interPos)*point0_y;
          }
          else if (sat_i <= point2_x) {
            interPos = (sat_i-point1_x) / (point2_x-point1_x);
            stretch_i = (interPos*point2_y) + (1-interPos)*point1_y;
          }
          else {
            interPos = (sat_i-point2_x) / (point3_x-point2_x);
            stretch_i = (interPos*point3_y) + (1-interPos)*point2_y;
          }

          double stretch_j;
          if (sat_j <= point1_x) {
            interPos = (sat_j-point0_x) / (point1_x-point0_x);
            stretch_j = (interPos*point1_y) + (1-interPos)*point0_y;
          }
          else if (sat_j <= point2_x) {
            interPos = (sat_j-point1_x) / (point2_x-point1_x);
            stretch_j = (interPos*point2_y) + (1-interPos)*point1_y;
          }
          else {
            interPos = (sat_j-point2_x) / (point3_x-point2_x);
            stretch_j = (interPos*point3_y) + (1-interPos)*point2_y;
          }

          if (sat_i > 0.0) {
            hueVecX_i = hueVecX_[i] * (stretch_i/sat_i);
            hueVecY_i = hueVecY_[i] * (stretch_i/sat_i);
          }
          else {
            hueVecX_i = 0.0;
            hueVecY_i = 0.0;
          }
          if (sat_j > 0.0) {
            hueVecX_j = hueVecX_[j] * (stretch_j/sat_j);
            hueVecY_j = hueVecY_[j] * (stretch_j/sat_j);
          }
          else {
            hueVecX_j = 0.0;
            hueVecY_j = 0.0;
          }
        }
        else {
          hueVecX_i = hueVecX_[i];
          hueVecY_i = hueVecY_[i];
          hueVecX_j = hueVecX_[j];
          hueVecY_j = hueVecY_[j];
        }

        double hueSatDifX = fabs(hueVecX_i - hueVecX_j);
        hueSatDifX *= perctFac;

        double hueSatDifY = fabs(hueVecY_i - hueVecY_j);
        hueSatDifY *= perctFac;

        double litDif = fabs(lightness_[i] - lightness_[j]);
        litDif *= perctFac;

        double distance = sqrt(hueSatDifX*hueSatDifX +
                               hueSatDifY*hueSatDifY +
                               litDif*litDif);
        intSim = sqrt(200.0*200.0 +
                      200.0*200.0 +
                      100.0*100.0) - distance;
        // to be tested (instead of the upper intSim)
//         double sigma = falloff_int * 100.0;
//         intSim = exp(-1.0 * pow(distance/sigma, 2.0));
        if (intSim < 0.0) intSim = 0.0;
        intSim = (1.0-intWeight)+(intSim*intWeight);
      }
      else
        intSim = 1.0;


      if (posWeight > 0.0) {
        if (neighbors_.val(i,j) > 0.5)
          posSim = 1.0;
        else {
          double sigma = falloff_pos * ((width_ < height_) ? width_:height_);
          double distance = sqrt((fabs(locationX_[i] - locationX_[j])*
                                  fabs(locationX_[i] - locationX_[j])) +
                                 (fabs(locationY_[i] - locationY_[j])*
                                  fabs(locationY_[i] - locationY_[j])) +
                                 (fabs(locationZ_[i] - locationZ_[j])*
                                  fabs(locationZ_[i] - locationZ_[j])));
          posSim = exp(-1.0 * pow(distance/sigma, 2.0));
        }
        if (posSim < 0.0) posSim = 0.0;
        posSim = (1.0-posWeight)+(posSim*posWeight);
      }
      else
        posSim = 1.0;

      this->similarity_.val(i,j) = this->similarity_.val(j,i) = intSim * posSim;
    }
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
int ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    push(const Image<IMG_ELM_TYPE>* frame, const Image<MSK_ELM_TYPE>* mask)
{
  if (frame == NULL || mask == NULL     ||
      frame->height() != mask->height() ||
      frame->width()  != mask->width()  ||
      mask->nChl()    != 1              ||
      frame->nChl()   != 3)
    throw ncutException("wrong image format");

  if (frameSeq_.length() == 0) {
    this->width_  = frame->width();
    this->height_ = frame->height();
    this->nChl_   = frame->nChl();
  }

  try {
    if (frameSeq_.length() < wSize_) {
      frameSeq_.pushFrame(*frame);
      maskSeq_.pushFrame(*mask);
    }
    else {
      frameSeq_.pushThrough(*frame);
      maskSeq_.pushThrough(*mask);
    }
  }
  catch(ncutException& e) {
    if (frameSeq_.length() > maskSeq_.length()) frameSeq_.popFrame();
    throw ncutException(
        (std::string("wrong image format : ")+e.what()).c_str() );
  }

  if (frameSeq_.length() < wSize_) {
    chartElements(maskSeq_.length()-1);

    return 1;
  }
  else {
    if (this->similarity_.dim() == 0) {
      chartElements(maskSeq_.length()-1);
      initInfo();
      getInfo(0, frameSeq_.length()-1);
      calcSim(0, frameSeq_.length()-1);
    }
    else {
      readaptBase();
      chartElements(maskSeq_.length()-1);
      readaptInfo();
      getInfo(frameSeq_.length()-1, frameSeq_.length()-1);
      calcSim(frameSeq_.length()-1, frameSeq_.length()-1);
    }

    return 0;
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    ElementProfile(unsigned int wSize, const Setting* setting)
  : Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>(setting),
    normFac(1.0/IMAGE_DATATYPE_MAXVAL),
    frameSeq_(), maskSeq_(),
    wSize_(wSize), width_(0), height_(0), nChl_(0),
    nElm_(0), pxls_(), maxElmIdx_(), neighbors_(),
    hue_(NULL), hueVariation_(NULL), hueVecX_(NULL), hueVecY_(NULL),
    lightness_(NULL), litVariation_(NULL), saturation_(NULL),
    satVariation_(NULL), intensity_(NULL), locationX_(NULL), locationY_(NULL),
    locationZ_(NULL), nPixels_(NULL), nElm_old_(0), nElm_front_(0)

{
  if (wSize_ < 2 || wSize_ % 2 != 1)
    throw ncutException("invalid window size, wSize must be >= 2 and odd");

  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    ElementProfile(const ElementProfile& clone)
  : Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>(clone.setting_),
    normFac(1.0/IMAGE_DATATYPE_MAXVAL),
    frameSeq_(NULL), maskSeq_(NULL),
    wSize_(0), width_(0), height_(0), nChl_(0), nElm_(0),
    pxls_(), maxElmIdx_(), neighbors_(),
    hue_(NULL), hueVariation_(NULL), hueVecX_(NULL), hueVecY_(NULL),
    lightness_(NULL), litVariation_(NULL), saturation_(NULL),
    satVariation_(NULL), intensity_(NULL), locationX_(NULL), locationY_(NULL),
    locationZ_(NULL), nPixels_(NULL), nElm_old_(0), nElm_front_(0)

{
  *this = clone;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::cleanup()
{
  if (hue_ != NULL) {
    delete [] hue_;
    hue_ = NULL;
  }
  if (hueVariation_ != NULL) {
    delete [] hueVariation_;
    hueVariation_ = NULL;
  }
  if (hueVecX_ != NULL) {
    delete [] hueVecX_;
    hueVecX_ = NULL;
  }
  if (hueVecY_ != NULL) {
    delete [] hueVecY_;
    hueVecY_ = NULL;
  }
  if (lightness_ != NULL) {
    delete [] lightness_;
    lightness_ = NULL;
  }
  if (litVariation_ != NULL) {
    delete [] litVariation_;
    litVariation_ = NULL;
  }
  if (saturation_ != NULL) {
    delete [] saturation_;
    saturation_ = NULL;
  }
  if (satVariation_ != NULL) {
    delete [] satVariation_;
    satVariation_ = NULL;
  }
  if (intensity_ != NULL) {
    for (unsigned int i=0; i < nElm_; i++) {
      if (intensity_[i] != NULL) delete [] intensity_[i];
    }
    delete [] intensity_;
    intensity_ = NULL;
  }
  if (locationX_ != NULL) {
    delete [] locationX_;
    locationX_ = NULL;
  }
  if (locationY_ != NULL) {
    delete [] locationY_;
    locationY_ = NULL;
  }
  if (locationZ_ != NULL) {
    delete [] locationZ_;
    locationZ_ = NULL;
  }
  if (nPixels_ != NULL) {
    delete [] nPixels_;
    nPixels_ = NULL;
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::~ElementProfile()
{
  cleanup();
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>&
ElementProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>::
    operator=(const ElementProfile& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  cleanup();

  // copy the similarity matrix
  this->similarity_ = clone.similarity_;

  // copy indices and identifiers
  this->indices_ = clone.indices_;
  this->identifiers_ = clone.identifiers_;

  // copy the parameters from clone
  this->frameSeq_ = clone.frameSeq_;
  this->maskSeq_  = clone.maskSeq_;

  this->width_  = clone.width_;
  this->height_ = clone.height_;
  this->nChl_   = clone.nChl_;
  this->wSize_  = clone.wSize_;
  this->nElm_   = clone.nElm_;

  this->pxls_ = clone.pxls_;
  this->maxElmIdx_ = clone.maxElmIdx_;

  this->neighbors_ = clone.neighbors_;

  // copy the element information
  if (clone.hue_ != NULL) {
    this->hue_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->hue_[elm] = clone.hue_[elm];
    }
  }
  else {
    this->hue_ = NULL;
  }

  if (clone.hueVariation_ != NULL) {
    this->hueVariation_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->hueVariation_[elm] = clone.hueVariation_[elm];
    }
  }
  else {
    this->hueVariation_ = NULL;
  }

  if (clone.hueVecX_ != NULL) {
    this->hueVecX_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->hueVecX_[elm] = clone.hueVecX_[elm];
    }
  }
  else {
    this->hueVecX_ = NULL;
  }

  if (clone.hueVecY_ != NULL) {
    this->hueVecY_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->hueVecY_[elm] = clone.hueVecY_[elm];
    }
  }
  else {
    this->hueVecY_ = NULL;
  }

  if (clone.lightness_ != NULL) {
    this->lightness_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->lightness_[elm] = clone.lightness_[elm];
    }
  }
  else {
    this->lightness_ = NULL;
  }

  if (clone.litVariation_ != NULL) {
    this->litVariation_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->litVariation_[elm] = clone.litVariation_[elm];
    }
  }
  else {
    this->litVariation_ = NULL;
  }

  if (clone.saturation_ != NULL) {
    this->saturation_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->saturation_[elm] = clone.saturation_[elm];
    }
  }
  else {
    this->saturation_ = NULL;
  }

  if (clone.satVariation_ != NULL) {
    this->satVariation_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->satVariation_[elm] = clone.satVariation_[elm];
    }
  }
  else {
    this->satVariation_ = NULL;
  }

  if (clone.intensity_ != NULL) {
    this->intensity_ = new double* [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      if (clone.intensity_[elm] == NULL) {
        this->intensity_[elm] = NULL;
        continue;
      }
      else {
        this->intensity_[elm] = new double [clone.frameSeq_.nChl()];
        for (unsigned int chl=0; chl<clone.frameSeq_.nChl(); chl++)
          this->intensity_[elm][chl] = clone.intensity_[elm][chl];
      }
    }
  }
  else {
    this->intensity_ = NULL;
  }

  if (clone.locationX_ != NULL) {
    this->locationX_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->locationX_[elm] = clone.locationX_[elm];
    }
  }
  else {
    this->locationX_ = NULL;
  }

  if (clone.locationY_ != NULL) {
    this->locationY_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->locationY_[elm] = clone.locationY_[elm];
    }
  }
  else {
    this->locationY_ = NULL;
  }

  if (clone.locationZ_ != NULL) {
    this->locationZ_ = new double [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->locationZ_[elm] = clone.locationZ_[elm];
    }
  }
  else {
    this->locationZ_ = NULL;
  }

  if (clone.nPixels_ != NULL) {
    this->nPixels_ = new unsigned int [clone.nElm_];
    for (unsigned int elm=0; elm<clone.nElm_; elm++) {
      this->nPixels_[elm] = clone.nPixels_[elm];
    }
  }
  else {
    this->nPixels_ = NULL;
  }

  // copy the setting
  this->setting_ = clone.setting_;

  return *this;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
const MSK_ELM_TYPE& SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    mask(const MSK_ELM_TYPE& elmId) const
{
  typename std::map<MSK_ELM_TYPE, MSK_ELM_TYPE>::const_iterator mask_p;

  mask_p = mask_.find(elmId);
  if (mask_p == mask_.end())
    throw ncutException("element not found");

  return mask_p->second;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
unsigned int SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    age(const MSK_ELM_TYPE& segId) const
{
  typename std::map<MSK_ELM_TYPE, unsigned int>::const_iterator age_p;

  age_p = age_.find(segId);
  if (age_p == age_.end())
    throw ncutException("segment not found");

  return age_p->second;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
bool SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    inactive(const MSK_ELM_TYPE& segId) const
{
  typename std::map<MSK_ELM_TYPE, bool>::const_iterator inactive_p;

  inactive_p = inactive_.find(segId);
  if (inactive_p == inactive_.end())
    throw ncutException("segment not found");

  return inactive_p->second;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
const MSK_ELM_TYPE& SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    simSeg(const MSK_ELM_TYPE& segId) const
{
  typename std::map<MSK_ELM_TYPE, MSK_ELM_TYPE>::const_iterator simSeg_p;

  simSeg_p = simSeg_.find(segId);
  if (simSeg_p == simSeg_.end())
    throw ncutException("segment not found");

  return simSeg_p->second;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
double SegmentProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>::
    dif(unsigned int idx1, unsigned int idx2,
        const std::vector<std::vector<MSK_ELM_TYPE> >* elements) const
{
  if (idx1 >= nSeg_ || idx2 >= nSeg_)
    throw ncutException("segment index out of range");
  if (idx1 == idx2)
    throw ncutException("segment indices must be different");

  if (elements == NULL)
    elements = &this->elements_;

  // true if the hue vectors should be stretched
  bool stretchSat = (this->setting_->stretch_saturation_merge == 1);

  // the points that define the piecewise linear saturation stretching function
  double point0_x = this->setting_->stretch_point0_x;
  double point0_y = this->setting_->stretch_point0_y;
  double point1_x = this->setting_->stretch_point1_x;
  double point1_y = this->setting_->stretch_point1_y;
  double point2_x = this->setting_->stretch_point2_x;
  double point2_y = this->setting_->stretch_point2_y;
  double point3_x = this->setting_->stretch_point3_x;
  double point3_y = this->setting_->stretch_point3_y;

  if (point0_x < 0.0 || point0_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 0 coordinate x");
  if (point0_y < 0.0 || point0_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 0 coordinate y");
  if (point1_x <= point0_x || point1_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 1 coordinate x");
  if (point1_y < 0.0 || point1_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 1 coordinate y");
  if (point2_x <= point1_x || point2_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 2 coordinate x");
  if (point2_y < 0.0 || point2_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 2 coordinate y");
  if (point3_x <= point2_x || point3_x > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 3 coordinate x");
  if (point3_y < 0.0 || point3_y > IMAGE_DATATYPE_MAXVAL)
    throw ncutException("invalid stretch point 3 coordinate y");

  // factor 100 / "maximum possible image intensity value"
  const double perctFac = 100.0*elmProfile_->normFac;

  // check if the segments touch
  unsigned int neighborPxls = 0;
  typename std::vector<MSK_ELM_TYPE>::const_iterator elm1;
  typename std::vector<MSK_ELM_TYPE>::const_iterator elm2;

  for (elm1  = (*elements)[idx1].begin();
       elm1 != (*elements)[idx1].end(); elm1++) {
    unsigned int elm1Idx = elmProfile_->idx(*elm1);

    for (elm2  = (*elements)[idx2].begin();
         elm2 != (*elements)[idx2].end(); elm2++) {
      unsigned int elm2Idx = elmProfile_->idx(*elm2);

      // add the number of neighboring pixels
      if (elmProfile_->neighbors()->val(elm1Idx,elm2Idx) > 0.5)
        neighborPxls += static_cast<unsigned int>
            (elmProfile_->neighbors()->val(elm1Idx,elm2Idx));
    }
  }

  // return with a maximum difference if they do not touch
  if (neighborPxls == 0)
    return 500.0; // calculated difference should not be bigger than 300

  // add the differences between all elements of segment 1
  // and all elements of segment 2
  double cutIntDif = 0.0;
  unsigned int pxlCount = 0;

  for (elm1  = (*elements)[idx1].begin();
       elm1 != (*elements)[idx1].end(); elm1++) {
    unsigned int elm1Idx = elmProfile_->idx(*elm1);

    for (elm2  = (*elements)[idx2].begin();
         elm2 != (*elements)[idx2].end(); elm2++) {
      unsigned int elm2Idx = elmProfile_->idx(*elm2);

      // stretch the hue vectors
      double hueVecX_i, hueVecY_i, hueVecX_j, hueVecY_j;
      if (stretchSat) {
        double sat_i = elmProfile_->saturation(elm1Idx);
        double sat_j = elmProfile_->saturation(elm2Idx);

        double interPos;
        double stretch_i;
        if (sat_i <= point1_x) {
          interPos = (sat_i-point0_x) / (point1_x-point0_x);
          stretch_i = (interPos*point1_y) + (1-interPos)*point0_y;
        }
        else if (sat_i <= point2_x) {
          interPos = (sat_i-point1_x) / (point2_x-point1_x);
          stretch_i = (interPos*point2_y) + (1-interPos)*point1_y;
        }
        else {
          interPos = (sat_i-point2_x) / (point3_x-point2_x);
          stretch_i = (interPos*point3_y) + (1-interPos)*point2_y;
        }

        double stretch_j;
        if (sat_j <= point1_x) {
          interPos = (sat_j-point0_x) / (point1_x-point0_x);
          stretch_j = (interPos*point1_y) + (1-interPos)*point0_y;
        }
        else if (sat_j <= point2_x) {
          interPos = (sat_j-point1_x) / (point2_x-point1_x);
          stretch_j = (interPos*point2_y) + (1-interPos)*point1_y;
        }
        else {
          interPos = (sat_j-point2_x) / (point3_x-point2_x);
          stretch_j = (interPos*point3_y) + (1-interPos)*point2_y;
        }

        if (sat_i > 0.0) {
          hueVecX_i = elmProfile_->hueVecX(elm1Idx) * (stretch_i/sat_i);
          hueVecY_i = elmProfile_->hueVecY(elm1Idx) * (stretch_i/sat_i);
        }
        else {
          hueVecX_i = 0.0;
          hueVecY_i = 0.0;
        }
        if (sat_j > 0.0) {
          hueVecX_j = elmProfile_->hueVecX(elm2Idx) * (stretch_j/sat_j);
          hueVecY_j = elmProfile_->hueVecY(elm2Idx) * (stretch_j/sat_j);
        }
        else {
          hueVecX_j = 0.0;
          hueVecY_j = 0.0;
        }
      }
      else {
        hueVecX_i = elmProfile_->hueVecX(elm1Idx);
        hueVecY_i = elmProfile_->hueVecY(elm1Idx);
        hueVecX_j = elmProfile_->hueVecX(elm2Idx);
        hueVecY_j = elmProfile_->hueVecY(elm2Idx);
      }

      // calculate hue vector and lightness difference
      double hueSatDifX = fabs(hueVecX_i - hueVecX_j);
      hueSatDifX *= perctFac;

      double hueSatDifY = fabs(hueVecY_i - hueVecY_j);
      hueSatDifY *= perctFac;

      double litDif = fabs(elmProfile_->lightness(elm1Idx) -
                           elmProfile_->lightness(elm2Idx));
      litDif *= perctFac;

      // calculate total difference
      double intDif = sqrt(hueSatDifX*hueSatDifX +
                           hueSatDifY*hueSatDifY +
                           litDif*litDif);

      // weigh total difference with the number of pixels in both elements
      unsigned int pxls = elmProfile_->nPixels(elm1Idx) +
                          elmProfile_->nPixels(elm2Idx);
      cutIntDif += intDif * pxls;
      pxlCount += pxls;
    }
  }

  return cutIntDif / pxlCount;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
int SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::calcSim()
{
  this->similarity_.changeDimension(nSeg_);

  try {
    for (unsigned int seg1=0; seg1<nSeg_; seg1++) {
      for (unsigned int seg2=seg1+1; seg2<nSeg_; seg2++) {
        this->similarity_.val(seg2,seg1) = this->similarity_.val(seg1,seg2) =
          500.0 - dif(seg1, seg2);
      }
    }
  }
  catch (ncutException& e) {
    this->similarity_.changeDimension(0);
    throw;
  }

  return 0;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    attachIds(const SegmentProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>& last)
{
  unsigned int adultAge = this->setting_->segment_adult_age;
  double ageWeight = this->setting_->segment_age_weight;
  double attachSim = this->setting_->min_segment_attach_similarity;

  if (adultAge == 0)
    throw ncutException("invalid segment adult age, must be >= 1.");
  if (ageWeight < 0.0 || ageWeight > 1.0)
    throw ncutException("invalid segment age weight, must be in [0,1].");
  if (attachSim < 0.0 || attachSim > 1.0)
    throw ncutException(
        "invalid minimum segment attach similarity, must be in [0,1].");

  // calculate the pixel area based similarity between the new segments and
  // the ones from the last iteration
  double similarity[last.elements_.size()][elements_.size()];
  unsigned int totalPxlArea_last[last.elements_.size()];
  unsigned int totalPxlArea[elements_.size()];

  for (unsigned int seg_last=0; seg_last<last.elements_.size(); seg_last++)
    for (unsigned int seg=0; seg<elements_.size(); seg++)
      similarity[seg_last][seg] = 0.0;

  for (unsigned int seg_last=0; seg_last<last.elements_.size(); seg_last++)
    totalPxlArea_last[seg_last] = 0;

  for (unsigned int seg=0; seg<elements_.size(); seg++)
    totalPxlArea[seg] = 0;


  // accumulate overlapping pixel areas of pairs of segments
  typename std::vector<std::vector<MSK_ELM_TYPE> >::iterator seg;
  typename std::vector<MSK_ELM_TYPE>::iterator elmId;
  unsigned int segIdx;
  unsigned long pxlArea;
  for (seg  = elements_.begin(), segIdx=0;
       seg != elements_.end(); seg++, segIdx++) {
    for(elmId = seg->begin(); elmId != seg->end(); elmId++) {
      unsigned int elmIdx = elmProfile_->idx(*elmId);

      if (static_cast<unsigned int>(elmProfile_->locationZ(elmIdx)) <
          elmProfile_->wSize()-1) {
        unsigned int segIdx_last = last.idx(last.mask(*elmId));

        pxlArea = elmProfile_->nPixels(elmIdx);
        similarity[segIdx_last][segIdx] += pxlArea;
        totalPxlArea_last[segIdx_last] += pxlArea;
        totalPxlArea[segIdx] += pxlArea;
      }
    }
  }


  // calculate percentage of overlap and account for old age privilege
  double age;
  for (unsigned int seg_last=0; seg_last<last.elements_.size(); seg_last++) {
    for (unsigned int seg=0; seg<elements_.size(); seg++) {
//       std::cout << "last id " << last.id(seg_last) << " area " <<
// last.nPixels_[seg_last] << std::endl << std::flush;
//       std::cout << "overlap seg idx " << seg << " to last id " <<
// last.id(seg_last) << " - " << similarity[seg_last][seg] << std::endl <<
// std::flush;

      similarity[seg_last][seg] *= 2.0;
      similarity[seg_last][seg] /= (totalPxlArea_last[seg_last] +
                                    totalPxlArea[seg]);

//       std::cout << "similarity seg idx " << seg << " to last id " <<
// last.id(seg_last) << " - " << similarity[seg_last][seg] << std::endl <<
// std::flush;

      // segments are more likely to be assigned to old ones
      age = last.age(last.id(seg_last));
      age /= adultAge;
      if (age > 1.0) age = 1.0;
      age = (1.0-ageWeight)+(age*ageWeight);
      similarity[seg_last][seg] *= age;
    }
  }

  // using the similarity, guess which new segment
  // is which old segment and assign segment identifiers accordingly

  // find unused identifiers
  // TODO: make identifiers and color numbers separate and let id's run
  //       continously
//   MSK_ELM_TYPE maxSegId = (last.indices_.rbegin())->first;

//   std::vector<int> freeSegIds;
//   for (MSK_ELM_TYPE id=maxSegId-1; id>=0; id--) {
//     if (last.indices_.find(id) == last.indices_.end())
//       freeSegIds.push_back(id);
//   }

 
// assign each segment an identifier, either of a segment from
  // the last iteration or an unsued one
  this->identifiers_.resize(nSeg_);
  unsigned int similarSegment;
  std::vector<double> maxSim(nSeg_,0.0);
  for (unsigned int seg=0; seg<nSeg_; seg++) {
    // find the most similar segment from the last iteration
    double currentMaxSim = -1.0;
    for (unsigned int seg_last=0; seg_last<last.nSeg_; seg_last++) {
      if (similarity[seg_last][seg] > currentMaxSim) {
        currentMaxSim = similarity[seg_last][seg];
        similarSegment = seg_last;
      }
    }

//     std::cout << "this seg " << seg << " similar to " <<
// last.id(similarSegment) << " (" << currentMaxSim << ")" << std::endl <<
// std::flush;
//     std::cout << "this seg " << seg << " pixel area " << this->nPixels_[seg]
// << std::endl << std::flush;

    if (currentMaxSim > attachSim) {
    // this segment is similar to a segment from the previous segmentation
      MSK_ELM_TYPE id = last.id(similarSegment);
      typename std::map<MSK_ELM_TYPE, unsigned int>::iterator otherSeg_p =
          this->indices_.find(id);

      if (otherSeg_p == this->indices_.end()) {
//         std::cout << "this seg " << seg << " gets the id " << id << std::endl
// << std::flush;
      // the global segment number wasn't assigned yet
        this->identifiers_[seg] = id;
        this->indices_[id] = seg;
        maxSim[seg] = currentMaxSim;
        continue;
      }
      else {
      // this global Number was already assigned to another segment
        int otherSeg = otherSeg_p->second;
//         std::cout << "the id " << id << " was assigned to " << otherSeg <<
// std::endl << std::flush;
        if (currentMaxSim > maxSim[otherSeg]) {
//           std::cout << "this seg " << seg << " gets the id " << id <<
// std::endl << std::flush;
          // this segment has higher similarity and gets the global number
          this->identifiers_[seg] = id;
          this->indices_[id] = seg;
          maxSim[seg] = currentMaxSim;

          // give the other segment a new global segment number
          if (nextSegId_ > MAX_SEG_ID)
            throw ncutCritical("not enough segment Id numbers");  
          this->identifiers_[otherSeg] = nextSegId_;
          this->indices_[nextSegId_] = otherSeg;
          nextSegId_++;
          continue;
        }
//         std::cout << "the other seg " << otherSeg << " keeps the id " << id
// << std::endl << std::flush;
      }
    }

    // give this segment a new global segment number
    if (nextSegId_ > MAX_SEG_ID)
      throw ncutCritical("not enough segment Id numbers");

    this->identifiers_[seg] = nextSegId_;
    this->indices_[nextSegId_] = seg;
    nextSegId_++;

//       std::cout << "this seg " << seg << " gets the new id " <<
// this->identifiers_[seg] << std::endl << std::flush;
//     }

  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    attach(const SegmentProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>& last)
{
  unsigned int segDelay = this->setting_->segment_delay;
  unsigned int adultAge = this->setting_->segment_adult_age;

  if (segDelay < 0)
    throw ncutException("invalid segment delay, must be >= 0.");
  if (adultAge == 0)
    throw ncutException("invalid segment adult age, must be >= 1.");

  if (last.nSeg_ == 0) {

    // build the trivial identifiers, indices, mask, active and simSeg maps
    // (only used for the first iteration)
    typename std::vector<MSK_ELM_TYPE>::iterator elm;
    for (; nextSegId_<elements_.size(); nextSegId_++) {
      this->identifiers_.push_back(static_cast<MSK_ELM_TYPE>(nextSegId_));
      this->indices_[this->id(nextSegId_)] = nextSegId_;
      for (elm  = elements_[nextSegId_].begin();
           elm != elements_[nextSegId_].end(); elm++) {
             mask_[*elm] = this->id(nextSegId_);
      }
      age_[this->id(nextSegId_)] = (adultAge>segDelay) ? adultAge+1 : segDelay+1;
      inactive_[this->id(nextSegId_)] = false;
      simSeg_[this->id(nextSegId_)] = this->id(nextSegId_);
    }
  }
  else {

    // get identifiers and indices using info from the last iteration as
    // reference
    attachIds(last);

//     typename std::map<MSK_ELM_TYPE, unsigned int>::iterator segIt;
//     for (segIt = this->indices_.begin(); segIt != this->indices_.end(); segIt++) {
//       std::cout << "nach attach id " << segIt->first << " - idx " << segIt->second << std::endl << std::flush;
//     }

    // build the segment mask
    typename std::vector<MSK_ELM_TYPE>::iterator elm;
    for (unsigned int seg=0; seg<nSeg_; seg++) {
      for (elm  = elements_[seg].begin(); elm != elements_[seg].end(); elm++)
        mask_[*elm] = this->id(seg);
    }

    // build the active segments map
    typename std::map<MSK_ELM_TYPE, bool>::iterator inactive;
    for (inactive = inactive_.begin(); inactive!=inactive_.end(); inactive++)
      inactive->second = true;
    typename std::map<MSK_ELM_TYPE, unsigned int>::const_iterator last_active;
    for (unsigned int seg=0; seg < nSeg_; seg++) {
      MSK_ELM_TYPE segId = this->id(seg);
      last_active = last.age_.find(segId);
      if (last_active == last.age_.end()) {
        age_[segId] = 1;
        inactive_[segId] = false;
      }
      else {
        age_[segId] += last_active->second + 1;
        inactive_[segId] = false;
      }

    }

//     typename std::map<MSK_ELM_TYPE, unsigned int>::iterator activeSeg;
//     for (activeSeg = age_.begin(); activeSeg != age_.end(); activeSeg++) {
//       std::cout << "age_[" << activeSeg->first << "] " << activeSeg->second
//         << std::endl << std::flush;
//     }

//     typename std::map<MSK_ELM_TYPE, bool>::iterator inactiveSeg;
//     for (inactiveSeg = inactive_.begin();
//          inactiveSeg != inactive_.end(); inactiveSeg++) {
//       std::cout << "inactive_[" << inactiveSeg->first << "] "
//         << inactiveSeg->second << std::endl << std::flush;
//     }

    // build a list of segments that are similar to the segments active for
    // less than segDelay frames
    for (unsigned int seg=0; seg<nSeg_; seg++) {
      if (age_[this->id(seg)] < segDelay) {

        // this segment has been active for less than segDelay frames,
        // find the most similar segment
        double minDif = 600.0;
        int minDifSeg = -1;
        for (unsigned int seg2=0; seg2<nSeg_; seg2++) {
          if (seg == seg2) continue;
          if (age_[this->id(seg2)] < segDelay) continue;
          double difference = dif(seg, seg2);
          if (difference < minDif) {
            minDifSeg = seg2;
            minDif = difference;
          }
        }
        if (minDifSeg == -1) // no similar segment found
          simSeg_[this->id(seg)] = this->id(seg);
        else
          simSeg_[this->id(seg)] = this->id(minDifSeg);
      }
      else {
        simSeg_[this->id(seg)] = this->id(seg);
      }
    }

  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
int SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    push(const ElementProfile<IMG_ELM_TYPE,MSK_ELM_TYPE>* elmProfilee,
         const std::vector<std::vector<MSK_ELM_TYPE> >& segmentss)
{
  if (elmProfilee == NULL)
    throw ncutException("sp invalid elment profile (NULL)");
  if (elmProfilee->nElm() == 0)
    throw ncutException("invalid elment profile (does not contain elements)");
  if (segmentss.empty())
    throw ncutException("invalid segments list (empty)");

  // save info from the last iteration
  SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE> last(*this);

  // free memory
  cleanup();
  this->indices_.clear();
  this->identifiers_.clear();
  mask_.clear();
  idxMask_.clear();
  age_.clear();
  simSeg_.clear();

  // initialize members
  elmProfile_ = elmProfilee;
  nChl_ = elmProfilee->nChl();
  nSeg_ = segmentss.size();
  elements_ = segmentss;
  idxMask_.resize(elmProfilee->nElm());
  neighbors_.changeDimension(nSeg_);
  intensity_ = new double* [nSeg_];
  for (unsigned int seg=0; seg<nSeg_; seg++)
    intensity_[seg] = new double[nChl_];
  locationX_ = new double [nSeg_];
  locationY_ = new double [nSeg_];
  nElements_ = new unsigned int [nSeg_];
  nPixels_   = new unsigned int [nSeg_];

  // set the segment info arrays to zero
  for (unsigned int seg=0; seg<nSeg_; seg++) {
    for (unsigned int chl = 0; chl<nChl_; chl++) {
      intensity_[seg][chl] = 0.0;
    }
    locationX_[seg] = 0.0;
    locationY_[seg] = 0.0;
    nElements_[seg] = 0;
    nPixels_[seg]   = 0;
  }

  // build the element index mask and calculate mean intensities and location
  try {
    unsigned int elmIdx;
    typename std::vector<MSK_ELM_TYPE>::iterator elmId_p;
    for (unsigned int segIdx=0; segIdx<elements_.size(); segIdx++) {
      for (elmId_p = elements_[segIdx].begin();
           elmId_p != elements_[segIdx].end(); elmId_p++) {

        // get the element index (may throw ncutException)
        elmIdx = elmProfile_->idx(*elmId_p);

        // build the element index mask
        idxMask_[elmIdx] = segIdx;

        // add intensities
        for (unsigned int chl = 0; chl<nChl_; chl++)
          intensity_[segIdx][chl] +=
              elmProfile_->intensity(elmIdx,chl) *
              elmProfile_->nPixels(elmIdx);
        nElements_[segIdx]++;
        nPixels_[segIdx] += elmProfile_->nPixels(elmIdx);

        // add pixel locations
        locationX_[segIdx] += elmProfile_->locationX(elmIdx) *
                              elmProfile_->nPixels(elmIdx);
        locationY_[segIdx] += elmProfile_->locationY(elmIdx) *
                              elmProfile_->nPixels(elmIdx);
      }
    }
  }
  catch (ncutException& e) {
    *this = last;
    throw;
  }



  // calculate the segment mean intensity and center of gravity
  for (unsigned int seg=0; seg<nSeg_; seg++) {
    if (nPixels_[seg] == 0) {
      for (unsigned int chl = 0; chl<nChl_; chl++)
        intensity_[seg][chl] = 0.0;
      locationX_[seg] = -1;
      locationY_[seg] = -1;

      continue;
    }

    for (unsigned int chl=0; chl<nChl_; chl++) {
      intensity_[seg][chl] /= nPixels_[seg];
    }

    locationX_[seg] /= nPixels_[seg];
    locationY_[seg] /= nPixels_[seg];
  }

  // calculate the neighborhood information
  for (unsigned int i=0; i<elmProfile_->nElm(); i++) {
    for (unsigned int j=i+1; j<elmProfile_->nElm(); j++) {
      if (idxMask_[i] != idxMask_[j] &&
          elmProfile_->neighbors()->val(i,j) > 0.5) {
        neighbors_.val(idxMask_[i],idxMask_[j]) +=
            elmProfile_->neighbors()->val(i,j);
        neighbors_.val(idxMask_[j],idxMask_[i]) =
            neighbors_.val(idxMask_[i],idxMask_[j]);
      }
    }
  }

  try {
    attach(last);
  }
  catch (ncutException& e) {
    *this = last;
    throw;
  }

  return 0;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    SegmentProfile(const SegmentProfile& clone)
  : Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>(clone.setting_),
    elements_(),
    elmProfile_(NULL),
    mask_(),
    idxMask_(),
    age_(),
    inactive_(),
    simSeg_(),
    nSeg_(0),
    nChl_(0),
    nextSegId_(0),
    neighbors_(),
    intensity_(NULL),
    locationX_(NULL),
    locationY_(NULL),
    nElements_(NULL),
    nPixels_(NULL)
{
  *this = clone;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    SegmentProfile(const Setting* setting)
  : Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>(setting),
    elements_(),
    elmProfile_(NULL),
    mask_(),
    idxMask_(),
    age_(),
    inactive_(),
    simSeg_(),
    nSeg_(0),
    nChl_(0),
    nextSegId_(0),
    neighbors_(),
    intensity_(NULL),
    locationX_(NULL),
    locationY_(NULL),
    nElements_(NULL),
    nPixels_(NULL)
{
  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::cleanup()
{
  if (nElements_ != NULL) {
    delete [] nElements_;
    nElements_ = NULL;
  }
  if (nPixels_ != NULL) {
    delete [] nPixels_;
    nPixels_ = NULL;
  }
  if (intensity_ != NULL) {
    for (unsigned int seg=0; seg < nSeg_; seg++) {
      if (intensity_[seg] != NULL) delete [] intensity_[seg];
    }
    delete [] intensity_;
    intensity_ = NULL;
  }
  if (locationX_ != NULL) {
    delete [] locationX_;
    locationX_ = NULL;
  }
  if (locationY_ != NULL) {
    delete [] locationY_;
    locationY_ = NULL;
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::~SegmentProfile()
{
  cleanup();
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>&
SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    operator=(const SegmentProfile& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  cleanup();

  // copy parameters
  this->indices_     = clone.indices_;
  this->identifiers_ = clone.identifiers_;
  this->similarity_  = clone.similarity_;
  this->elements_    = clone.elements_;
  this->elmProfile_  = clone.elmProfile_;
  this->mask_        = clone.mask_;
  this->idxMask_     = clone.idxMask_;
  this->age_         = clone.age_;
  this->inactive_    = clone.inactive_;
  this->simSeg_      = clone.simSeg_;
  this->nChl_        = clone.nChl_;
  this->nSeg_        = clone.nSeg_;
  this->nextSegId_   = clone.nextSegId_;
  this->neighbors_   = clone.neighbors_;

  // copy segment info arrays
  if (clone.intensity_ != NULL) {
    this->intensity_ = new double* [clone.nSeg_];
    for (unsigned int seg=0; seg<clone.nSeg_; seg++) {
      if (clone.intensity_[seg] == NULL) {
        this->intensity_[seg] = NULL;
        continue;
      }
      else {
        this->intensity_[seg] = new double [clone.nChl_];
        for (unsigned int chl=0; chl<clone.nChl_; chl++) {
          this->intensity_[seg][chl] = clone.intensity_[seg][chl];
        }
      }
    }
  }
  else {
    this->intensity_ = NULL;
  }

  if (clone.locationX_ != NULL) {
    this->locationX_ = new double [clone.nSeg_];
    for (unsigned int seg=0; seg<clone.nSeg_; seg++) {
      this->locationX_[seg] = clone.locationX_[seg];
    }
  }
  else {
    this->locationX_ = NULL;
  }

  if (clone.locationY_ != NULL) {
    this->locationY_ = new double [clone.nSeg_];
    for (unsigned int seg=0; seg<clone.nSeg_; seg++) {
      this->locationY_[seg] = clone.locationY_[seg];
    }
  }
  else {
    this->locationY_ = NULL;
  }

  if (clone.nElements_ != NULL) {
    this->nElements_ = new unsigned int [clone.nSeg_];
    for (unsigned int seg=0; seg<clone.nSeg_; seg++) {
      this->nElements_[seg] = clone.nElements_[seg];
    }
  }
  else {
    this->nElements_ = NULL;
  }

  if (clone.nPixels_ != NULL) {
    this->nPixels_ = new unsigned int [clone.nSeg_];
    for (unsigned int seg=0; seg<clone.nSeg_; seg++) {
      this->nPixels_[seg] = clone.nPixels_[seg];
    }
  }
  else {
    this->nPixels_ = NULL;
  }

  // copy the setting
  this->setting_ = clone.setting_;

  return *this;
}

} // namespace ncut

#endif // MOTIONPROFILE_H
