#ifndef NCUT_CORE_H
#define NCUT_CORE_H 1

#include "motionProfile.h"

#define ARPACK_SILENT_MODE 1

#include <arbsmat.h>
#include <arbgsym.h>

// #include <time.h>

// #include <arbsmat.h>
// #include <arbgsym.h>
// #include <arerror.h>

// #include <assert.h>

namespace ncut
{

/**
 * @brief Ncut represents one ncut of a similarity matrix.
 *
 * This class is used by NcutNode to calculate
 * the ncuts and construct the tree accordingly.
 * Do not use @c Ncut directly, use the functions provided by the class
 * @c Segmentation.
 */
class Ncut
{
  public:

    // functions for read-only access to data members ------

    /**
     * @brief Get a pointer to the similarity matrix.
     * @return Pointer to the similarity matrix.
     */
    inline const Matrix* similarity() const
    {return similarity_;};
    /**
     * @brief Get the total association of element with index i.
     * @param i element index
     * @return Total association of element with index i.
     */
    inline double assoc(unsigned int i) const
    {return assoc_[i];};
    /**
     * @brief Get a pointer to the total association array.
     * @return Pointer to the total association array.
     */
    inline const double* assoc() const
    {return assoc_;};
    /**
     * @brief Get the dimension of this ncut.
     * @return Dimension of this ncut.
     */
    inline unsigned int dim() const
    {return dim_;};
    /**
     * @brief Get the eigenvalue of the eigenvector with index i.
     * @param i eigenvector index
     * @return Eigenvalue of the eigenvector with index i.
     */
    inline double eigVal(unsigned int i) const
    {return eigVal_[i];};
    /**
     * @brief Get a pointer to the eigenvalue array.
     * @return Pointer to the eigenvalue array.
     */
    inline const double* eigVal() const
    {return eigVal_;};
    /**
     * @brief Get the eigenvector component with index i.
     * @param i component index, the components of all eigenvectors are in this array.
     * @return Eigenvector component with index i.
     */
    inline double eigVec(unsigned int i) const
    {return eigVec_[i];};
    /**
     * @brief Get a pointer to the eigenvectors array.
     * @return Pointer to the eigenvectors array.
     */
    inline const double* eigVec() const
    {return eigVec_;};
    /**
     * @brief Get the number of converged eigenvectors.
     * @return Number of converged eigenvectors.
     */
    inline unsigned int nConv() const
    {return nConv_;};

    // class-specific functions ------

    /**
     * @brief Calculate one minimal ncut.
     *
     * @param nEigVecs The desired number of eigenvectors.
     * @param initVec An initialization vector for arpack, defaults to @c NULL.
     * @return 0 if successful.
     *
     * Calculates the minimal ncut of the similarity matrix pointed to by the
     * member similarity and tries to produce @c nEigVecs eigenvectors and
     * eigenvalues, but @c nEigVecs eigenvectors are @e not guaranteed. Use
     * nConv to obtain the actual number of calculated eigenvectors.
     */
    int calculate(unsigned int nEigVecs, double* initVec = NULL);

    // standard functions ------

    /**
     * @brief %Ncut can be constructed from a similarity matrix.
     *
     * @param similarity A pointer to the similarity matrix on which
     *                   the ncut will be calculated.
     */
    Ncut(const Matrix* similarity);

    /**
     * @brief %Ncut copy constructor.
     *
     * @param clone The ncut to be cloned.
     *
     * Creates a completely independent copy of @c clone.
     */
    Ncut(const Ncut& clone);

    /**
     * @brief %Ncut destructor.
     */
    virtual ~Ncut();

    /**
     * @brief %Ncut assignment operator.
     *
     * @param clone An @c Ncut to be cloned.
     * @return This object.
     */
    virtual Ncut& operator=(const Ncut& clone);

  protected:

    const Matrix* similarity_; /**< the similarity matrix on which
                                    the ncut is calculated*/
    unsigned int dim_; /**< the dimension of the ncut
                            (= similarity matrix dimension)*/

    double* assoc_; /**< an array containing the total association
                         of each node*/
    double* eigVal_; /**< an array of the calculated eigenvalues
                          (its length is @c nConv_)*/
    double* eigVec_; /**< an array of the calculated eigenvectors
                          (its length is @c nConv_ * @c dim_)*/
    unsigned int nConv_; /**< the actual number of converged eigenvalues
                              and -vectors*/

  private:

    /**
     * @brief Frees all memory dynamically allocated directly by this object.
     */
    void cleanup();
};

/**
 * @brief NcutNode is one Node in the ncut dendogram.
 *
 * A node represents one bipartition of a segment into two halves.
 * The root @c NcutNode bipartitions (splits) the whole image.
 * @c leftNode and @c rightNode point to the @c NcutNode s (=partitions) of
 * the two resulting segments or to @c NULL if the resulting segments aren't
 * partitioned any further. An eigenvector, calculated by the class @c Ncut, is
 * used as an indicator vector to bipartition a segment. The node does not
 * neccessarily contain a new ncut, an eigenvector of an ncut calculated in a
 * level closer to the root may also be used. If this is the case, @c ncut_
 * points to the old ncut, but @c eigVec_ points to the eigenvector that is
 * used to partition this segment.
 * Segmentation is not performed on pixels, but on elements, for example
 * the segments of a pre-segmentation. Elements have global identifiers and
 * local indices. @c indices_ is a map from element identifiers to element
 * indices in each @c NcutNode. This map is also a list of elements that are
 * part of the current segment.
 * Do not use @c NcutNode directly, construct the ncut dendogram using the
 * functions provided by the class @c Segmentation.
 */
template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
class NcutNode
{
  public:

    // functions for read-only access to data members ------

    /**
     * @brief Get the Profile this Node uses.
     * @return Profile this Node uses.
     */
    inline const greedy_ptr<const Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >
    profile() const
    {return profile_;};
    /**
     * @brief Get the index map of the elements in this NcutNode.
     * @return Index map of the elements in this NcutNode.
     */
    inline const std::map<MSK_ELM_TYPE, unsigned int>* indices() const
    {return &indices_;};
    /**
     * @brief Get the Ncut calculated by this NcutNode.
     * @return Ncut calculated by this NcutNode.
     */
    inline const greedy_ptr<const Ncut> ncut() const
    {return ncut_;};
    /**
     * @brief Get the eigenvalue of the eigenvector this NcutNode uses for partitioning.
     * @return Eigenvalue of the eigenvector this NcutNode uses for partitioning.
     */
    inline double eigVal() const
    {return eigVal_;};
    /**
     * @brief Get a pointer to the eigenvector this NcutNode uses for partitioning.
     * @return Pointer to the eigenvector this NcutNode uses for partitioning.
     */
    inline const double* eigVec() const
    {return eigVec_;};
    /**
     * @brief Get the splitting point this NcutNode uses for partitioning.
     * @return Splitting point this NcutNode uses for partitioning.
     */
    inline double splitPoint() const
    {return splitPoint_;};
    /**
     * @brief Get the number of the eigenvector that was used for partitioning.
     * @return Number of the eigenvector that was used for partitioning.
     */
    inline double eigVecNum() const
    {return eigVecNum_;};
    /**
     * @brief Get the left child NcutNode of this NcutNode.
     * @return Left child NcutNode of this NcutNode.
     */ 
    inline const greedy_ptr<const NcutNode> leftNode() const
    {return leftNode_;};
    /**
     * @brief Get the right child NcutNode of this NcutNode.
     * @return Right child NcutNode of this NcutNode.
     */ 
    inline const greedy_ptr<const NcutNode> rightNode() const
    {return rightNode_;};

    // class-specific functions ------

    /**
     * @brief Constructs the ncut dendogram.
     *
     * @param maxDepth The maximum depth of the ncut dendogram.
     * @param depth The current depth this node is in.
     * @param nEigVecs The number of eigenvectors that should be used of each
     *                 ncut.
     * @param eigvn The number of the current eigenvector used to
     *              bipartition this segment.
     * @param initVec The starting vector used for calculating the ncut.
     * @return 0 if a valid Node was constructed,
     *         1 if the node is invalid and has to be deleted.
     *
     * Bipartitions the elements contained in @c indices_ (either using an
     * existing @c Ncut or calculating a new one).
     * Child nodes are recursively constructed until the maximum depth is
     * reached. If one of the child nodes needs a new @c Ncut, a new element
     * profile with reduced similarity matrix (only containing the
     * elements in the respective segment) is constructed and passed to
     * the child node.
     */
    int calculate(unsigned int maxDepth, unsigned int depth,
                  unsigned int nEigVecs, unsigned int eigvn,
                  double* initVec = NULL);

    /**
     * @brief Checks to which half an element is assigned in this node.
     *
     * @param idx The index of the element to be checked.
     * @return @c true if the element is on the left half, @c false otherwise.
     *
     * Checks if the element with index @c idx belongs to the left half of this
     * node's bipartition.
     */
    bool isLeft(unsigned int idx) const;

    // standard functions ------

    /**
     * @brief %NcutNode owning constructor, takes ownership of @c profile.
     *
     * @param profile the profile used for segmentation
     * @param indices a list of elements still active in this node and their
     *                indices
     * @param ncut the ncut from the parent node
     * @param setting The parameters of the current segmentation.
     *
     * Creates a new NcutNode with the Profile @c profile, element indices list
     * @c indices and optionally with Ncut @c ncut.
     * @c profile @e is owned by this NcutNode.
     */
    NcutNode(std::auto_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& profile,
             const std::map<MSK_ELM_TYPE, unsigned int>& indices,
             greedy_ptr<Ncut>& ncut, const Setting* setting);

    /**
     * @brief %NcutNode non-owning constructor, does not own @c profile.
     *
     * @param profile the profile used for segmentation
     * @param indices a list of elements still active in this node and their
                      indices
     * @param ncut the ncut this node should use 
     * @param setting The parameters of the current segmentation.
     *
     * Creates a new NcutNode with the Profile @c profile, element indices list
     * @c indices and optionally with Ncut @c ncut.
     * @c profile is @e not owned by this NcutNode.
     */
    NcutNode(greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& profile,
             const std::map<MSK_ELM_TYPE, unsigned int>& indices,
             greedy_ptr<Ncut>& ncut, const Setting* setting);

    /**
     * @brief %NcutNode copy constructor.
     *
     * @param clone The root of the subtree to be cloned.
     *
     * Clones the NcutNode @c clone and all its child nodes (e.g. clones the
     * subtree @c clone is root of).
     */
    NcutNode(const NcutNode& clone);

    /**
     * @brief %NcutNode custom copy constructor.
     *
     * @param clone The root of the subtree to be cloned.
     * @param pclone The Profile this subtree should point to.
     * @param nclone The Ncut this subtree should point to.
     *
     * This copy constructor uses a pointer to a different @c Profile and
     * a pointer to a different @c Ncut.
     * These pointers are handed down the tree to make sure every child node of
     * this node will be independent of @c clone.
     */
    NcutNode(const NcutNode& clone,
             greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& pclone,
             greedy_ptr<Ncut>& nclone);

    /**
     * @brief %NcutNode destructor.
     *
     * Deletes the subtree this @c NcutNode is root of.
     */
    virtual ~NcutNode();

    /**
     * @brief %NcutNode assignment operator.
     *
     * @param clone The root of the subtree to be cloned.
     * @return This object.
     *
     * Clones the NcutNode @c clone and all its child nodes (e.g. clones the
     * subtree @c clone is root of).
     */
    virtual NcutNode& operator=(const NcutNode& clone);

  protected:

    // element info
    greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> > profile_; /**<
        the profile used in this Node*/
    std::map<MSK_ELM_TYPE, unsigned int> indices_; /**<
        the list of elements bipartitioned in this node and their indices*/


    // ncut info
    greedy_ptr<Ncut> ncut_; /**< the @c Ncut that calculated the current
                                 eigenvector*/
    double eigVal_; /**< the eigenvalue corresponding to the eigenvector used
                         to bipartition this segment*/
    const double* eigVec_; /**< the eigenvector used to bipartition this
                                segment*/
    double splitPoint_; /**< the threshold for bipartitioning the graph*/
    unsigned int eigVecNum_; /**< the number of the eigenvector that
                                  was used for bipartitioning*/

    // child nodes
    greedy_ptr<NcutNode> leftNode_; /**< the left child node represents the
                                         further bipartition of one of the two
                                         segments resulting from this
                                         bipartition*/
    greedy_ptr<NcutNode> rightNode_; /**< the left child node represents the
                                          further bipartition of the other of
                                          the two segments resulting from
                                          this bipartition*/

    const Setting* setting_; /**< Contains all needed parameters*/

  private:

    /**
     * @brief %NcutNode custom copy function.
     *
     * @param clone The @c NcutNode to be cloned.
     * @param pclone A pointer to a different @c Profile.
     * @param nclone A pointer to a different @c Ncut.
     * @return This object.
     *
     * This function is needed to clone a complete subtree starting from a root
     * @c NcutNode. For the subtree to be completely independent, there
     * must be no pointer pointing to a @c Profile or an @c Ncut of the original
     * subtree. To achieve this, pointers to @c Profile and @c Ncut that were
     * constructed in the cloned subtree have to be handed down the cloned
     * subtree while it is being cloned.
     */
    NcutNode&
    copy(const NcutNode& clone,
         greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& pclone,
         greedy_ptr<Ncut>& nclone);
};

/**
 * @brief Segmentation represents the complete segmentation of the elements in an
 * image sequence.
 *
 * The datastructure used for segmentation is a dendogram
 * consisting of @c NcutNodes . @c Segmentation contains a pointer to the root
 * @c NcutNode. Each @c NcutNode represents a bipartition of a part of the image
 * into two halves. Each @c NcutNode may have two child NcutNodes representing
 * the further bipartition of each of the two halves. The dendogram has a
 * maximum depth and a desired number of eigenvectors per ncut.
 * Use @c push to push furher images and correspondings masks for segmentation.
 */
template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
class Segmentation
{
  public:

    // functions for read & write access to data members ------

    /**
     * @brief Get a pointer to the settings (parameters) used by this Segmentation.
     * @return Pointer to the settings (parameters) used by this Segmentation.
     * This function is for read&write access.
     */
    inline Setting* setting()
    {return &setting_;};

    // functions for read-only access to data members ------

    /**
     * @brief Get a pointer to one frame of the frame sequence.
     * @param num The sequence number of the frame.
     * @return Pointer to one frame of the frame sequence.
     */
    inline const Image<IMG_ELM_TYPE>* frame(unsigned int num) const
    {return elmProfile_.frame(num);};
    /**
     * @brief Get a pointer to one frame of the mask sequence.
     * @param num The sequence number of the frame.
     * @return Pointer to one frame of the mask sequence.
     */
    inline const Image<MSK_ELM_TYPE>* elmMask(unsigned int num) const
    {return elmProfile_.mask(num);};
    /**
     * @brief Get the value of one pixel of a frame in the mask sequence.
     * @param num The sequence number of the frame.
     * @param pxl The number of the pixel (starting from top left).
     * @return Value of one pixel of a frame in the mask sequence.
     */
    inline const MSK_ELM_TYPE& elmMask(unsigned int num, unsigned int pxl) const
    {return elmProfile_.mask(num, pxl);};
    /**
     * @brief Get the value of one pixel of a frame in the frame sequence.
     * @param num The sequence number of the frame.
     * @param pxl The number of the pixel (starting from top left).
     * @return Value of one pixel of a frame in the frame sequence.
     */
    inline const IMG_ELM_TYPE& frameMask(unsigned int num,
                                         unsigned int pxl) const
    {return elmProfile_.frame(num, pxl);};
    /**
     * @brief Get the width of the video sequences.
     * @return Width of the video sequences.
     */
    inline unsigned int width() const
    {return elmProfile_.width();};
    /**
     * @brief Get the height of the video sequences.
     * @return Height of the video sequences.
     */
    inline unsigned int height() const
    {return elmProfile_.height();};
    /**
     * @brief Get the number of channels of the video sequence.
     * @return Number of channels of the video sequence.
     */
    inline unsigned int nChl() const
    {return elmProfile_.nChl();};	
    /**
     * @brief Get the number of elements in the current frame window.
     * @return Number of elements in the current frame window.
     */
    inline unsigned int nElm() const
    {return elmProfile_.nElm();};

    /**
     * @brief Get the x-midpoint position of a segment in the current frame window.
     * @param seg The index of the segment in the current frame window.
     * @return X-midpoint position of a segment in the current frame window.
     */
    inline double locationX(unsigned int seg) const
    {return segProfile_.locationX(seg);};
    /**
     * @brief Get the y-midpoint position of a segment in the current frame window.
     * @param seg The index of the segment in the current frame window.
     * @return Y-midpoint position of a segment in the current frame window.
     */
    inline double locationY(unsigned int seg) const
    {return segProfile_.locationY(seg);};

    /**
     * @brief Get the number of segments in the current frame window.
     * @return Number of segments in the current frame window.
     */
    inline unsigned int nSeg() const
    {return segProfile_.nSeg();};
    /**
     * @brief Get the number of segmented frames since the creation of this Segmentation.
     * @return Number of segmented frames since the creation of this Segmentation.
     */
    inline unsigned int nFrm() const
    {return elmProfile_.nFrm();};
    /**
     * @brief Get the size of the frame window in frames.
     * @return Size of the frame window in frames.
     */
    inline unsigned int wSize() const
    {return elmProfile_.wSize();};

    /**
     * @brief Get the index map of the segments in the current frame window.
     * @return Index map of the segments in the current frame window.
     */
    inline const std::map<MSK_ELM_TYPE, unsigned int>* indices() const
    {return segProfile_.indices();}

    /**
     * @brief Get the index of a segment in the current frame window.
     * @param segId Unique identifier of the segment.
     * @return Index of a segment in the current frame window.
     */
    inline unsigned int idx(const MSK_ELM_TYPE& segId) const
    {return segProfile_.idx(segId);};
    /**
     * @brief Get the unique identifier of a segment in the current frame window.
     * @param segIdx Index of the segment in the current frame window.
     * @return Unique identifier of a segment in the current frame window.
     */
    inline const MSK_ELM_TYPE& id(unsigned int segIdx) const
    {return segProfile_.id(segIdx);};
    /**
     * @brief Get the segment an element is part of.
     * @param elmId Identifier of the element.
     * @return Identifier of the segment the element is part of.
     */
    inline const MSK_ELM_TYPE& segMask(const MSK_ELM_TYPE& elmId) const
    {return segProfile_.mask(elmId);};
    /**
     * @brief Get a list of pixels in an element.
     * @param elmIdx Index of the element.
     * @return List of pixels in an element.
     */
    inline const std::vector<unsigned int>* pixels(unsigned int elmIdx) const
    {return elmProfile_.pixels(elmIdx);};
    /**
     * @brief Get the number of pixels in a segment.
     * @param seg Index of the segment.
     * @return Number of pixels in a segment.
     */
    inline double nPixels(int seg) const
    {return segProfile_.nPixels(seg);};
    /**
     * @brief Get a list of elements in a segment.
     * @param segIdx Index of the segment.
     * @return List of the indices of the elements in the segment.
     */
    inline const std::vector<unsigned int>* elements(unsigned int segIdx) const
    {return segProfile_.elements(segIdx);};
    /**
     * @brief Get the frame number in the current frame window an element is in.
     * @param elmIdx Index of the element.
     * @return Number of the frame in the current frame window (not the sequence number of the frame).
     */
    inline unsigned int frameOf(unsigned int elmIdx) const
    {return elmProfile_.frameOf(elmIdx);};
    /**
     * @brief Get the age of a segment.
     * @param segId Identifier of the segment.
     * @return Age of a segment.
     */
    inline unsigned int age(const MSK_ELM_TYPE& segId) const
    {return segProfile_.age(segId);};
    /**
     * @brief Get a pointer to the age map.
     * @return Pointer to the age map.
     */
    inline const std::map<MSK_ELM_TYPE, unsigned int>* age() const
    {return segProfile_.age();};
    // inactive segments
    /**
     * @brief Check if a segment is active (present) in the current frame window.
     * @param segId Identifier of the segment.
     * @return true if inactive, false otherwise.
     */
    inline const bool inactive(const MSK_ELM_TYPE& segId) const
    {return segProfile_.inactive(segId);};
    /**
     * @brief Get a pointer to the inactive map.
     * @return Pointer to the inactive map.
     */
    inline const std::map<MSK_ELM_TYPE, bool>* inactive() const
    {return segProfile_.inactive();};
    /**
     * @brief Get the most similar segment to a given segment.
     * @param segId Identifier of the segment.
     * @return The most similar segment to the given segment.
     */
    inline const MSK_ELM_TYPE& simSeg(const MSK_ELM_TYPE& segId) const
    {return segProfile_.simSeg(segId);};
    /**
     * @brief Get the root NcutNode of the dendogram of the current frame window.
     * @return Root NcutNode of the dendogram of the current frame window.
     */
    inline const NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>* dendogram() const
    {return dendogram_;};
    /**
     * @brief Get the ElementProfile of the current frame window.
     * @return ElementProfile of the current frame window.
     */
    inline const ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>* elmProfile() const
    {return &elmProfile_;};
    /**
     * @brief Get the SegmentProfile of the current frame window.
     * @return SegmentProfile of the current frame window.
     */
    inline const SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>* segProfile() const
    {return segProfile_;};
    /**
     * @brief Get a pointer to the settings (parameters) used by this Segmentation.
     * @return Pointer to the settings (parameters) used by this Segmentation.
      * This function is for read access only.
     */
    inline const Setting* setting() const
    {return &setting_;};

    /**
     * @brief Push frame and mask onto the sequence for segmentation.
     *
     * @param frame The frame image that should be pushed to
     *              the frame sequence for segmentation.
     * @param mask The pre-segmentation mask for the frame image.
     * @return 1 if there aren't enough frames in the sequence to start
     *         segmentation yet, 0 otherwise.
     *
     * Pushes frame and mask to the back of the frame and mask sequences,
     * calculates new profiles, constructs a new ncut dendogram and updates
     * segment information.
     */
    int push(const Image<IMG_ELM_TYPE>* frame,
             const Image<MSK_ELM_TYPE>* mask);

    /**
     * @brief %Segmentation constructor.
     *
     * @param wSize The size of the time window in frames.
     * @param setting The parameters of the current segmentation.
     *
     * Creates a new Segmentation that moves a frame window of @c wSize frames
     * over the whole sequence and calculates ncuts for the window in
     * every step.
     */
    Segmentation(unsigned int wSize = 3, Setting setting = Setting());

    /**
     * @brief %Segmentation copy constructor.
     *
     * @param clone The Segmentation to be cloned.
     */
    Segmentation(const Segmentation& clone);

    /**
     * @brief %Segmentation destructor.
     */
    virtual ~Segmentation();

    /**
     * @brief %Segmentation assignment operator.
     *
     * @param clone The Segmentation to be cloned.
     * @return This object.
     */
    virtual Segmentation& operator=(const Segmentation& clone);

  protected:

    Setting setting_; /**< Contains all needed parameters*/

    NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>* dendogram_; /**<
        the root node of the ncut dendogram*/

    ElementProfile<IMG_ELM_TYPE, MSK_ELM_TYPE> elmProfile_; /**<
        holds the element information and the similarity matrix*/
    SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE> segProfile_; /**<
        holds the segment information*/

  private:

    /**
     * @brief Gathers segment information from the subtree starting at @c node.
     *
     * @param node Segment information will be fetched from this node on.
     * @param elements The segment information is saved into this vector.
     *
     * The dendogram is traversed depth-first and information is
     * gathered from every leaf node.
     */
    void getInfo(const NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>* node,
                 std::vector<std::vector<MSK_ELM_TYPE> >& elements);


    /**
     * @brief Merges similar segments.
     *
     * @param profile The @c SegmentProfile containing information about the
     *                segments to be merged.
     * @return The updated @c SegmentProfile.
     *
     * Merges the segments of a @c SegmentProfile until there are no more
     * segments with a difference smaller than a given threshold, or until a
     * minimum segment count is reached.
     */
    std::vector<std::vector<MSK_ELM_TYPE> >
    merge(const SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>& profile);

    /**
     * @brief Finds connected elements starting from a given element.
     *
     * @param segProfile the profile of the segments that we be processed
     * @param startElmIdx The search will start from this element.
     * @param reached True if an element can be reached from @c startElmIdx,
     *                false otherwise.
     *
     * Uses depth-first search to find the connected component containing the
     * element with index startElmIdx. Elements are only connected if
     * they are neighbors in the same segment. Note that entries in
     * reached will not be modified if not reachable from startElmIdx.
     */
    void connected(const SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>& segProfile,
                   unsigned int startElmIdx, std::vector<bool>& reached);


    /**
     * @brief Splits segments consisting of multiple unconnected components.
     *
     * @param profile The @c SegmentProfile containing information about the
     *                segments to be split.
     * @return The updated @c SegmentProfile.
     *
     * Searches for segments in @c profile , that consist of two or more
     * components that aren't touching and splits those segments into their
     * components.
     */
    std::vector<std::vector<MSK_ELM_TYPE> >
    split(const SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>& profile);

    /**
     * @brief Frees all memory dynamically allocated directly by this object.
     */
    void cleanup();
};


/* template implementations ------------------------------------------------- */

int Ncut::calculate(unsigned int nEigVecs, double* initVec)
{
  nEigVecs = 4;
  if (similarity_ == NULL)
    throw ncutException("invalid similarity matrix (NULL)");
  if (dim_ == 0 || similarity_->dim() != dim_)
    throw ncutException("invalid dimension, must be > 0 and equal to "
                        "similarity matrix dimension");

  // free old data
  cleanup();

  // calculate the total association for each element
  Matrix assocMatrix(dim_);
  for (unsigned int i=0; i<dim_; i++)
    for (unsigned int j=0; j<dim_; j++)
      assocMatrix.val(i,i) += similarity_->val(i,j);

  // save the association values in an array
  assoc_ = new double [dim_];
  for (unsigned int i=0; i<dim_; i++)
    assoc_[i] = assocMatrix.val(i,i);

  // calculate the second matrix for the generalized eigenvalue problem
  Matrix assocSubSimil(assocMatrix);
  assocSubSimil -= *similarity_;

  // convert matrices to band format for use with arpack
  int assocSubSimilN;
  int assocSubSimilNsdiag;
  auto_array<double> assocSubSimilBand(assocSubSimil.band(assocSubSimilN,
                                                          assocSubSimilNsdiag));
  int assocN;
  int assocNsdiag;
  auto_array<double> assocBand(assocMatrix.band(assocN, assocNsdiag));

  try {
    // construct arpack matrix objects
    ARbdSymMatrix<double> assocSubSimilARP(assocSubSimilN, assocSubSimilNsdiag,
                                           assocSubSimilBand.get());
    ARbdSymMatrix<double> assocARP(assocN, assocNsdiag, assocBand.get());

    // calculate eigenvalues- and vectors of the motion profile using arpack
    int ncv = ((2*nEigVecs)+1<dim_-1) ? (2*nEigVecs)+1 : dim_;
//     int ncv = ((2*3)+1<dim_-1) ? (2*3)+1 : dim_;
    ARluSymGenEig<double> problem(nEigVecs, assocSubSimilARP, assocARP, "SA",
                                  ncv, 0.0, nEigVecs*100000, initVec);

for (unsigned int i=0; i<dim_; i++)
{
  for (unsigned int j=0; j<dim_; j++)
  {
    std::cout << "," << assocSubSimil.val(i,j) << " ";
  }
  std::cout << std::endl << std::flush;
}

    // get eigenvalues and eigenvectors
    eigVal_ = new double[nEigVecs];
    eigVec_ = new double[nEigVecs*dim_];
    nConv_ = problem.EigenValVectors(eigVec_, eigVal_);
//     for (int i=0; i<nEigVecs; i++) {
//       std::cout << "EigVec[" << i << "] - (" << std::flush;
//       for (int j=0; j<dim_; j++) {
//         std::cout  << eigVec_[dim_*i+j] << "," << std::flush;
//       }
//       std::cout << ")" << std::endl << std::flush;
//     }
  }
  catch (ArpackError& e) {
    throw ncutException(
        (std::string("ArpackError: error code ") +
         std::string(1,e.Status()) ).c_str() );
  }

  return 0;
}

Ncut::Ncut(const Matrix* similarity)
  : similarity_(similarity),
    dim_( (similarity==NULL) ? 0 : similarity->dim() ),
    assoc_(NULL), eigVal_(NULL), eigVec_(NULL), nConv_(0)
{
  if (similarity_ == NULL)
    throw ncutException("invalid similarity matrix (NULL)");
  return;
}

Ncut::Ncut(const Ncut& clone)
  : similarity_(NULL), dim_(0),
    assoc_(NULL), eigVal_(NULL), eigVec_(NULL), nConv_(0)
{
  *this = clone;
}

void Ncut::cleanup()
{
  if (assoc_ != NULL) {
    delete [] assoc_;
    assoc_ = NULL;
  }
  if (eigVal_ != NULL) {
    delete [] eigVal_;
    eigVal_ = NULL;
  }
  if (eigVec_ != NULL) {
    delete [] eigVec_;
    eigVec_ = NULL;
  }
}

Ncut::~Ncut()
{
  cleanup();
}

Ncut& Ncut::operator=(const Ncut& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  cleanup();

  // copy ncut info from the clone
  this->similarity_ = clone.similarity_;
  this->dim_        = clone.dim_;
  this->nConv_      = clone.nConv_;

  // initialize the local structures
  assoc_  = new double[clone.dim_];
  eigVal_ = new double[clone.nConv_];
  eigVec_ = new double[clone.nConv_*clone.dim_];

  // copy the total association values
  for (unsigned int i=0; i<clone.dim_; i++) {
    this->assoc_[i] = clone.assoc_[i];
  }

  // copy Eigenvalues and Eigenvectors from the clone
  for (unsigned i=0; i<clone.nConv_; i++) {
    this->eigVal_[i] = clone.eigVal_[i];

    for (unsigned j=0; j<clone.dim_; j++) {
      this->eigVec_[i*j] = clone.eigVec_[i*j];
    }
  }

  return *this;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
bool NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::isLeft(unsigned int idx) const
{
  if (eigVec_ == NULL || ncut_.get() == NULL || idx >= ncut_->dim() || idx < 0)
    throw ncutException("invalid element index");
  if (eigVec_[idx] > splitPoint_)
    return true;
  else
    return false;
}


template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
int NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    calculate(unsigned int maxDepth, unsigned int depth,
              unsigned int nEigVecs, unsigned int eigvn,
              double* initVec)
{
#ifdef debugout
  std::cout << "::calculate depth " << depth << " eigvn " << eigvn
  << " indices_.size() " << indices_.size() << std::endl << std::flush;
#endif

  if (profile_.get() == NULL)
    throw ncutException("invalid profile (NULL)");
  if (depth >= maxDepth ||
      eigvn >= nEigVecs ||
      eigvn < 1)
    throw ncutException("invalid arguments");

  unsigned int nStep = setting_->number_of_eigenvector_splittest_steps;
  if (nStep < 1)
    throw ncutException("number of split-test steps invalid, must be >= 1");

  // calculate data for this node ---------------

  // save the eigenvector number of this node
  eigVecNum_ = eigvn;

  // calculate new ncut if needed
  if (eigvn == 1 || ncut_.get() == NULL) {
    ncut_ =  new Ncut(profile_->similarity());
    try {
      ncut_->calculate(nEigVecs, initVec);
    }
    catch (ncutException& e) {
      #ifdef debugout
      std::cerr << e.what() << std::endl << std::flush;
      #endif
      return 1;
    }
  }

  // not enough converged eigenvectors, this node will be deleted
  if (ncut_->nConv() <= eigvn) return 1;

  // get eigenvalue and eigenvector
  eigVal_ = ncut_->eigVal(eigvn);
  eigVec_ = &( (ncut_->eigVec())[eigvn*ncut_->dim()] );

  // find a good splitting point
  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator elm;
  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator elm2;
  double minVal, maxVal;
  minVal = maxVal = eigVec_[0];
  for (unsigned int i=0; i<ncut_->dim(); i++) {
    if (eigVec_[i] < minVal) minVal = eigVec_[i];
    if (eigVec_[i] > maxVal) maxVal = eigVec_[i];
  }
  double step = (maxVal - minVal) / (double) nStep;
  if (step <= 0.0) return 1;

  double bestNcut = 2.0; // maximum (worst) possible ncut value
  double assoc_A, assoc_B, cut, ncutValue;
  for (double split = minVal+(step/2.0); split<maxVal; split+=step) {
    assoc_A = assoc_B = cut = 0.0;
    for (elm=indices_.begin(); elm!=indices_.end(); elm++) {
      if (eigVec_[elm->second] < split)
        assoc_A += ncut_->assoc(elm->second);
      else
        assoc_B += ncut_->assoc(elm->second);

      for (elm2=elm, elm2++; elm2!=indices_.end(); elm2++) {
        if ((eigVec_[elm->second]  < split) !=
            (eigVec_[elm2->second] < split)) {
          cut += profile_->similarity()->val(elm->second,elm2->second);
        }
      }
    }
    ncutValue = (cut/assoc_A) + (cut/assoc_B);
    if (ncutValue < bestNcut) {
      splitPoint_ = split;
      bestNcut = ncutValue;
    }
  }
#ifdef debugout
  std::cout << "splitPoint - " << splitPoint_ << std::endl << std::flush;
#endif
  
  // find a good splitting point v2 (to be tested)
//   typename std::map<MSK_ELM_TYPE, unsigned int>::iterator elm;
//   typename std::map<MSK_ELM_TYPE, unsigned int>::iterator elm2;
//   double minVal, maxVal;
//   minVal = maxVal = eigVec_[0];
//   for (unsigned int i=0; i<ncut_->dim(); i++) {
//     if (eigVec_[i] < minVal) minVal = eigVec_[i];
//     if (eigVec_[i] > maxVal) maxVal = eigVec_[i];
//   }
//   double step = (maxVal - minVal) / (double) (nStep+1);
//   if (step <= 0.0) return 1;
// 
//   double bestNcut = 2.0; // maximum (worst) possible ncut value
//   double assoc_A, assoc_B, cut, ncutValue, last_split;
//   assoc_A = assoc_B = cut = 0.0;
//   last_split = minVal - (step/2.0);
// 
//   for (double split = minVal+(step/2.0); split<maxVal; split+=step) {
//     if (last_split < minVal) {
//     // first ncut value calculation goes through all elements
//       for (elm=indices_.begin(); elm!=indices_.end(); elm++) {
//         if (eigVec_[elm->second] < split)
//           assoc_A += ncut_->assoc(elm->second);
//         else
//           assoc_B += ncut_->assoc(elm->second);
// 
//         for (elm2=elm, elm2++; elm2!=indices_.end(); elm2++) {
//           if ((eigVec_[elm->second]  < split) !=
//               (eigVec_[elm2->second] < split)) {
//             cut += profile_->similarity()->val(elm->second,elm2->second);
//           }
//         }
//       }
//     }
//     else {
//     // following ncut value calculations only consider elements that changed
//     // sides
//       for (elm=indices_.begin(); elm!=indices_.end(); elm++) {
//         if (eigVec_[elm->second]<split && eigVec_[elm->second]>last_split) {
//         // elm changed sides from B to A
//           assoc_A += ncut_->assoc(elm->second);
//           assoc_B -= ncut_->assoc(elm->second);
// 
//           for (elm2=elm, elm2++; elm2!=indices_.end(); elm2++) {
//             if (eigVec_[elm2->second] > split)
//             // elm2 is still on the B side
//               cut += profile_->similarity()->val(elm->second,elm2->second);
//             else
//             // both elements are on the A side
//               cut -= profile_->similarity()->val(elm->second,elm2->second);
//           }
//         }
//         else {
//           for (elm2=elm, elm2++; elm2!=indices_.end(); elm2++) {
//             if (eigVec_[elm2->second] < split &&
//                 eigVec_[elm2->second] > last_split) {
//             // elm2 changed sides from B to A
//               if (eigVec_[elm->second] > split)
//               // elm is still on the B side
//                 cut += profile_->similarity()->val(elm->second,elm2->second);
//               else
//               // both elements are on the A side
//                 cut -= profile_->similarity()->val(elm->second,elm2->second);
//             }
//           }
//         }
//       }
//     }
//     ncutValue = (cut/assoc_A) + (cut/assoc_B);
//     if (ncutValue < bestNcut) {
//       splitPoint_ = split;
//       bestNcut = ncutValue;
//     }
//     last_split = split;
//   }
#ifdef debugout
  std::cout << "---splitPoint " << splitPoint_ << std::endl << std::flush;
#endif

  // all done if this node has no child nodes
  if (depth >= maxDepth-1) return 0;


  // prepare data for child nodes ---------------------

  // calculate eigenvector numbers for both child nodes
  unsigned int left_eigvn  = (eigvn*2);
  unsigned int right_eigvn = left_eigvn+1;
  if (left_eigvn  >= nEigVecs || right_eigvn >= nEigVecs);
  left_eigvn = right_eigvn = 1;

  // the element indices map for the left and right child nodes
  std::map<MSK_ELM_TYPE, unsigned int> left_indices;
  std::map<MSK_ELM_TYPE, unsigned int> right_indices;

  if (left_eigvn != 1 && right_eigvn != 1) {
  // child nodes will only be more eigenvectors of the same ncut
  // (profiles stay the same)

    // construct the element indices lists for both child nodes
    for (elm=indices_.begin(); elm!=indices_.end(); elm++) {
      if (eigVec_[elm->second] > splitPoint_)
        left_indices[elm->first] = elm->second;
      else if (eigVec_[elm->second] < splitPoint_)
        right_indices[elm->first] = elm->second;
    }

    // this node doesn't do anything, it will be deleted
    if (left_indices.empty() || right_indices.empty()) return 1;

    // profiles stay the same
    greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> > left_profile;
    greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> > right_profile;
    left_profile = right_profile = profile_.get();

    // construct the left child node (ncut_ is ignored if left_eigvn == 1)
    leftNode_ = new NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>(left_profile,
                                                         left_indices,
                                                         ncut_, setting_);
    int r = leftNode_->calculate(maxDepth,depth+1,nEigVecs,left_eigvn);
    if (r == 1) leftNode_.reset(NULL);

    // construct the right child node (ncut_ is ignored if right_eigvn == 1)
    rightNode_ = new NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>(right_profile,
                                                          right_indices,
                                                          ncut_, setting_);
    r = rightNode_->calculate(maxDepth,depth+1,nEigVecs,right_eigvn);
    if (r == 1) rightNode_.reset(NULL);
  }
  else {
  // both child node will be a new ncut
  // (profiles change)

    // gather element indices for left and right child nodes in a list
    std::list<unsigned int> left_elements;
    std::list<unsigned int> right_elements;
    for (elm=indices_.begin(); elm!=indices_.end(); elm++) {
//       std::cout << "eigVec_[" << elm->second << "] - " << eigVec_[elm->second] << " - splitPoint " << splitPoint_ << std::endl << std::flush;
      if      (eigVec_[elm->second] > splitPoint_)
        left_elements.push_back(elm->second);
      else if (eigVec_[elm->second] < splitPoint_)
        right_elements.push_back(elm->second);
    }

    // this node doesn't do anything, it will be deleted
    if (left_elements.empty() || right_elements.empty()) return 1;

    // split the similarity matrix of this node twice (for both child nodes)
    std::vector<unsigned int> left_indexMap;
    std::vector<unsigned int> right_indexMap;
    Matrix left_sim(0);
    Matrix right_sim(0);
    Matrix dummy(0);
    profile_->similarity()->split(left_sim, dummy, left_indexMap,
                                  left_elements);
    profile_->similarity()->split(right_sim, dummy, right_indexMap,
                                  right_elements);

    // construct the element indices map for both child nodes
    for (elm=indices_.begin(); elm!=indices_.end(); elm++) {
      if (eigVec_[elm->second] > splitPoint_) {
        left_indices[elm->first] = left_indexMap[elm->second];
      }
      else if (eigVec_[elm->second] < splitPoint_) {
        right_indices[elm->first] = right_indexMap[elm->second];
      }
    }

    // construct the element identifiers map for both child nodes
    std::vector<MSK_ELM_TYPE> left_identifiers(left_indices.size());
    std::vector<MSK_ELM_TYPE> right_identifiers(right_indices.size());
    typename std::map<MSK_ELM_TYPE, unsigned int>::iterator indicesItr;
    for (indicesItr = left_indices.begin();
         indicesItr != left_indices.end(); indicesItr++) {
      left_identifiers[indicesItr->second] = indicesItr->first;
    }
    for (indicesItr = right_indices.begin();
         indicesItr != right_indices.end(); indicesItr++) {
      right_identifiers[indicesItr->second] = indicesItr->first;
    }

    // construct the profiles for both child nodes
    std::auto_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >
      left_profile(new Profile<IMG_ELM_TYPE,MSK_ELM_TYPE>(left_sim,
                                                          left_indices,
                                                          left_identifiers,
                                                          setting_));
    std::auto_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >
      right_profile(new Profile<IMG_ELM_TYPE,MSK_ELM_TYPE>(right_sim,
                                                           right_indices,
                                                           right_identifiers,
                                                           setting_));

    // construct the left child node (ncut_ is ignored if left_eigvn == 1)
    leftNode_ = new NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>(left_profile,
                                                         left_indices,
                                                         ncut_, setting_);
    int r = leftNode_->calculate(maxDepth,depth+1,nEigVecs,left_eigvn);
    if (r == 1) leftNode_.reset(NULL);

    // construct the right child node (ncut_ is ignored if right_eigvn == 1)
    rightNode_ = new NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>(right_profile,
                                                          right_indices,
                                                          ncut_, setting_);
    r = rightNode_->calculate(maxDepth,depth+1,nEigVecs,right_eigvn);
    if (r == 1) rightNode_.reset(NULL);
  }

  return 0;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    NcutNode(std::auto_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& profile,
             const std::map<MSK_ELM_TYPE, unsigned int>& indices,
             greedy_ptr<Ncut>& ncut, const Setting* setting)
  : profile_(profile), indices_(indices), ncut_(ncut), eigVal_(0),
    eigVec_(NULL), splitPoint_(0.0), eigVecNum_(0),
    leftNode_(NULL), rightNode_(NULL), setting_(setting)
{
  if (profile_.get() == NULL)
    throw ncutException("invalid profile (NULL)");

  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    NcutNode(greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& profile,
             const std::map<MSK_ELM_TYPE, unsigned int>& indices,
             greedy_ptr<Ncut>& ncut, const Setting* setting)
  : profile_(profile), indices_(indices), ncut_(ncut), eigVal_(0),
    eigVec_(NULL), splitPoint_(0.0), eigVecNum_(0),
    leftNode_(NULL), rightNode_(NULL), setting_(setting)
{
  if (profile_.get() == NULL)
    throw ncutException("invalid profile (NULL)");

  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
    NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::~NcutNode()
{
  return; // everything done with greedy_ptr
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    NcutNode(const NcutNode& clone)
  : profile_(NULL), indices_(), ncut_(NULL), eigVal_(0),
    eigVec_(NULL), splitPoint_(0.0), eigVecNum_(0),
    leftNode_(NULL), rightNode_(NULL), setting_(clone.setting_)
{
  *this = clone;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    NcutNode(const NcutNode& clone,
             greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& pclone,
             greedy_ptr<Ncut>& nclone)
  : profile_(NULL), indices_(), ncut_(NULL), eigVal_(0),
    eigVec_(NULL), splitPoint_(0.0), eigVecNum_(0),
    leftNode_(NULL), rightNode_(NULL), setting_(clone.setting_)
{
  copy(clone,pclone,nclone);
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>& NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    operator=(const NcutNode& clone)
{
  greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> > pclone(NULL);
  greedy_ptr<Ncut> nclone(NULL);
  return copy(clone,pclone,nclone);
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>& NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    copy(const NcutNode& clone,
        greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> >& pclone,
        greedy_ptr<Ncut>& nclone)
{
  // copy simple members
  this->indices_    = clone.indices_;
  this->eigVal_     = clone.eigVal_;
  this->splitPoint_ = clone.splitPoint_;
  this->eigVecNum_  = clone.eigVecNum_;

  // copy the profile:
  // if the clone owns it, copy the profile
  // if the clone does not own it and a profile pointer was handed down, use it
  // if no profile pointer was handed down, use the clone's profile pointer
  if (clone.profile_.owns())
    this->profile_ = new Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>(*clone.profile_);
  else if (pclone.get() != NULL)
    this->profile_ = pclone;
  else
    this->profile_ = clone.profile_;

  // copy the ncut:
  // if the clone owns it, copy the ncut
  // if the clone does not own it and an ncut pointer was handed down, use it
  // if no ncut pointer was handed down, use the clone's ncut pointer
  if (clone.ncut_.owns()) {
    this->ncut_ = new Ncut(*clone.ncut_);
    this->eigVec_ =
        &(this->ncut_->eigVec()[this->eigVecNum_*this->ncut_->dim()]);
  }
  else if (nclone.get() != NULL) {
    this->ncut_ = nclone;
    this->eigVec_ =
        &(this->ncut_->eigVec()[this->eigVecNum_*this->ncut_->dim()]);
  }
  else {
    this->ncut_ = clone.ncut_;
    this->eigVec_ = clone.eigVec_;
  }

  // copy the child nodes and make sure they use pointers
  // to this node's profile and ncut, not to the clone's.
  if (clone.leftNode_.get() != NULL)
    this->leftNode_ =
        new NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>(*clone.leftNode_,
                                                 this->profile_,
                                                 this->ncut_);
  if (clone.rightNode_.get() != NULL)
    this->rightNode_ =
        new NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>(*clone.rightNode_,
                                                 this->profile_,
                                                 this->ncut_);

  // copy the setting
  this->setting_ = clone.setting_;

  return *this;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    getInfo(const NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>* node,
            std::vector<std::vector<MSK_ELM_TYPE> >& elements)
{
  if (node == NULL || node->eigVec() == NULL)
    throw ncutException("invalid node, node corrupt");

  // get pointers to both child nodes
  const NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>* leftChild =
    node->leftNode().get();
  const NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>* rightChild =
    node->rightNode().get();

  if (leftChild == NULL || rightChild == NULL) {
  // this is a (half-)leaf node

    // get current segment numbers of both segments
    // and create vectors to hold the element id's
    unsigned int leftSegNum;
    unsigned int rightSegNum;
    if (leftChild == NULL) {
      leftSegNum = elements.size();
      elements.push_back(std::vector<MSK_ELM_TYPE>());
    }
    if (rightChild == NULL) {
      rightSegNum = elements.size();
      elements.push_back(std::vector<MSK_ELM_TYPE>());
    }

    // save element id's for the two segments in separate vectors
    unsigned int segNum;
    typename std::map<MSK_ELM_TYPE, unsigned int>::const_iterator elmItr;
    for (elmItr = node->indices()->begin();
         elmItr != node->indices()->end(); elmItr++) {

      if (node->eigVec()[elmItr->second] > node->splitPoint()) {
        if (leftChild != NULL) continue;
        segNum = leftSegNum;
      }
      else {
        if (rightChild != NULL) continue;
        segNum = rightSegNum;
      }

      elements[segNum].push_back(elmItr->first);
    }
  }

  // recurse to child nodes
  if (leftChild != NULL)
    getInfo(leftChild, elements);
  if (rightChild != NULL)
    getInfo(rightChild, elements);
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
std::vector<std::vector<MSK_ELM_TYPE> >
Segmentation<IMG_ELM_TYPE,MSK_ELM_TYPE>::
    merge(const SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>& profile)
{
  unsigned int minSegCount = setting_.minimum_segment_count;
  double mergeThreshold = setting_.merge_threshold;

  std::vector<std::vector<MSK_ELM_TYPE> > elements = *profile.elements();

  if (profile.nSeg() <= 1)
    return elements;
  if (profile.similarity()->dim() != profile.nSeg())
    throw ncutException(
        "invalid segment profile (similarity matrix not calculated)." );
  if (profile.elmProfile() == NULL)
    throw ncutException("invalid element profile (NULL)");
  if (minSegCount <= 1)
    throw ncutException("invalid minimum segment count, must be >= 2");
  if (mergeThreshold < 0.0 || mergeThreshold >= 500.0)
    throw ncutException("invalid merge threshold, must be in [0,500)");

// TODO: use already calculated profile similarity matrix instead of
//       calculating a new one (to be tested)
//   Matrix difference(*profile->similarity());

  // calculate the differences of each pair of segments
  Matrix difference(profile.nSeg());
  try {
    for (unsigned int seg1=0; seg1<profile.nSeg(); seg1++) {
      for (unsigned int seg2=seg1+1; seg2<profile.nSeg(); seg2++) {
        difference.val(seg2,seg1) = difference.val(seg1,seg2) =
            profile.dif(seg1, seg2);
      }
    }
  }
  catch (ncutException& e) {
    throw ncutCritical(
        (std::string("Error: Segmentation left in inconsistent state: ")+
        e.what() ).c_str() );
  }

  // merge segments until one of the stopping criteria is met
  bool active [profile.nSeg()];
  for (unsigned int seg=0; seg<profile.nSeg(); seg++)
    active[seg] = true;
  unsigned int mergeSeg1;
  unsigned int mergeSeg2;
  unsigned int segCount = profile.nSeg();
  while(segCount >= minSegCount) {

    // search for the pair of segments with the smallest difference
    double minDif = 600.0; // dif shouldn't be bigger than 500
    for (unsigned int seg1=0; seg1<profile.nSeg(); seg1++) {
      if (!active[seg1]) continue;
      for (unsigned int seg2=seg1+1; seg2<profile.nSeg(); seg2++) {
        if (!active[seg2]) continue;
        if (difference.val(seg1,seg2) < minDif) {
          mergeSeg1 = seg1;
          mergeSeg2 = seg2;
          minDif = difference.val(seg1,seg2);
        }
      }
    }

    // if the smallest difference exceeds the merge threshold stop merging
#ifdef debugout
    std::cout << "would merge " << minDif << std::endl << std::flush;
#endif

    if (minDif > mergeThreshold) break;
#ifdef debugout
    std::cout << " ... and will" << minDif << std::endl << std::flush;
#endif

    // else merge the two segments with smallest difference
    for (unsigned int elm=0; elm<elements[mergeSeg2].size(); elm++)
      elements[mergeSeg1].push_back(elements[mergeSeg2][elm]);
    active[mergeSeg2] = false;

    // update the difference matrix
    try {
      for (unsigned int seg=0; seg<elements.size(); seg++) {
        if ((!active[seg]) || seg == mergeSeg1) continue;

        difference.val(seg,mergeSeg1) = difference.val(mergeSeg1,seg) =
            profile.dif(mergeSeg1, seg, &elements);
      }
    }
    catch (ncutException& e) {
      throw ncutCritical(
          (std::string("Error: Segmentation left in inconsistent state: ")+
          e.what() ).c_str() );
    }

    segCount--;
  }

  // delete inactive segments
  // (segments that have been merged into another segment)
  unsigned int segIdx;
  typename std::vector<std::vector<MSK_ELM_TYPE> >::iterator seg;
  for (seg=elements.begin(), segIdx=0; seg!=elements.end(); segIdx++) {
    if (!active[segIdx])
      seg = elements.erase(seg);
    else
      seg++;
  }

  return elements;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    connected(const SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>& segProfile,
              unsigned int startElmIdx, std::vector<bool>& reached)
{
  if (startElmIdx >= segProfile.nElm())
    throw ncutException(
        "invalid start element index or invalid segment profile" );

  const Matrix* neighbors = segProfile.elmProfile()->neighbors();
  unsigned int nElm = segProfile.nElm();

  unsigned int startSegIdx = segProfile.idxMask(startElmIdx);

  reached[startElmIdx] = true;

  for (unsigned int elm=0; elm<nElm; elm++) {
    if (neighbors->val(startElmIdx,elm) > 0.5 && !reached[elm]) {
    // is neighbor and not reached
      if (segProfile.idxMask(elm) == startSegIdx ) {
      // same segment
        connected(segProfile, elm, reached);
      }
    }
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
std::vector<std::vector<MSK_ELM_TYPE> >
Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    split(const SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE>& profile)
{
  if (profile.elmProfile() == NULL)
    throw ncutException("invalid element profile (NULL)");

  // split parts of segments that aren't touching
  std::vector<std::vector<MSK_ELM_TYPE> > elements;
  std::map<MSK_ELM_TYPE, unsigned int>
      unassigned(*profile.elmProfile()->indices());
  std::vector<bool> reached(profile.nElm(), false);
  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator elmItr;
  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator eraseItr;
  while (!unassigned.empty()) {

    // mark elements as reached if they can be reached from this element
    // without crossing other segments
    connected(profile, unassigned.begin()->second, reached);

    // save those elements marked as reached as one segment and remove them
    // from the unassigned list
    elements.push_back(std::vector<MSK_ELM_TYPE>());
    for (elmItr = unassigned.begin(); elmItr!=unassigned.end();) {
      if (reached[elmItr->second]) {
        elements.back().push_back(elmItr->first);
        eraseItr = elmItr++;
        unassigned.erase(eraseItr);
      }
      else {
        elmItr++;
      }
    }
  }

  return elements;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
int Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    push(const Image<IMG_ELM_TYPE>* frame, const Image<MSK_ELM_TYPE>* mask)
{
  unsigned int depth    = setting_.ncut_tree_depth;
  unsigned int nEigVecs = setting_.number_of_eigenvectors;
  bool doMergeSegments  = (setting_.merge_segments == 1);
  bool doSplitSegments  = (setting_.connected_segments == 1);

  if (depth == 0)
    throw ncutException("invalid depth, must be >= 1");
  if (nEigVecs < 2)
    throw ncutException(
        "invalid number of eigenvectors per ncut, must be >= 2" );

  // save information needed to calculate the ncut starting vector
  unsigned int nElm_last = 0;
  unsigned int nElm_front = 0;
  double* assoc_last = NULL;
  if (dendogram_ != NULL) {
    nElm_last = elmProfile_.nElm();
    nElm_front = elmProfile_.maxIdx(0)+1;
    assoc_last = new double [nElm_last];
    for (unsigned int elm_last=0; elm_last<nElm_last; elm_last++)
      assoc_last[elm_last] = dendogram_->ncut()->assoc(elm_last);
  }

  // push frame and mask to the element profile
  if (elmProfile_.push(frame, mask) == 1)
    return 1; // frame window is not filled yet

  // calculate the ncut starting vector (to be tested)
  double* initVec = NULL;
//   if (dendogram_ != NULL) {
// 
//     // calculate the matrix of the associated non-generalized
//     // eigenvalue problem
//     Matrix a(elmProfile_.nElm());
//     double inv_sqrt_assoc [elmProfile_.nElm()];
//     for (unsigned int i=0; i<elmProfile_.nElm(); i++) {
//       if (dendogram_->ncut()->assoc(i) == 0.0)
//         throw ncutCritical("invalid total association value (0), "
//                            "Segmentation left in inconsistent state");
//       inv_sqrt_assoc[i] = 1.0/sqrt(dendogram_->ncut()->assoc(i));
//     }
//     for (unsigned int i=0; i<elmProfile_.nElm(); i++)
//       for (unsigned int j=0; j<elmProfile_.nElm(); j++)
//         a.val(i,j) = ( ((i==j)?dendogram_->ncut()->assoc(i):0.0) -
//                       dendogram_->ncut()->similarity()->val(i,j)) *
//                      inv_sqrt_assoc[j] * inv_sqrt_assoc[i];
// 
//     // calculate the eigenvector of the associated non-generalized
//     // eigenvalue problem of the last iteration,
//     // shifted to match the indices in this iteration
//     initVec = new double [elmProfile_.nElm()];
//     unsigned int elm, elm_last;
//     for (elm_last=nElm_front, elm=0; elm_last<nElm_last; elm_last++, elm++)
//       initVec[elm] = dendogram_->eigVec()[elm_last] *
//                      sqrt(assoc_last[elm_last]);
//     delete [] assoc_last;
//     for (elm=elm;elm<elmProfile_.nElm(); elm++)
//       initVec[elm] = 0.0;
// 
//     // transform the non-generalized eigenvector of the last iteration
//     // with the matrix of the non-generalized eigenvalue problem of this
//     // iteration
//     a.transform(initVec);
//   }

  // construct the ncut dendogram
  delete dendogram_;
  greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> > elmProfile_p(
      *dynamic_cast<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE>*>(&elmProfile_) );
  greedy_ptr<Ncut> ncut_p (NULL);
  dendogram_ = new NcutNode<IMG_ELM_TYPE, MSK_ELM_TYPE>(elmProfile_p,
                                                        *elmProfile_.indices(),
                                                        ncut_p, &setting_);
  if (dendogram_->calculate(depth,0,nEigVecs,1,initVec) == 1) {
    delete dendogram_;
    delete [] initVec;
    throw ncutException("Couldn't find any ncut");
  }
  delete [] initVec;

  // fetch segment information from nodes
  std::vector<std::vector<MSK_ELM_TYPE> > elements;
  getInfo(dendogram_, elements);

  // merge segments
  if (doMergeSegments) {

    SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE> preMergeProfile(&setting_);
    preMergeProfile.push(&elmProfile_, elements);
    preMergeProfile.calcSim();

    elements = merge(preMergeProfile);
  }

  // split parts of segments that aren't touching
  if (doSplitSegments) {

    SegmentProfile<IMG_ELM_TYPE, MSK_ELM_TYPE> preSplitProfile(&setting_);
    preSplitProfile.push(&elmProfile_, elements);

    elements = split(preSplitProfile);
  }

  segProfile_.push(&elmProfile_, elements);
  
#ifdef debugout  
  for (unsigned int i=0; i<nSeg(); i++) {
    std::cout << "id " << id(i) << " area " << segProfile_.nPixels(i) << std::endl << std::flush;
  }
#endif

  return 0;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    Segmentation(unsigned int wSize, Setting setting)
  : setting_(setting), dendogram_(NULL), elmProfile_(wSize, &setting_),
    segProfile_(&setting_)
{
  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    Segmentation(const Segmentation& clone)
  : setting_(clone.setting_), dendogram_(NULL), elmProfile_(3), segProfile_()
{
  *this=clone;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::cleanup()
{
  delete dendogram_;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::~Segmentation()
{
  cleanup();
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>&
Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    operator=(const Segmentation& clone)
{
  // TODO: this does not create a completely independent copy, all setting
  //       pointers will still point to clone.setting_

  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  cleanup();

  // copy the element profile and the segment profiles
  this->elmProfile_ = clone.elmProfile_;
  this->segProfile_ = clone.segProfile_;

  // copy the dendogram with pointers to this profile (not the clone's)
  greedy_ptr<Profile<IMG_ELM_TYPE, MSK_ELM_TYPE> > pclone(&this->elmProfile_);
  greedy_ptr<Ncut> nclone(NULL);
  if (clone.dendogram_ != NULL)
    this->dendogram_ =
        new NcutNode<IMG_ELM_TYPE,MSK_ELM_TYPE>(*clone.dendogram_,
                                                pclone, nclone);
  else
    this->dendogram_ = NULL;

  this->setting_ = clone.setting_;

  return *this;
}

} // namespace ncut

#endif // NCUT_CORE_H
