#include "ncutImage.h"

#include <math.h>
#include <assert.h>
#include <string>
#include <sstream>
#include <iostream>
#include <stdlib.h>

namespace ncut
{

int PNGImage::load(const char *filename)
{
  // free old data
  if (!imageData_.empty()) imageData_.clear();
  width_ = 0;
  height_ = 0;
  nChl_ = 0;

  // open the file
  FILE *fp = fopen(filename, "rb");
  if (!fp) {
    throw ncutException(
      (std::string("Could not open file ")+filename).c_str() );
  }

  // check if file is a .png file
  int numBytes = 8;
  png_byte *header = new png_byte[numBytes];
  fread(header, 1, numBytes, fp);
  bool is_png = !png_sig_cmp(header, 0, numBytes);
  if (!is_png) {
    fclose(fp);
    throw ncutException(
        (filename+std::string(" is not a png file")).c_str() );
  }

  if (header != NULL) {
    delete [] header;
    header = NULL;
  }

  // initialize the data structures
  png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
                                               (png_voidp)NULL,
                                               NULL, NULL);
  if (!png_ptr) {
    fclose(fp);
    throw ncutException(
      (std::string("Could not read png file ")+filename).c_str() );
  }

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
    fclose(fp);
    throw ncutException(
        (std::string("Could not read png file ")+filename).c_str() );
  }

  png_infop end_info = png_create_info_struct(png_ptr);
  if (!end_info) {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    fclose(fp);
    throw ncutException(
        (std::string("Could not read png file ")+filename).c_str() );
  }

  png_init_io(png_ptr, fp);
  png_set_sig_bytes(png_ptr, numBytes);

  // load the image
  png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

  // get image info
  width_  = png_get_image_width(png_ptr, info_ptr);
  height_ = png_get_image_height(png_ptr, info_ptr);
  nChl_   = png_get_channels(png_ptr, info_ptr);

  // get pointers to image data
  // (array of pointers to rows in the image)
  png_bytep *row_pointers = new png_bytep[height_];
  row_pointers = png_get_rows(png_ptr, info_ptr);

  // initialize the local structure
  imageData_.resize(width_*height_*nChl_);

  // copy image data to local structure
  for (unsigned int row=0; row<height_; row++) {
    for (unsigned int elm=0; elm<width_*nChl_; elm++) {
      imageData_[(row*width_*nChl_)+elm] = row_pointers[row][elm];
    }
  }

  // free the libpng data structures
  png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);

  // close the .png file
  fclose(fp);

  return 0;
}

int PNGImage::save(const char *filename) const
{
  FILE *fp = fopen(filename, "wb");
  if (!fp) {
    throw ncutException(
        (std::string("Could not open file ")+filename).c_str() );
  }

  // initialize the data structures
  png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
                                                (png_voidp)NULL,
                                                NULL, NULL);
  if (!png_ptr) {
    fclose(fp);
    throw ncutException(
        (std::string("Could not write png file ")+filename).c_str() );
  }

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
    fclose(fp);
    throw ncutException(
        (std::string("Could not write png file ")+filename).c_str() );
  }

  if (setjmp(png_jmpbuf(png_ptr))) {
    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fp);
    throw ncutException(
        (std::string("Could not write png file ")+filename).c_str() );
  }

  png_init_io(png_ptr, fp);

  info_ptr->width=width_;
  info_ptr->height=height_;
  info_ptr->bit_depth=8;
  info_ptr->color_type=PNG_COLOR_TYPE_RGB;
  info_ptr->interlace_type=1;
  info_ptr->valid=0;
  info_ptr->channels=nChl_;

  png_write_info(png_ptr,info_ptr);

  png_byte *row_pointers[height_];

  for(unsigned int y=0; y<height_; y++)
  {
    row_pointers[y] = new png_byte[info_ptr->channels*width_];

    for(unsigned int element=0; element<width_*info_ptr->channels;element++) {
      row_pointers[y][element] =
        imageData_[(y*width_*info_ptr->channels)+element];
    }
  }

  png_write_image(png_ptr,row_pointers);

  for(unsigned int y=0; y<height_; y++)
  {
    if (row_pointers[y] != NULL) 
      delete [] row_pointers[y];
  }

  png_write_end(png_ptr, NULL);
  png_destroy_write_struct(&png_ptr, &info_ptr);

  fclose(fp);
  return 0;
}

PNGImage::PNGImage()
  : Image<PNG_IMAGE_DATATYPE>()
{
  return;
}

PNGImage::PNGImage(const PNGImage& clone)
  : Image<PNG_IMAGE_DATATYPE>()
{
  *this = clone;
}

PNGImage::PNGImage(const char *filename)
  : Image<PNG_IMAGE_DATATYPE>()
{
  load(filename);
}


PNGImage::PNGImage(int width, int height, int nChl)
  : Image<PNG_IMAGE_DATATYPE>(width, height, nChl)
{
  return;
}


PNGImage::~PNGImage()
{
  if (!imageData_.empty()) imageData_.clear();
}

PNGImage& PNGImage::operator=(const PNGImage& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // clear old data
  if (!imageData_.empty()) imageData_.clear();

  // copy image info from clone
  this->width_ = clone.width_;
  this->height_ = clone.height_;
  this->nChl_ = clone.nChl_;

  // initialize the local structure
  imageData_.resize(width_*height_*nChl_);

  // copy image data from the clone
  for (unsigned int i=0; i<width_*height_*nChl_; i++) {
    this->imageData_[i] = clone.imageData_[i];
  }

  return *this;
}

int BSIImage::load(const char *filename, unsigned long modNum)
{
  // free old data
  if (!imageData_.empty()) imageData_.clear();
  width_    = 0;
  height_   = 0;
  nChl_     = 0;
  nSeg_     = 0;

  // load the BSI Image into memory
  segmentation::TImageSegmentation<signed long> *seg =
    new segmentation::TImageSegmentation<signed long>();
  long r = segmentation::CImageSegmentationIndexFile::Load(filename, *seg);
  if (r != 1) {
    throw ncutException(
        (std::string("Could not load BSI Image ")+filename).c_str() );
  }

  // copy the image info
  width_  = seg->Width();
  height_ = seg->Height();
  nChl_   = 1;
  nSeg_   = seg->GetSegmentCount();

  // prepare the local structure
  imageData_.resize(width_*height_*nChl_);

  std::map <unsigned long, unsigned long> labelMask;
  // copy the data into the local structure
  BSI_IMAGE_DATATYPE label;
  for (unsigned int row=0; row<height_; row++) {
    for (unsigned int col=0; col<width_; col++) {
      if (labelMask[((*seg)(col,row))] == 0) {
        label = modNum++; 
        labelMask[((*seg)(col,row))] = label;
      }
      else {
        label = labelMask[((*seg)(col,row))];
      }
      imageData_[(row*width_)+col] = label;
    }
  }

  // free temporary memory
  delete seg;
  seg = NULL;

  return modNum;
}

BSIImage::BSIImage()
  : Image<BSI_IMAGE_DATATYPE>(), nSeg_(0)
{
  nChl_ = 1;
}

BSIImage::BSIImage(const BSIImage& clone)
  : Image<BSI_IMAGE_DATATYPE>(), nSeg_(0)
{
  *this = clone;
}

BSIImage::BSIImage(const char *filename, unsigned long& modNum)
  : Image<BSI_IMAGE_DATATYPE>(), nSeg_(0)
{
  modNum = load(filename, modNum);
}

BSIImage::~BSIImage()
{
  if (!this->imageData_.empty()) this->imageData_.clear();
}

BSIImage& BSIImage::operator=(const BSIImage& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // clear old data
  if (!this->imageData_.empty()) this->imageData_.clear();

  // copy image info from clone
  this->width_ = clone.width_;
  this->height_ = clone.height_;
  this->nChl_ = clone.nChl_;
  this->nSeg_ = clone.nSeg_;

  // initialize the local structure
  this->imageData_.resize(width_*height_*nChl_);

  // copy image data from the clone
  for (unsigned int i=0; i<width_*height_*nChl_; i++) {
    this->imageData_[i] = clone.imageData_[i];
  }

  return *this;
}

int PNGSequence::pushThrough(const Image<PNG_IMAGE_DATATYPE>& frame)
{
  const PNGImage* frame_p;
  try {
    frame_p = dynamic_cast<const PNGImage*>(&frame);
  }
  catch (std::bad_cast& e) {
    throw ncutException(e.what());
  }
  if (frame_p == NULL) throw ncutException("cast to PNGImage failed");

  return pushThrough(*frame_p);
}

int PNGSequence::pushThrough(const PNGImage& frame)
{
  if (!validateFrame(frame)) {
    throw ncutException("frame does not fit into the sequence (wrong format?)");
  }

  if (frames_.empty()) {
    pushFrame(frame);
    return 0;
  }

  if(frames_.front() != NULL) {
    PNGImage *firstFrame = static_cast<PNGImage*>(frames_.front());

    delete firstFrame;
  }

  frames_.pop_front();

  PNGImage *clone = new PNGImage(frame);

  frames_.push_back(clone);

  return 0;
}

int PNGSequence::pushFrame(const Image<PNG_IMAGE_DATATYPE>& frame)
{
  const PNGImage* frame_p;
  try {
    frame_p = dynamic_cast<const PNGImage*>(&frame);
  }
  catch (std::bad_cast& e) {
    throw ncutException(e.what());
  }
  if (frame_p == NULL) throw ncutException("cast to PNGImage failed");

  return pushFrame(*frame_p);
}

int PNGSequence::pushFrame(const PNGImage& frame)
{
  if (!validateFrame(frame)) {
    throw ncutException("frame does not fit into the sequence (wrong format?)");
  }

  PNGImage *clone = new PNGImage(frame);

  frames_.push_back(clone);

  return 0;
}

void PNGSequence::popFrame()
{
  if (!frames_.empty()) {
    if (frames_.back() != NULL) {
      PNGImage *lastFrame = static_cast<PNGImage*>(frames_.back());

      delete lastFrame;
    }
    frames_.pop_back();
  }
}

int PNGSequence::save(const char* filename) const
{
  for (unsigned long i=0; i<length(); i++) {
    std::stringstream currentFilename;
    if (i < 10) {
      currentFilename << filename << "00" << i << ".png";
    }
    else if (i < 100) {
      currentFilename << filename << "0" << i << ".png";
    }
    else {
      currentFilename << filename << i << ".png";
    }
// 
//     int r = (*this)[i].save(currentFilename.str().c_str());
//     if (r != 0) return r;
  }

//     int r = (*this)[1].save(CurrentFilename.str().c_str());

    (*this)[1]->save(filename);

//     std::stringstream currentFilename0;
//     std::stringstream currentFilename1;
//     std::stringstream currentFilename2;
//     currentFilename0 << filename << "-0";
//     currentFilename1 << filename << "-1";
//     currentFilename2 << filename << "-2";
//     int r = (*this)[0]->save(currentFilename0.str().c_str());
//     if (r != 0) return r;
//     r = (*this)[1]->save(currentFilename1.str().c_str());
//     if (r != 0) return r;
//     r = (*this)[2]->save(currentFilename2.str().c_str());
//     if (r != 0) return r;

  return 0;
}

PNGSequence::PNGSequence()
  : Sequence<PNG_IMAGE_DATATYPE>()
{
  return;
}

PNGSequence::PNGSequence(const PNGSequence& clone)
  : Sequence<PNG_IMAGE_DATATYPE>()
{
  *this = clone;
}

PNGSequence::~PNGSequence()
{
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!this->frames_.empty()) this->frames_.clear();
}

PNGSequence& PNGSequence::operator=(const PNGSequence& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!this->frames_.empty()) this->frames_.clear();

  // copy sequence info from clone
  this->width_  = clone.width_;
  this->height_ = clone.height_;
  this->nChl_   = clone.nChl_;

  // initialize the local structure
  this->frames_.resize(clone.frames_.size());

  // copy the sequence data from the clone
  for (unsigned int i=0; i<clone.frames_.size(); i++) {
    if (clone.frames_[i] != NULL) {
      PNGImage* cloneFrame;
      try {
        cloneFrame = dynamic_cast<PNGImage*>(clone.frames_[i]);
      }
      catch (std::bad_cast& e) {
        throw ncutException(e.what());
      }
      if(cloneFrame == NULL) throw ncutException("cast to PNGImage failed");
      this->frames_[i] = new PNGImage(*(cloneFrame));
    }
    else {
      this->frames_[i] = NULL;
    }
  }

  return *this;
}

PNGImage* PNGSequence::operator[](unsigned long pos)
{
  assert(pos < frames_.size()); // alle asserts kommen am ende weg
  if(pos >= frames_.size()) pos = frames_.size()-1;

  return static_cast<PNGImage*>(frames_[pos]);
}

const PNGImage* PNGSequence::operator[](unsigned long pos) const
{
  assert(pos < frames_.size()); // alle asserts kommen am ende weg
  if(pos >= frames_.size()) pos = frames_.size()-1;

  return static_cast<PNGImage*>(frames_[pos]);
}

int BSISequence::pushThrough(const Image<BSI_IMAGE_DATATYPE>& frame)
{
  const BSIImage* frame_p;
  try {
    frame_p = dynamic_cast<const BSIImage*>(&frame);
  }
  catch (std::bad_cast& e) {
    throw ncutException(e.what());
  }
  if (frame_p == NULL) throw ncutException("cast to BSIImage failed");

  return pushThrough(*frame_p);
}

int BSISequence::pushThrough(const BSIImage& frame)
{
  if (!validateFrame(frame)) {
    throw ncutException("frame does not fit into the sequence (wrong format?)");
  }

  if (frames_.empty()) {
    pushFrame(frame);
    return 0;
  }

  if(frames_.front() != NULL) {
    BSIImage *firstFrame = static_cast<BSIImage*>(frames_.front());

    if (nSeg_ < firstFrame->nSeg()) nSeg_ = 0;
    else                            nSeg_ -= firstFrame->nSeg();

    delete firstFrame;
  }

  frames_.pop_front();

  nSeg_ += frame.nSeg();

  BSIImage *clone = new BSIImage(frame);

  frames_.push_back(clone);

  return 0;
}

int BSISequence::pushFrame(const Image<BSI_IMAGE_DATATYPE>& frame)
{
  const BSIImage* frame_p;
  try {
    frame_p = dynamic_cast<const BSIImage*>(&frame);
  }
  catch (std::bad_cast& e) {
    throw ncutException(e.what());
  }
  if (frame_p == NULL) throw ncutException("cast to BSIImage failed");

  return pushFrame(*frame_p);
}

int BSISequence::pushFrame(const BSIImage& frame)
{
  if (!validateFrame(frame)) {
    throw ncutException("frame does not fit into the sequence (wrong format?)");
  }

  nSeg_ += frame.nSeg();

  BSIImage *clone = new BSIImage(frame);

  frames_.push_back(clone);

  return 0;
}

void BSISequence::popFrame()
{
  if (!frames_.empty()) {
    if (frames_.back() != NULL) {
      BSIImage *lastFrame = static_cast<BSIImage*>(frames_.back());

      if (nSeg_ < lastFrame->nSeg()) nSeg_ = 0;
      else                           nSeg_ -= lastFrame->nSeg();

      delete lastFrame;
    }
    frames_.pop_back();
  }
}

int BSISequence::clear()
{
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!frames_.empty()) frames_.clear();

  width_    = 0;
  height_   = 0;
  nChl_     = 0;
  nSeg_     = 0;

  return 0;
}

BSISequence::BSISequence()
  : Sequence<BSI_IMAGE_DATATYPE>(), nSeg_(0)
{
  return;
}

BSISequence::BSISequence(const BSISequence& clone)
  : Sequence<BSI_IMAGE_DATATYPE>(), nSeg_(0)
{
  *this = clone;
}

BSISequence::~BSISequence()
{
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!this->frames_.empty()) this->frames_.clear();
}

BSISequence& BSISequence::operator=(const BSISequence& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!this->frames_.empty()) this->frames_.clear();

  // copy sequence info from clone
  this->width_  = clone.width_;
  this->height_ = clone.height_;
  this->nChl_   = clone.nChl_;
  this->nSeg_   = clone.nSeg_;

  // initialize the local structure
  this->frames_.resize(clone.frames_.size());

  // copy the sequence data from the clone
  for (unsigned int i=0; i<clone.frames_.size(); i++) {
    if (clone.frames_[i] != NULL) {
      BSIImage* cloneFrame;
      try {
        cloneFrame = dynamic_cast<BSIImage*>(clone.frames_[i]);
      }
      catch (std::bad_cast& e) {
        throw ncutException(e.what());
      }
      if (cloneFrame == NULL) throw ncutException("cast to BSIImage failed");

      this->frames_[i] = new BSIImage(*(cloneFrame));
    }
    else {
      this->frames_[i] = NULL;
    }
  }

  return *this;
}

BSIImage* BSISequence::operator[](unsigned long pos)
{
  assert(pos < frames_.size()); // alle asserts kommen am ende weg
  if (pos >= frames_.size()) pos = frames_.size()-1;

  return static_cast<BSIImage*>(frames_[pos]);
}

const BSIImage* BSISequence::operator[](unsigned long pos) const
{
  assert(pos < frames_.size()); // alle asserts kommen am ende weg
  if (pos >= frames_.size()) pos = frames_.size()-1;

  return static_cast<BSIImage*>(frames_[pos]);
}

} //namespace ncut
