#ifndef NCUT_IMAGE_H
#define NCUT_IMAGE_H 1

#include "TImageSegmentation.h"
#include "CImageSegmentationIndexFile.h"
#include "ncutException.h"

#include <png.h>
#include <vector>
#include <deque>
#include <iostream>

#define PNG_IMAGE_DATATYPE png_byte
#define BSI_IMAGE_DATATYPE signed long

namespace ncut
{

const PNG_IMAGE_DATATYPE PNG_IMAGE_DATATYPE_MAXVAL = 255;
const PNG_IMAGE_DATATYPE PNG_IMAGE_DATATYPE_MINVAL = 0;
const unsigned int IMAGE_DATATYPE_MAXVAL = 255;

/**
 * @c Image is the base class for @c PNGImage and @c BSIImage.
 * @c Image is a small wrapper around a STL vector containing the raw image
 * values, providing additional information, such as width and height, and more
 * convenient image element access functions.
 */
template <class ELEMENT_TYPE = unsigned char>
class Image
{
  public:

    // functions for read & write access to data members ------

    /**
     * @brief Get the value of the colour channel of the specified element.
     * @param elm the element
     * @return the value of the colour channel
     */
    inline ELEMENT_TYPE& val(unsigned int elm)
    {return imageData_[elm];};
    /**
     * @brief Get the value of the colour channel of the specified pixel and channel.
     * @param pxl the specified pixel
     * @param chl the colour channel
     * @return the value of the colour channel
     */
    inline ELEMENT_TYPE& val(unsigned int pxl, unsigned int chl)
    {return imageData_[(pxl*nChl_)+chl];};

    // functions for read-only access to data members ------

    /**
     * @brief Get the constant value of the colour channel of the specified element.
     * @param elm the element
     * @return The constant value of the colour channel
     */
    inline const ELEMENT_TYPE& val(unsigned int elm) const
    {return imageData_[elm];};
    /**
     * @brief Get the value of the colour channel of the specified pixel and channel.
     * @param pxl the specified pixel
     * @param chl the colour channel
     * @return The value of the colour channel
     */
    inline const ELEMENT_TYPE& val(unsigned int pxl, unsigned int chl) const
    {return imageData_[(pxl*nChl_)+chl];};
    /**
     * @brief Get the with of the image.
     * @return The width of the image
     */
    inline unsigned int width() const
    {return width_;};
    /**
     * @brief Get the height of the image.
     * @return The height of the image
     */
    inline unsigned int height() const
    {return height_;};
    /**
     * @brief Get the number of channels of the image.
     * @return The number of channles 
     */
    inline unsigned int nChl() const
    {return nChl_;};

    // standard functions ------

    /**
     * @brief %Image empty constructor.
     *
     * Creates an empty image with width=height=channels=0.
     */
    Image();
    /**
     * @brief %Image custom constructor.
     *
     * @param width  The width of the image.
     * @param height The height of the image.
     * @param nChl   The number of channels of the image.
     *
     * Creates an empty image with the specified dimensions
     * and initializes the data to zero.
     */
    Image(int width, int height, int nChl);
              
    /**
     * @brief %Image copy constructor.
     *
     * @param clone The @c Image to be cloned.
     */
    Image(const Image& clone);

    /**
     * @brief %Image destructor.
     */
    virtual ~Image();

    /**
     * @brief %Image assignment operator.
     *
     * @param clone The @c Image to be cloned.
     * @return This object.
     */
    virtual Image& operator=(const Image& clone);

  protected:

    std::vector<ELEMENT_TYPE> imageData_; /**< The raw image values stored
        sequentially in the form pxl0chl0, pxl0chl1, ... , pxl1chl0, ...*/
    unsigned int width_;/**< The width of this image.*/
    unsigned int height_;/**< The height of this image.*/
    unsigned int nChl_; /**< The number of
        channels of this image.*/
};

/**
 * @c PNGImage adds PNG-specific load and save functionality to its base class
 * @c Image.
 */
class PNGImage : public Image<PNG_IMAGE_DATATYPE>
{
  public:

    // class-specific functions ------

    /**
     * @brief Load the PNG image file "filename".
     *
     * @param filename A char array specifying the location of a PNG image.
     * @return 0 if successful.
     *
     * Replaces this PNG image with the PNG image located at @c filename.
     */
    int load(const char* filename);
    /**
     * @brief Save this %PNGImage into the file "filename".
     *
     * @param filename A char array specifying the destination of this
     *                 @c PNGimage.
     * @return 0 if successful.
     */
    int save(const char* filename) const;

    // standard functions ------

    /**
     * @brief %PNGImage empty constructor.
     *
     * Creates an empty image with width=height=channels=0.
     */
    PNGImage();
    /**
     * @brief %PNGImage copy constructor.
     *
     * @param clone The @c PNGImage to be cloned.
     */
    PNGImage(const PNGImage& clone);
    /**
     * @brief A %PNGImage can be contructed using a PNG image stored on disk.
     *
     * @param filename A char array specifying the location of a PNG image.
     *
     * Creates a @c PNGImage from the PNG image located at @c filename.
     */
    PNGImage(const char* filename);

              
    /**
     * @brief %PNGImage custom constructor.
     *
     * @param width  The width of the image.
     * @param height The height of the image.
     * @param nChl   The number of channels of the image.
     *
     * Creates an empty PNG image with the specified dimensions
     * and initializes the data to zero.
    */
    PNGImage(int width, int height, int nChl);

    /**
     * @brief %PNGImage destructor.
     */
    virtual ~PNGImage();

    /**
     * @brief %PNGImage assignment operator.
     *
     * @param clone The @c PNGImage to be cloned.
     * @return This object.
     */
    virtual PNGImage& operator=(const PNGImage& clone);
};

/**
 * @c BSIImage adds BSI-specific load and functionality to its base class
 * @c Image. It also provides information about the number of segments and the
 * maximum label number stored.
 */
class BSIImage : public Image<BSI_IMAGE_DATATYPE>
{
  public:

    // functions for read-only access to data members ------
    /**
     * @brief Get the number of segments of the images.
     * @return The number of segemtns of the images
     */
    inline unsigned long nSeg() const
    {return nSeg_;};

    // class-specific functions ------

    /**
     * @brief Load the BSI image file "filename".
     *
     * @param filename A char array specifying the location of a BSI image.
     * @param modNum A constant to be added to all labels.
     * @return 0 if successful.
     *
     * Replaces this BSI image with the BSI image located at @c filename and
     * optionally adds the constant @c modNum to all labels in the BSI image to
     * avoid duplicate labels in a sequence of BSI images.
     */
    int load(const char* filename, unsigned long modNum = 0);

    // standard functions ------

    /**
     * @brief %BSIImage empty constructor.
     *
     * Creates an empty image with width=height=channels=0.
     */
    BSIImage();
    /**
     * @brief %BSIImage copy constructor.
     *
     * @param clone The @c BSIImage to be cloned.
     */
    BSIImage(const BSIImage& clone);
    /**
     * @brief A %BSIImage can be constructed using a BSI image stored on disk.
     *
     * @param filename A char array specifying the location of a BSI image.
     * @param modNum A constant to be added to all labels.
     *
     * Creates a @c BSIImage from the BSI image located at @c filename and
     * optionally adds the constant @c modNum to all labels in the BSI image to
     * avoid duplicate labels in a sequence of BSI images.
     */
    BSIImage(const char* filename, unsigned long& modNum);

    /**
     * @brief &BSIImage destructor.
     */
    virtual ~BSIImage();

    /**
     * @brief %BSIImage assignment operator.
     *
     * @param clone The @c BSIImage to be cloned.
     * @return This object.
     */
    virtual BSIImage& operator=(const BSIImage& clone);

  protected:

    unsigned long nSeg_; /**< The number of segments in this BSI image*/
};

/**
 * @c Sequence is the base class for @c PNGSequence and @c BSISequence.
 * It stores pointers to a sequence images.
 * @c Sequence is a small wrapper around a STL deque containing pointers to all
 * @c Sequence s in the sequence, providing additional information, such as
 * width, height and length, and special sequence manipulation functions.
 */
template <class ELEMENT_TYPE = unsigned char>
class Sequence
{
  public:

    // functions for read & write access to data members ------

    virtual Image<ELEMENT_TYPE>* operator[](unsigned long);

    // functions for read-only access to data members ------

    virtual const Image<ELEMENT_TYPE>* operator[](unsigned long) const;

    /**
     * @brief Get the with of the images.
     * @return The width of the images
     */
    inline unsigned int width() const
    {return width_;};
    /**
     * @brief Get the height of the images.
     * @return The height of the images
     */
    inline unsigned int height() const
    {return height_;};
    /**
     * @brief Get the number of channels of the images.
     * @return The number of channles of the images
     */
    inline unsigned int nChl() const
    {return nChl_;};
    /**
     * @brief Get the length of the sequence.
     * @return The length of the sequence
     */
    inline unsigned int length() const
    {return frames_.size();};

    // class-specific functions ------

    /**
     * @brief Push a frame onto the sequence and pop the first frame.
     *
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image onto the sequence and removes the first
     * frame of the sequence. The length of the sequence remains unchanged.
     * Only images of the same dimension as all other images in the sequence
     * are accepted.
     */
    virtual int pushThrough(const Image<ELEMENT_TYPE>& frame);
    /**
     * @brief Push a frame to the back of the sequence.
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image to the back of the sequence. The length of the
     * sequence is increased by one.
     * Only images of the same dimension as all other images in the sequence
     * are accepted.
     */
    virtual int pushFrame(const Image<ELEMENT_TYPE>& frame);
    /**
     * @brief Pop the last frame of the sequence.
     *
     * Deletes the last frame and removes it from the sequence. This does
     * nothing if the sequence is empty.
     */
    virtual void popFrame();
    /**
     * @brief Remove all frames from the sequence.
     * @return 0 if successful.
     *
     * Deletes all frames in the sequence and sets width, height and number of
     * channels to zero. The length of the sequence is zero.
     */
    virtual int clear();

    // standard functions ------

    /**
     * @brief %Sequence empty constructor
     *
     * Creates an empty @c Sequence with=height=channels=0.
     */
    Sequence();
    /**
     * @brief %Sequence copy constructor.
     *
     * @param clone The @c Sequence to be cloned.
     */
    Sequence(const Sequence& clone);

    /**
     * @brief %Sequence destructor.
     */
    virtual ~Sequence();

    /**
     * @brief %Sequence assignment operator.
     *
     * @param clone The @c Sequence to be cloned.
     * @return This object.
     */
    virtual Sequence& operator=(const Sequence& clone);


  protected:

    std::deque<Image<ELEMENT_TYPE>* > frames_; /**< Contains pointers to all
        images in this sequence. All images are owned by @c Sequence*/
    unsigned int width_;/**< The width of all images in this sequence.*/
    unsigned int height_;/**< The height of all images in this sequence.*/
    unsigned int nChl_; /**< The number of
        channels of all images in this sequence.*/

    /**
     * @brief Checks if a frame can be added to this sequence.
     *
     * @param frame The frame to be checked.
     * @return @c true if the frame has the right dimension, @c false otherwise.
     *
     * Every frame in a sequence has to be of the same dimension. This function
     * checks if width, height and number of channels match.
     */
    bool validateFrame(const Image<ELEMENT_TYPE>& frame);
};

/**
 * @c PNGSequence adds PNG-specific save functionality to its base class
 * @c Sequence. Only images of type @c PNGImage are allows in a @c PNGSequence.
 */
class PNGSequence : public Sequence<PNG_IMAGE_DATATYPE>
{
  public:

    // functions for read & write access to data members ------

    PNGImage* operator[](unsigned long);

    // functions for read-only access to data members ------

    const PNGImage* operator[](unsigned long) const;

    // class-specific functions ------

    /**
     * @brief Push a frame onto the sequence and pop the first frame.
     *
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image onto the sequence and removes the first
     * frame of the sequence. The length of the sequence remains unchanged.
     * Only PNG images of the same dimension as all other images in the sequence
     * are accepted. @c frame is cast to a @c PNGImage.
     */
    int pushThrough(const Image<PNG_IMAGE_DATATYPE>& frame);
    /**
     * @brief Push a frame onto the sequence and pop the first frame.
     *
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image onto the sequence and removes the first
     * frame of the sequence. The length of the sequence remains unchanged.
     * Only PNG images of the same dimension as all other images in the sequence
     * are accepted.
     */
    int pushThrough(const PNGImage& frame);
    /**
     * @brief Push a frame to the back of the sequence.
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image to the back of the sequence. The length of the
     * sequence is increased by one.
     * Only PNG images of the same dimension as all other images in the sequence
     * are accepted. @c frame is cast to a @c PNGImage.
     */
    int pushFrame(const Image<PNG_IMAGE_DATATYPE>& frame);
    /**
     * @brief Push a frame to the back of the sequence.
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image to the back of the sequence. The length of the
     * sequence is increased by one.
     * Only PNG images of the same dimension as all other images in the sequence
     * are accepted.
     */
    int pushFrame(const PNGImage& frame);
    /**
     * @brief Pop the last frame of the sequence.
     *
     * Deletes the last frame and removes it from the sequence. This does
     * nothing if the sequence is empty.
     */
    void popFrame();
    /**
     * @brief Save all PNG images in this sequence.
     *
     * @param filename The base filename for all images.
     * @return 0 if successful.
     *
     * The sequence number of each frame is appended to the filename and the
     * frame is saved.
     */
    int save(const char* filename) const;

    // standard functions ------

    /**
     * @brief %PNGSequence empty constructor
     *
     * Creates an empty @c PNGSequence with=height=channels=0.
     */
    PNGSequence();
    /**
     * @brief %PNGSequence copy constructor.
     *
     * @param clone The @c PNGSequence to be cloned.
     */
    PNGSequence(const PNGSequence& clone);

    /**
     * @brief %PNGSequence destructor.
     */
    virtual ~PNGSequence();

    /**
     * @brief %PNGSequence assignment operator.
     *
     * @param clone The @c PNGSequence to be cloned.
     * @return This object.
     */
    virtual PNGSequence& operator=(const PNGSequence& clone);
};

/**
 * A @c BSISequence keeps track of the total number of segments in the sequence.
 * Only images of type @c BSIImage are allowed in a @c BSISequence.
 */
class BSISequence : public Sequence<BSI_IMAGE_DATATYPE>
{
  public:

    // functions for read & write access to data members ------

    BSIImage* operator[](unsigned long);

    // functions for read-only access to data members ------

    const BSIImage* operator[](unsigned long) const;

    /**
     * @brief Get the number of segments in the sequence
     * @return The number of segments in the seuqence
     */
    inline unsigned long nSeg() const
    {return nSeg_;};

    // class-specific functions ------

    /**
     * @brief Push a frame onto the sequence and pop the first frame.
     *
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image onto the sequence and removes the first
     * frame of the sequence. The length of the sequence remains unchanged.
     * Only BSI images of the same dimension as all other images in the sequence
     * are accepted. @c frame is cast to a @c BSIImage.
     */
    int pushThrough(const Image<BSI_IMAGE_DATATYPE>& frame);
    /**
     * @brief Push a frame onto the sequence and pop the first frame.
     *
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image onto the sequence and removes the first
     * frame of the sequence. The length of the sequence remains unchanged.
     * Only BSI images of the same dimension as all other images in the sequence
     * are accepted.
     */
    int pushThrough(const BSIImage& frame);
    /**
     * @brief Push a frame to the back of the sequence.
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image to the back of the sequence. The length of the
     * sequence is increased by one.
     * Only BSI images of the same dimension as all other images in the sequence
     * are accepted. @c frame is cast to a @c BSIImage.
     */
    int pushFrame(const Image<BSI_IMAGE_DATATYPE>& frame);
    /**
     * @brief Push a frame to the back of the sequence.
     * @param frame The @c Image to be pushed onto the sequence.
     * @return 0 if successful.
     *
     * Pushes a copy of @c Image to the back of the sequence. The length of the
     * sequence is increased by one.
     * Only BSI images of the same dimension as all other images in the sequence
     * are accepted.
     */
    int pushFrame(const BSIImage& frame);
    /**
     * @brief Pop the last frame of the sequence.
     *
     * Deletes the last frame and removes it from the sequence. This does
     * nothing if the sequence is empty.
     */
    void popFrame();
    /**
     * @brief Remove all frames from the sequence.
     * @return 0 if successful.
     *
     * Deletes all frames in the sequence and sets width, height, number of
     * channels and number of segments to zero. The length of the sequence is
     * zero.
     */
    int clear();

    // standard functions ------

    /**
     * @brief %BSISequence empty constructor
     *
     * Creates an empty @c BSISequence with=height=channels=segments=0.
     */
    BSISequence();
    /**
     * @brief %BSISequence copy constructor.
     *
     * @param clone The @c BSISequence to be cloned.
     */
    BSISequence(const BSISequence& clone);

    /**
     * @brief BSISequence destructor.
     */
    virtual ~BSISequence();

    /**
     * @brief %BSISequence assignment operator.
     *
     * @param clone The @c BSISequence to be cloned.
     * @return This object.
     */
    virtual BSISequence& operator=(const BSISequence& clone);

  protected:

    unsigned long nSeg_; /**< The total number of segments in the sequence.*/
};


/* template implementations ------------------------------------------------- */

template <class ELEMENT_TYPE>
Image<ELEMENT_TYPE>::Image()
  : imageData_(), width_(0), height_(0), nChl_(0)
{
  return;
}

template <class ELEMENT_TYPE>
Image<ELEMENT_TYPE>::Image(const Image& clone)
  : imageData_(), width_(0), height_(0), nChl_(0)
{
  *this = clone;
}

template <class ELEMENT_TYPE>
Image<ELEMENT_TYPE>::Image(int width, int height, int nChl)
  : imageData_(), width_(width), height_(height), nChl_(nChl)
{
  imageData_.resize(width_*height_*nChl_);
  return;
}

template <class ELEMENT_TYPE>
Image<ELEMENT_TYPE>::~Image()
{
  return;
}

template <class ELEMENT_TYPE>
Image<ELEMENT_TYPE>& Image<ELEMENT_TYPE>::operator=(const Image& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // clear old data
  if (!imageData_.empty()) imageData_.clear();

  // copy image info from clone
  this->width_ = clone.width_;
  this->height_ = clone.height_;
  this->nChl_ = clone.nChl_;

  // copy image data from the clone
  imageData_ = clone.imageData_;

  return *this;
}

template <class ELEMENT_TYPE>
bool Sequence<ELEMENT_TYPE>::validateFrame(const Image<ELEMENT_TYPE>& frame)
{
  if (frames_.empty()) {
    width_  = frame.width();
    height_ = frame.height();
    nChl_   = frame.nChl();
  }
  else {
    if (width_  != frame.width() ||
        height_ != frame.height() ||
        nChl_   != frame.nChl()) {
      return false; // dimensions of all frames must agree
    }
  }

  return true;
}

template <class ELEMENT_TYPE>
int Sequence<ELEMENT_TYPE>::pushThrough(const Image<ELEMENT_TYPE>& frame)
{
  if (!validateFrame(frame)) {
    throw ncutException("frame does not fit into the sequence (wrong format?)");
  }

  if (frames_.empty()) {
    pushFrame(frame);
    return 0;
  }

  delete frames_.front();
  frames_.pop_front();

  Image<ELEMENT_TYPE> *clone = new Image<ELEMENT_TYPE>(frame);

  frames_.push_back(clone);

  return 0;
}

template <class ELEMENT_TYPE>
int Sequence<ELEMENT_TYPE>::pushFrame(const Image<ELEMENT_TYPE>& frame)
{
  if (!validateFrame(frame)) {
    throw ncutException("frame does not fit into the sequence (wrong format?)");
  }

  Image<ELEMENT_TYPE> *clone = new Image<ELEMENT_TYPE>(frame);

  frames_.push_back(clone);

  return 0;
}

template <class ELEMENT_TYPE>
void Sequence<ELEMENT_TYPE>::popFrame()
{
  if (!frames_.empty()) {
    delete frames_.back();
    frames_.pop_back();
  }
}

template <class ELEMENT_TYPE>
int Sequence<ELEMENT_TYPE>::clear()
{
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!frames_.empty()) frames_.clear();

  width_  = 0;
  height_ = 0;
  nChl_   = 0;

  return 0;
}

template <class ELEMENT_TYPE>
Sequence<ELEMENT_TYPE>::Sequence()
  : frames_(), width_(0), height_(0), nChl_(0)
{
  return;
}

template <class ELEMENT_TYPE>
Sequence<ELEMENT_TYPE>::Sequence(const Sequence& clone)
  : frames_(), width_(0), height_(0), nChl_(0)
{
  *this = clone;
}

template <class ELEMENT_TYPE>
Sequence<ELEMENT_TYPE>::~Sequence()
{
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!frames_.empty()) frames_.clear();
}

template <class ELEMENT_TYPE>
Sequence<ELEMENT_TYPE>& Sequence<ELEMENT_TYPE>::operator=(const Sequence& clone)
{
  // check for self-assignment
  if (this == &clone) return *this;

  // free old memory
  for (unsigned int i=0; i<frames_.size(); i++) {
    delete frames_[i];
    frames_[i] = NULL;
  }
  if (!frames_.empty()) frames_.clear();

  // copy sequence info from clone
  this->width_  = clone.width_;
  this->height_ = clone.height_;
  this->nChl_   = clone.nChl_;

  // initialize the local structure
  frames_.resize(clone.frames_.size());

  // copy the sequence data from the clone
  for (unsigned int i=0; i<frames_.size(); i++) {
    if (clone.frames_[i] != NULL) {
      this->frames_[i] = new Image<ELEMENT_TYPE>(*(clone.frames_[i]));
    }
    else {
      this->frames_[i] = NULL;
    }
  }

  return *this;
}

template <class ELEMENT_TYPE>
Image<ELEMENT_TYPE>* Sequence<ELEMENT_TYPE>::
  operator[](unsigned long pos)
{
  assert(pos < frames_.size()); // alle asserts kommen am ende weg
  if(pos >= frames_.size()) pos = frames_.size()-1;

  return frames_[pos];
}

template <class ELEMENT_TYPE>
const Image<ELEMENT_TYPE>* Sequence<ELEMENT_TYPE>::
  operator[](unsigned long pos) const
{
  assert(pos < frames_.size()); // alle asserts kommen am ende weg
  if(pos >= frames_.size()) pos = frames_.size()-1;

  return frames_[pos];
}

} // namespace ncut

#endif // NCUT_IMAGE_H
