/***************************************************************************
 *   Copyright (C) 2005 by Paul Guerrero   *
 *   paul@Glendin   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef NCUTMEMORY_H
#define NCUTMEMORY_H 1

#include <memory>

/**
  *  A wrapper class to provide auto_array with reference semantics.  For
  *  example, an auto_array can be assigned (or constructed from) the result of
  *  a function which returns an auto_array by value.
  *
  *  All the auto_array_ref stuff should happen behind the scenes.
  */
template<typename _Tp1>
struct auto_array_ref
{
  _Tp1* _M_ptr;

  explicit
  auto_array_ref(_Tp1* __p): _M_ptr(__p) { }
};

/**
 *  @brief  A simple smart array providing strict ownership semantics.
 *
 *  An @c auto_array owns the array it holds a pointer to.  Copying an
 *  @c auto_array copies the pointer and transfers ownership to the
 *  destination.
 *  If more than one @c auto_array owns the same array at the same time the
 *  behavior of the program is undefined.
 *
 *  The uses of @c auto_array include providing temporary exception-safety for
 *  dynamically allocated memory, passing ownership of dynamically allocated
 *  memory to a function, and returning dynamically allocated memory from a
 *  function.  @c auto_array does not meet the CopyConstructible and
 *  Assignable requirements for Standard Library
 *  <a href="tables.html#65">container</a>  elements and thus instantiating a
 *  Standard Library container with an @c auto_array results in undefined
 *  behavior.
 *
 */
template<typename _Tp>
class auto_array
{
  private:

    _Tp* _M_ptr;

  public:

    /// The pointed-to type.
    typedef _Tp element_type;

    /**
      *  @brief  An %auto_array is usually constructed from a raw array.
      *  @param  __p  A pointer (defaults to NULL).
      *
      *  This object now @e owns the array pointed to by @a p.
      */
    explicit
    auto_array(element_type* __p = 0) throw() : _M_ptr(__p) { }

    /**
      *  @brief  An %auto_array can be constructed from another %auto_array.
      *  @param  __a  Another %auto_array of the same type.
      *
      *  This object now @e owns the object previously owned by @a a,
      *  which has given up ownsership.
      */
    auto_array(auto_array& __a) throw() : _M_ptr(__a.release()) { }

    /**
      *  @brief  An %auto_array can be constructed from another %auto_array.
      *  @param  __a  Another %auto_array of a different but related type.
      *
      *  A pointer-to-Tp1 must be convertible to a pointer-to-Tp/element_type.
      *
      *  This object now @e owns the object previously owned by @a a,
      *  which has given up ownsership.
      */
    template<typename _Tp1>
    auto_array(auto_array<_Tp1>& __a) throw() : _M_ptr(__a.release()) { }

    /**
      *  @brief  %auto_array assignment operator.
      *  @param  __a  Another %auto_array of the same type.
      *
      *  This object now @e owns the object previously owned by @a a,
      *  which has given up ownsership.  The object that this one @e
      *  used to own and track has been deleted.
      */
    auto_array& operator=(auto_array& __a) throw()
    {
      reset(__a.release());
      return *this;
    }

    /**
      *  @brief  %auto_array assignment operator.
      *  @param  __a  Another %auto_array of a different but related type.
      *
      *  A pointer-to-Tp1 must be convertible to a pointer-to-Tp/element_type.
      *
      *  This object now @e owns the object previously owned by @a a,
      *  which has given up ownsership.  The object that this one @e
      *  used to own and track has been deleted.
      */
    template<typename _Tp1>
    auto_array& operator=(auto_array<_Tp1>& __a) throw()
    {
      reset(__a.release());
      return *this;
    }

    /**
      *  When the %auto_array goes out of scope, the object it owns is deleted.
      *  If it no longer owns anything (i.e., @c get() is @c NULL), then this
      *  has no effect.
      *
      *  The C++ standard says there is supposed to be an empty throw
      *  specification here, but omitting it is standard conforming.  Its
      *  presence can be detected only if _Tp::~_Tp() throws, but this is
      *  prohibited.  [17.4.3.6]/2
      */
    ~auto_array() { delete[] _M_ptr; }

    /**
      *  @brief  Smart array element access.
      *
      *  If this %auto_array no longer owns anything, then this
      *  operation will crash.  (For a smart pointer, "no longer owns
      *  anything" is the same as being a null pointer, and you know
      *  what happens when you dereference one of those...)
      */
    element_type& operator[](unsigned int i) const throw() { return _M_ptr[i]; }

    /**
      *  @brief  Smart array dereferencing.
      *
      *  If this %auto_array no longer owns anything, then this
      *  operation will crash.  (For a smart pointer, "no longer owns
      *  anything" is the same as being a null pointer, and you know
      *  what happens when you dereference one of those...)
      */
    element_type& operator*() const throw() { return *_M_ptr; }

    /**
      *  @brief  Smart array dereferencing.
      *
      *  This returns the pointer itself, which the language then will
      *  automatically cause to be dereferenced.
      */
    element_type* operator->() const throw() { return _M_ptr; }

    /**
      *  @brief  Bypassing the smart array.
      *  @return  The raw pointer being managed.
      *
      *  You can get a copy of the pointer that this object owns, for
      *  situations such as passing to a function which only accepts a raw
      *  pointer.
      *
      *  @note  This %auto_array still owns the memory.
      */
    element_type* get() const throw() { return _M_ptr; }

    /**
      *  @brief  Bypassing the smart array.
      *  @return  The raw pointer being managed.
      *
      *  You can get a copy of the pointer that this object owns, for
      *  situations such as passing to a function which only accepts a raw
      *  pointer.
      *
      *  @note  This %auto_array no longer owns the memory.  When this object
      *  goes out of scope, nothing will happen.
      */
    element_type* release() throw() {
      element_type* __tmp = _M_ptr;
      _M_ptr = 0;
      return __tmp;
    }

    /**
      *  @brief  Forcibly deletes the managed array.
      *  @param  __p  A pointer (defaults to NULL).
      *
      *  This object now @e owns the object pointed to by @a p.  The previous
      *  object has been deleted.
      */
    void reset(element_type* __p = 0) throw() {
      if (__p != _M_ptr) {
        delete[] _M_ptr;
        _M_ptr = __p;
      }
    }

    /** @{
      *  @brief  Automatic conversions
      *
      *  These operations convert an %auto_array into and from an auto_arary_ref
      *  automatically as needed.  This allows constructs such as
      *  @code
      *    auto_array<Derived>  func_returning_auto_array(.....);
      *    ...
      *    auto_array<Base> ptr = func_returning_auto_array(.....);
      *  @endcode
      */
    auto_array(auto_array_ref<element_type> __ref) throw()
      : _M_ptr(__ref._M_ptr) { }

    auto_array& operator=(auto_array_ref<element_type> __ref) throw() {
      if (__ref._M_ptr != this->get()) {
        delete[] _M_ptr;
        _M_ptr = __ref._M_ptr;
      }
      return *this;
    }

    template<typename _Tp1>
    operator auto_array_ref<_Tp1>() throw()
    { return auto_array_ref<_Tp1>(this->release()); }

    template<typename _Tp1>
    operator auto_array<_Tp1>() throw()
    { return auto_array<_Tp1>(this->release()); }
    /** @}  */
};

/**
  *  A wrapper class to provide greedy_ptr with reference semantics.  For
  *  example, an greedy_ptr can be assigned (or constructed from) the result of
  *  a function which returns an greedy_ptr by value.
  *
  *  All the greedy_ptr_ref stuff should happen behind the scenes.
  */
template<typename _Tp1>
struct greedy_ptr_ref
{
  _Tp1* _M_ptr;

  explicit
  greedy_ptr_ref(_Tp1* __p): _M_ptr(__p) { }
};


/**
  *  @brief  A simple smart pointer providing strict ownership semantics.
  *
  *  A @c greedy_ptr only owns the object it holds a pointer to, if it was
  *  constructed from a raw pointer (i.e. if it was the first greedy_ptr to
  *  manage the pointed-to object). Copying a @c greedy_ptr copies the
  *  pointer but does not transfer ownership to the destination, the source
  *  @c greedy_ptr retains ownership.
  *  If more than one @c greedy_ptr owns the same object at the same time the
  *  behavior of the program is undefined.
  */
template<typename _Tp>
class greedy_ptr
{
  private:
    _Tp* _M_ptr;
    bool _owns;

  public:
    /// The pointed-to type.
    typedef _Tp element_type;

    /**
      *  @brief  A %greedy_ptr is usually constructed from a raw pointer.
      *  @param  __p  A pointer (defaults to NULL).
      *
      *  This object now @e owns the object pointed to by @a p.
      */
    explicit
    greedy_ptr(element_type* __p = 0) throw()
      : _M_ptr(__p), _owns((__p != 0) ? true:false)
    { }

    /**
      *  @brief  A %greedy_ptr can be constructed from a reference.
      *  @param  __p  A reference (defaults to NULL).
      *
      *  This object does @e not own the object @a p.
      */
    explicit
    greedy_ptr(element_type& __p) throw()
      : _M_ptr(&__p), _owns(false)
    { }

    /**
      *  @brief  A %greedy_ptr can be constructed from another %greedy_ptr.
      *  @param  __a  Another %greedy_ptr of the same type.
      *
      *  This object does @e not own the object owned by @a a,
      *  which still holds ownership.
      */
    greedy_ptr(const greedy_ptr& __a) throw()
      : _M_ptr(__a.get()), _owns(false)
    { }

    /**
      *  @brief  A %greedy_ptr can be constructed from an %auto_ptr.
      *  @param  __a  An %auto_ptr of a different but related type.
      *
      *  A pointer-to-Tp1 must be convertible to a pointer-to-Tp/element_type.
      *
      *  This object now @e owns the object previously owned by @a a,
      *  which has given up ownership.
      */
    template<typename _Tp1>
    greedy_ptr(std::auto_ptr<_Tp1>& __a) throw()
      : _M_ptr(__a.release()), _owns(true)
    { }

    /**
      *  @brief  A %greedy_ptr can be constructed from an %auto_ptr.
      *  @param  __a  An %auto_ptr of the same type.
      *
      *  This object now @e owns the object previously owned by @a a,
      *  which has given up ownership.
      */
    greedy_ptr(std::auto_ptr<element_type>& __a) throw()
      : _M_ptr(__a.release()), _owns(true)
    { }

    /**
      *  @brief  A %greedy_ptr can be constructed from another %greedy_ptr.
      *  @param  __a  Another %greedy_ptr of a different but related type.
      *
      *  A pointer-to-Tp1 must be convertible to a pointer-to-Tp/element_type.
      *
      *  This object does @e not own the object owned by @a a,
      *  which still holds ownership.
      */
    template<typename _Tp1>
    greedy_ptr(const greedy_ptr<_Tp1>& __a) throw()
      :_M_ptr(__a.get()), _owns(false)
    { }

    /**
      *  @brief  %greedy_ptr assignment operator.
      *  @param  __a  Another %greedy_ptr of the same type.
      *
      *  This object does @e not own the object owned by @a a,
      *  which still holds ownership.
      *  If this greedy_ptr @e used to own an object, that object has been
      *  deleted.
      */
    greedy_ptr&
    operator=(const greedy_ptr& __a) throw()
    {
      reset(__a.get());
      _owns = false;
      return *this;
    }

    /**
      *  @brief  %greedy_ptr assignment operator.
      *  @param  __a  Another %greedy_ptr of a different but related type.
      *
      *  A pointer-to-Tp1 must be convertible to a pointer-to-Tp/element_type.
      *
      *  This object does @e not own the object owned by @a a,
      *  which still holds ownership.
      *  If this greedy_ptr @e used to own an object, that object has been
      *  deleted.
      */
    template<typename _Tp1>
    greedy_ptr&
    operator=(const greedy_ptr<_Tp1>& __a) throw()
    {
      reset(__a.get());
      _owns = false;
      return *this;
    }

     /**
      *  @brief  %greedy_ptr assignment operator.
      *  @param  __p  A pointer.
      *
      *  A convenience function, does exactly the same thing as @c reset(__p).
      */
    greedy_ptr&
    operator=(element_type* __p) throw()
    {
      reset(__p);
      return *this;
    }

    /**
      *  @brief  %greedy_ptr assignment operator.
      *  @param  __p  A reference.
      *
      *  A convenience function, does exactly the same thing as @c reset(__p).
      */
    greedy_ptr&
    operator=(element_type& __p) throw()
    {
      reset(__p);
      return *this;
    }

    /**
      *  When the %greedy_ptr goes out of scope, the object it owns is deleted.
      *  If it no longer owns anything then this has no effect.
      *
      *  The C++ standard says there is supposed to be an empty throw
      *  specification here, but omitting it is standard conforming.  Its
      *  presence can be detected only if _Tp::~_Tp() throws, but this is
      *  prohibited.  [17.4.3.6]/2
      */
    ~greedy_ptr() { if (_owns) delete _M_ptr; }

    /**
      *  @brief  Smart pointer dereferencing.
      *
      *  If this %greedy_ptr is a null pointer, then this
      *  operation will crash.
      */
    element_type&
    operator*() const throw() { return *_M_ptr; }

    /**
      *  @brief  Smart pointer dereferencing.
      *
      *  This returns the pointer itself, which the language then will
      *  automatically cause to be dereferenced.
      */
    element_type*
    operator->() const throw() { return _M_ptr; }

    /**
      *  @brief  Bypassing the smart pointer.
      *  @return  The raw pointer being managed.
      *
      *  You can get a copy of the pointer that this object owns, for
      *  situations such as passing to a function which only accepts a raw
      *  pointer.
      *
      *  @note  If this %greedy_ptr was owning the memory, it still does.
      */
    element_type*
    get() const throw() { return _M_ptr; }

    /**
      *  @brief  Getting ownership status.
      *  @return  @c true if this %greedy_ptr owns the object being managed,
      *           @c false otherwise.
      *
      *  You can query the %greedy_ptr for ownership of the object being
      *  managed. @c NULL pointers are never owned.
      */
    bool
    owns() const throw() {return _owns;}

    /**
      *  @brief  Bypassing the smart pointer.
      *  @return  The raw pointer being managed.
      *
      *  You can get a copy of the pointer that this object owns, for
      *  situations such as passing to a function which only accepts a raw
      *  pointer.
      *
      *  @note  This %greedy_ptr no longer owns the memory.  When this object
      *  goes out of scope, nothing will happen.
      */
    element_type*
    release() throw()
    {
      _owns = false;
      return _M_ptr;
    }

    /**
      *  @brief  Forcibly deletes the managed object.
      *  @param  __p  A pointer (defaults to NULL).
      *
      *  This object now @e owns the object pointed to by @a p.  The previous
      *  object has been deleted.
      */
    void
    reset(element_type* __p = 0) throw()
    {
      if (__p != _M_ptr) {
        if (_owns) delete _M_ptr;
        _M_ptr = __p;
        if (__p != 0) _owns = true;
        else _owns = false;
      }
    }

    /**
      *  @brief  Forcibly deletes the managed object.
      *  @param  __p  A reference.
      *
      *  This object now holds a pointer to @a p , but does @e not own it.
      *  The previously held object has been deleted.
      */
    void
    reset(element_type& __p = 0) throw()
    {
      if (&__p != _M_ptr) {
        if (_owns) delete _M_ptr;
        _M_ptr = &__p;
        _owns = false;
      }
    }

    /** @{
      *  @brief  Automatic conversions
      *
      *  These operations convert an %greedy_ptr into and from an greedy_ptr_ref
      *  automatically as needed.  This allows constructs such as
      *  @code
      *    greedy_ptr<Derived>  func_returning_greedy_ptr(.....);
      *    ...
      *    greedy_ptr<Base> ptr = func_returning_greedy_ptr(.....);
      *  @endcode
      */
    greedy_ptr(greedy_ptr_ref<element_type> __ref) throw()
      : _M_ptr(__ref._M_ptr), _owns(false) { }

    greedy_ptr&
    operator=(greedy_ptr_ref<element_type> __ref) throw()
    {
      if (__ref._M_ptr != this->get()) {
        if (_owns) delete _M_ptr;
        _M_ptr = __ref._M_ptr;
        _owns = false;
      }
      return *this;
    }

    template<typename _Tp1>
    operator greedy_ptr_ref<_Tp1>() throw()
    { return greedy_ptr_ref<_Tp1>(this->get()); }

    template<typename _Tp1>
    operator greedy_ptr<_Tp1>() throw()
    { return greedy_ptr<_Tp1>(*this); }
    /** @}  */
};

#endif // NCUTMEMORY_H
