#ifndef NCUT_VIS_H
#define NCUT_VIS_H 1

#include "ncutCore.h"
#include <math.h>
#include <time.h>
#include <iostream>

namespace ncut
{

struct PNG_color_t {
  PNG_IMAGE_DATATYPE red;
  PNG_IMAGE_DATATYPE green;
  PNG_IMAGE_DATATYPE blue;
};

/**
 * @brief Visualization of a segmenation
 * This class is the abstract class for visualization
*/
template <class IMG_ELM_TYPE, class MSK_ELM_TYPE, class VIS_ELM_TYPE>
class Vis
{
  public:
    
    /**
     * @brief Get the segmentation used by this visualization.
     * @return The segmentation used by this visualization.
     */
    inline const Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>* seg() const
    {return seg_;}

    
    /**
     * Creates a visualization for the current frame window of the sequence
     * @param resultSequence the sequence where the visualization should be saved
     * @param topFrame current Framenumber
     */
    virtual void create(Sequence<VIS_ELM_TYPE>& resultSequence, int topFrame = -1);

    /**
     * Vis is constructed by passing a reference to a Segmentation
     * @param  segmentation reference to a segmenation which should be visualized 
     */
    Vis(const Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>* segmentation);
//     Vis(const Vis& clone);


    /**
     * Vis destructor
     */
    virtual ~Vis();

  protected:

    const Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>* seg_; /**< segmentation which
                                                           * should be visualized*/

//     Vis& operator=(const Vis&);
};

/**
 * @brief A visualization of a segmentation using png images
 * Creates a visualization of a given segmentation using png images
 */
template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
class PNGVis : public Vis<IMG_ELM_TYPE, MSK_ELM_TYPE, PNG_IMAGE_DATATYPE>
{
  public:

    inline const std::map<MSK_ELM_TYPE,std::vector<int> >& rememberedPosX() const
    { return rememberedPosX_; }
    inline const std::map<MSK_ELM_TYPE,std::vector<int> >& rememberedPosY() const
    { return rememberedPosY_; }
    inline const std::map<MSK_ELM_TYPE,double>& segmentSumArea() const
    { return segmentSumArea_; };
    int firstFrame(const MSK_ELM_TYPE& segId) const;

    /**
     * Creates a visualization for the current frame window of the sequence
     * @param resultSequence the sequence where the visualization should be saved
     * @param topFrame current Framenumber
    */
    virtual void create(PNGSequence& resultSequence, int topFrame = -1);

    /**
     * Draws a line from (x1,y1) to (x2,y2) in the given image
     * @param x1 x-coordinate of the first point
     * @param y1 y-coordinate of the first point
     * @param x2 x-coordinate of the second point
     * @param y2 y-coordinate of the second point
     * @param image image where line should be drawn
     * @return 0 if successful
     */
    virtual int drawLine(int x1, int y1, int x2, int y2, PNGImage *image);

    /**
     * PNGVis is constructed by passing a reference to a Segmentation
     * @param  segmentation reference to a segmenation which should be visualized
    */
    PNGVis(const Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>* segmentation);

    
    /**
     * PNGVis Destructor
     */
    virtual ~PNGVis();


  protected:

    int baseFrame_; /**< framenumber of first segmented frame */

    std::vector<PNG_color_t> colors_; /**< color of final segment*/
    std::map<MSK_ELM_TYPE, unsigned int> colorMap_; /**< list of possible colors */

    std::map<MSK_ELM_TYPE,std::vector<int> > rememberedPosX_; /**< history of
         midpoint positions of a final segment (x-coordinate) */
    std::map<MSK_ELM_TYPE,std::vector<int> > rememberedPosY_;/**< history of
    midpoint positions of a final segment (y-coordinate) */
    std::map<MSK_ELM_TYPE,double > segmentSumArea_; /**< total sum of an final
                       * segment in all frames */

    std::map<MSK_ELM_TYPE, int> firstFrame_; /**< framenumber of first occurence
                                              * of a segment */

  private:

    virtual void mode_segmentedObjects(PNGImage* image, int cmp,
                                       int pxl, int frm);
    virtual void mode_original(PNGImage* image, int cmp, int frm);

    virtual unsigned long calculateElementNr(int x, int y,
                                             const PNGImage* image);

};


/* template implementations ---------------------------------------------- */

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE, class VIS_ELM_TYPE>
void Vis<IMG_ELM_TYPE, MSK_ELM_TYPE, VIS_ELM_TYPE>::
    create(Sequence<VIS_ELM_TYPE>& seq, int topFrame)
{
  // dummy function

  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE, class VIS_ELM_TYPE>
Vis<IMG_ELM_TYPE, MSK_ELM_TYPE, VIS_ELM_TYPE>::
    Vis(const Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>* _seg)
  : seg_(_seg)
{
  if (this->seg_ == NULL)
    throw ncutException("invalid segmentation (NULL)");
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE, class VIS_ELM_TYPE>
Vis<IMG_ELM_TYPE, MSK_ELM_TYPE, VIS_ELM_TYPE>::~Vis()
{
  return;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
int PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    firstFrame(const MSK_ELM_TYPE& segId) const
{
  typename std::map<MSK_ELM_TYPE, int>::const_iterator firstFrame_p;

  firstFrame_p = firstFrame_.find(segId);
  if (firstFrame_p == firstFrame_.end())
    throw ncutException("segment not found");

  return firstFrame_p->second;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::create(PNGSequence& seq, int topFrame)
{
#ifdef debugout
  clock_t timerStart, timerEnd;

  std::cout << "building the visualization..." << std::flush;
  timerStart = clock();
#endif

  // update the frame number
  if (topFrame == -1) baseFrame_++;
  else                baseFrame_ = topFrame - this->seg()->nFrm() + 1;
  if (baseFrame_ < 0)
    throw ncutException(
        "invalid frame number, must be >= number of frames in frame window" );

  // create new empty PNG image
  PNGImage *segmentedImage = new
      PNGImage(this->seg_->width(),this->seg_->height(),this->seg_->nChl());

  // create a list of unused colors
  std::list<unsigned int> freeColors;
  for (unsigned int cIdx = 1; cIdx < colors_.size(); cIdx++)
    freeColors.push_back(cIdx);

  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator segColor =
      colorMap_.begin();
  typename std::map<MSK_ELM_TYPE, unsigned int>::iterator eraseItr;
  while (segColor != colorMap_.end()) {
    if (this->seg_->indices()->find(segColor->first) ==
        this->seg_->indices()->end()) {
      // segment inactive
      eraseItr = segColor++;
      colorMap_.erase(eraseItr);
    }
    else {
      // segment still active
      freeColors.remove(segColor->second);
      segColor++;
    }
  }

  // iterator used to iterate through the segment indices map
  typename std::map<MSK_ELM_TYPE, unsigned int>::const_iterator segment;

  // assign a color to each segment (black if there aren't enough colors)
  for (segment =  this->seg_->indices()->begin();
       segment != this->seg_->indices()->end(); segment++) {
    if (colorMap_.find(segment->first) == colorMap_.end()) {
      if (freeColors.empty()) {
        colorMap_[segment->first] = 0;
        std::cerr << "not enough colors, using black" << std::endl;
      }
      else {
        colorMap_[segment->first] = freeColors.back();
        freeColors.pop_back();
      }
    }
  }

  unsigned long cmp;
  for (unsigned int frm = 0; frm < this->seg_->nFrm(); frm++) {

    // draw each pixel of the visualization into the PNG image
    for (unsigned int pxl = 0;
         pxl < this->seg_->width()*this->seg_->height(); pxl++) {

      cmp = pxl*this->seg_->nChl();

      switch(ncut_Setting.visualisation_mode) {
        //case 0: is default
        case 1:   mode_original(segmentedImage, cmp, frm);
        break;
        default:  mode_segmentedObjects(segmentedImage, cmp, pxl, frm);
        break;
      }
    }


    if (frm == 1) {

      // save the current position of each active segment
      for (segment =  this->seg_->indices()->begin();
           segment != this->seg_->indices()->end(); segment++) {

        int x = static_cast<int>
            (round (this->seg_->locationX(segment->second)));
        int y = static_cast<int>
            (round (this->seg_->locationY(segment->second)));
#ifdef debugout
        std::cout << "seg " << segment->first << " - x " << x
            << " - y " << y << std::endl << std::flush;
#endif

        rememberedPosX_[segment->first].push_back(x);
        rememberedPosY_[segment->first].push_back(y);

        segmentSumArea_[segment->first] += this->seg_->nPixels(segment->second);

        if (rememberedPosX_[segment->first].size() == 1)
          firstFrame_[segment->first] = baseFrame_+this->seg()->nFrm()-1;

        // draw a cross at the current position of each segment
        if (ncut_Setting.drawCross == 1) {
          int k=-1;
          for (int j=-1; j < 2; j++,k++) { // draw cross
            segmentedImage->val(calculateElementNr(x+j,y,segmentedImage)) = 0;
            segmentedImage->val(calculateElementNr(x+j,y,segmentedImage)+1) = 0;
            segmentedImage->val(calculateElementNr(x+j,y,segmentedImage)+2) = 0;

            segmentedImage->val(calculateElementNr(x,y+k,segmentedImage)) = 0;
            segmentedImage->val(calculateElementNr(x,y+k,segmentedImage)+1) = 0;
            segmentedImage->val(calculateElementNr(x,y+k,segmentedImage)+2) = 0;
          }
        }
      }
    }

    // draw the rememberedPosition in Frame #1
    if (frm==1 && ncut_Setting.drawTrace==1) { 

      for (segment  = this->seg_->indices()->begin();
           segment != this->seg_->indices()->end(); segment++) {

        typename std::vector<int>::reverse_iterator rItrLocationY;
        typename std::vector<int>::reverse_iterator rItrLocationX;
        int count;
        for (rItrLocationY = rememberedPosY_[segment->first].rbegin(),
             rItrLocationX = rememberedPosX_[segment->first].rbegin(),
             count=0;
             rItrLocationY != rememberedPosY_[segment->first].rend() &&
             count < ncut_Setting.tracelength;
             rItrLocationY++, rItrLocationX++, count++) {

          segmentedImage->val(calculateElementNr(*rItrLocationX,
                                                 *rItrLocationY,
                                                 segmentedImage)  ) = 0;
          segmentedImage->val(calculateElementNr(*rItrLocationX,
                                                 *rItrLocationY,
                                                 segmentedImage)+1) = 0;
          segmentedImage->val(calculateElementNr(*rItrLocationX,
                                                 *rItrLocationY,
                                                 segmentedImage)+2) = 0;

        }
      }
    }

    //draw the rememberedPosition in Frame #1
    if (frm==1 && ncut_Setting.drawLine==1) {

      for (segment  = this->seg_->indices()->begin();
           segment != this->seg_->indices()->end(); segment++) {

        typename std::vector<int>::reverse_iterator rItrLocationY;
        typename std::vector<int>::reverse_iterator rItrLocationX;
        int count;
        for (rItrLocationY = rememberedPosY_[segment->first].rbegin(),
             rItrLocationX = rememberedPosX_[segment->first].rbegin(),
             count=0;
             rItrLocationY != rememberedPosY_[segment->first].rend() &&
             count < ncut_Setting.tracelength;
             count++) {

          int x1 = *rItrLocationX;
          int y2 = *rItrLocationY;
          rItrLocationX++;
          rItrLocationY++;

          if (rItrLocationX != rememberedPosX_[segment->first].rend()) {
            drawLine(x1,y2,*rItrLocationX,*rItrLocationY, segmentedImage);
          }
        }
      }
    }

    // save the visualization image in the visualization sequence
    seq.pushFrame(*segmentedImage);
  }
#ifdef debugout
  timerEnd = clock();
  std::cout << "done visualisation(" << ((double)(timerEnd - timerStart))/CLOCKS_PER_SEC
    << "seconds)" << std::endl << std::flush;
#endif

  // free dynamically allocated memory
  if (segmentedImage != NULL) {
    delete segmentedImage;
    segmentedImage = NULL;
  }
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
unsigned long PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::calculateElementNr(int x, int
y, const PNGImage* image)
{
  if (x<0 || y<0 || image==NULL) return 0;
  return image->nChl() * ((image->width() * y) + x);
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
int PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::drawLine(int x1, int y1, int x2, int y2,
PNGImage *image)
{
//   std::cout << "drawLine: x1:" << x1 << " y1:" << y1 << " x2:" << x2 << "
// y2:" << y2 << std::endl << std::flush; 
  if (x1 == x2 && y1 == y2) return 0;
  // bresenham algorithmus

  bool swap=false;

  if(abs(x1-x2)>abs(y1-y2)) {
    swap=(x1>x2);
  }
  else {
    swap=(y1>y2);
  }

  if(swap) {
    int i=x1;
    x1=x2;
    x2=i;

    i=y1;
    y1=y2;
    y2=i;
  }

  int dx=abs(x2-x1);
  int dy=abs(y2-y1);

  if(dx==0) {
    for(int y=y1; y<=y2; y++) {

      image->val(calculateElementNr(x1,y,image)) = 0;
      image->val(calculateElementNr(x1,y,image)+1) = 0;
      image->val(calculateElementNr(x1,y,image)+2) = 0;

    }
  }
  else if(dy==0) {
    for(int x=x1; x<=x2; x++) {
      image->val(calculateElementNr(x,y1,image)) = 0;
      image->val(calculateElementNr(x,y1,image)+1) = 0;
      image->val(calculateElementNr(x,y1,image)+2) = 0;
    }
  }
  else if(dx==dy) {
    int xi=(x1<x2 ? +1 : -1);
    int yi=(y1<y2 ? +1 : -1);

    for(int i=0,x=x1,y=y1; i<=dx; i++,x+=xi,y+=yi) {
      image->val(calculateElementNr(x,y,image)) = 0;
      image->val(calculateElementNr(x,y,image)+1) = 0;
      image->val(calculateElementNr(x,y,image)+2) = 0;

    }
  }
  else {
    bool horiz=(dx>dy);

    int p =(horiz ? 2*dy-dx : 2*dx-dy);
    int h1=(horiz ? 2*dy : 2*dx);
    int h2=(horiz ? 2*(dy-dx) : 2*(dx-dy));

    int dir=(horiz ? (y2>y1 ? 1 : -1) : (x2>x1 ? 1 : -1));

    int x=x1,y=y1;

    image->val(calculateElementNr(x,y,image)) = 0;
    image->val(calculateElementNr(x,y,image)+1) = 0;
    image->val(calculateElementNr(x,y,image)+2) = 0;

    while(horiz ? x<x2 : y<y2) {
      if(horiz) {
        x++;
      }
      else {
        y++;
      }

      if(p<0) {
        p+=h1;
      }
      else {
        p+=h2;

        if(horiz) {
          y+=dir;
        }
        else {
          x+=dir;
        }
      }
      image->val(calculateElementNr(x,y,image)) = 0;
      image->val(calculateElementNr(x,y,image)+1) = 0;
      image->val(calculateElementNr(x,y,image)+2) = 0;

    }
  }
  return 0;
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    mode_segmentedObjects(PNGImage* segmentedImage, int cmp, int pxl, int frm)
{
  PNG_color_t myColor;
  MSK_ELM_TYPE segId = this->seg_->segMask(this->seg_->elmMask(frm,pxl));

  if (this->seg_->age(segId) < ncut_Setting.segment_delay)
    myColor = colors_[ colorMap_[ this->seg_->simSeg( segId ) ] ];
  else
    myColor = colors_[ colorMap_[ segId ] ];

  segmentedImage->val(cmp)   = myColor.red;
  segmentedImage->val(cmp+1) = myColor.green;
  segmentedImage->val(cmp+2) = myColor.blue;

}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
void PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    mode_original(PNGImage* segmentedImage, int cmp, int frm)
{
  segmentedImage->val(cmp)   = this->seg_->frameMask(frm,cmp);
  segmentedImage->val(cmp+1) = this->seg_->frameMask(frm,cmp+1);
  segmentedImage->val(cmp+2) = this->seg_->frameMask(frm,cmp+2);
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::
    PNGVis(const Segmentation<IMG_ELM_TYPE, MSK_ELM_TYPE>* _seg)
  : Vis<IMG_ELM_TYPE, MSK_ELM_TYPE, PNG_IMAGE_DATATYPE>(_seg),
    baseFrame_(-1),
    colors_(),
    colorMap_(),
    rememberedPosX_(),
    rememberedPosY_(),
    firstFrame_()
{
  if (this->seg_ == NULL)
    throw ncutCritical("invalid segmentation (NULL)");

  PNG_color_t color;

  color.red   = 0;
  color.green = 0;
  color.blue  = 0;
  colors_.push_back(color);

  // alles + 1
  // 0
  color.red   = 255;
  color.green = 255;
  color.blue  = 255;
  colors_.push_back(color);

  // 1
  color.red   = 88;
  color.green = 88;
  color.blue  = 88;
  colors_.push_back(color);
  // 2
  color.red   = 128;
  color.green = 0;
  color.blue  = 128;
  colors_.push_back(color);

  // 3
  color.red   = 255;
  color.green = 168;
  color.blue  = 88;
  colors_.push_back(color);
  // 4
  color.red   = 220;
  color.green = 220;
  color.blue  = 220;
  colors_.push_back(color);
  // 5
  color.red   = 195;
  color.green = 195;
  color.blue  = 195;
  colors_.push_back(color);
  // 6
  color.red   = 160;
  color.green = 160;
  color.blue  = 160;
  colors_.push_back(color);

  // 7
  color.red   = 128;
  color.green = 128;
  color.blue  = 128;
  colors_.push_back(color);
  // 8
  color.red   = 128;
  color.green = 0;
  color.blue  = 0;
  colors_.push_back(color);
  // 9
  color.red   = 192;
  color.green = 0;
  color.blue  = 0;
  colors_.push_back(color);
  // 10
  color.red   = 0;
  color.green = 128;
  color.blue  = 0;
  colors_.push_back(color);
  // 11
  color.red   = 255;
  color.green = 0;
  color.blue  = 0;
  colors_.push_back(color);
  // 12
  color.red   = 255;
  color.green = 192;
  color.blue  = 192;
  colors_.push_back(color);
  // 13
  color.red   = 0;
  color.green = 64;
  color.blue  = 0;
  colors_.push_back(color);
  // 14
  color.red   = 0;
  color.green = 192;
  color.blue  = 0;
  colors_.push_back(color);
  // 15
  color.red   = 0;
  color.green = 255;
  color.blue  = 0;
  colors_.push_back(color);

  // 16
  color.red   = 192;
  color.green = 255;
  color.blue  = 192;
  colors_.push_back(color);

  color.red   = 0;
  color.green = 0;
  color.blue  = 128;
  colors_.push_back(color);

  color.red   = 0;
  color.green = 0;
  color.blue  = 192;
  colors_.push_back(color);

  color.red   = 0;
  color.green = 0;
  color.blue  = 255;
  colors_.push_back(color);

  color.red   = 192;
  color.green = 192;
  color.blue  = 255;
  colors_.push_back(color);

  color.red   = 64;
  color.green = 64;
  color.blue  = 0;
  colors_.push_back(color);

  color.red   = 128;
  color.green = 128;
  color.blue  = 0;
  colors_.push_back(color);

  color.red   = 192;
  color.green = 192;
  color.blue  = 0;
  colors_.push_back(color);

  color.red   = 255;
  color.green = 255;
  color.blue  = 0;
  colors_.push_back(color);

  color.red   = 255;
  color.green = 255;
  color.blue  = 192;
  colors_.push_back(color);

  color.red   = 0;
  color.green = 128;
  color.blue  = 128;
  colors_.push_back(color);

  color.red   = 0;
  color.green = 192;
  color.blue  = 192;
  colors_.push_back(color);

  color.red   = 192;
  color.green = 255;
  color.blue  = 255;
  colors_.push_back(color);

  color.red   = 192;
  color.green = 0;
  color.blue  = 192;
  colors_.push_back(color);

  color.red   = 255;
  color.green = 0;
  color.blue  = 255;
  colors_.push_back(color);

  color.red   = 255;
  color.green = 192;
  color.blue  = 255;
  colors_.push_back(color);

  color.red   = 192;
  color.green = 88;
  color.blue  = 0;
  colors_.push_back(color);

  color.red   = 255;
  color.green = 128;
  color.blue  = 0;
  colors_.push_back(color);

  color.red   = 255;
  color.green = 220;
  color.blue  = 168;
  colors_.push_back(color);
}

template <class IMG_ELM_TYPE, class MSK_ELM_TYPE>
PNGVis<IMG_ELM_TYPE, MSK_ELM_TYPE>::~PNGVis()
{
  return;
}

} // namespace ncut

#endif // NCUT_VIS_H
